@@ -1,9 +1,9 @@
 #ifndef _server_h
 #define _server_h
 
-#include <gnustep/base/preface.h>
+#include <base/preface.h>
 #include <Foundation/NSConnection.h>
-#include <gnustep/base/Array.h>
+#include <base/Array.h>
 
 typedef struct _small_struct {
   unsigned char z;
@@ -36,12 +36,18 @@ struct myarray {
 - sendStruct: (foo)f;
 - sendSmallStruct: (small_struct)small;
 - (foo) returnStruct;
+- (foo) returnSetStruct: (int)x;
+- (small_struct) returnSmallStruct;
+- (small_struct) returnSetSmallStruct: (int)x;
 - sendArray: (int[3])a;
 - sendStructArray: (struct myarray)ma;
 - sendDouble: (double)d andFloat: (float)f;
 - (double*) doDoublePointer: (double*)d;
 - sendCharPtrPtr: (char**)sp;
 - sendBycopy: (bycopy id)o;
+#ifdef	_F_BYREF
+- sendByref: (byref id)o;
+#endif
 - manyArgs: (int)i1 : (int)i2 : (int)i3 : (int)i4 : (int)i5 : (int)i6
 : (int)i7 : (int)i8 : (int)i9 : (int)i10 : (int)i11 : (int)i12;
 - (float) returnFloat;
