@@ -22,12 +22,12 @@
    */ 
 
 #include <config.h>
-#include <gnustep/base/preface.h>
-#include <gnustep/base/Coder.h>
-#include <gnustep/base/CoderPrivate.h>
-#include <gnustep/base/MemoryStream.h>
-#include <gnustep/base/StdioStream.h>
-#include <gnustep/base/BinaryCStream.h>
+#include <base/preface.h>
+#include <base/Coder.h>
+#include <base/CoderPrivate.h>
+#include <base/MemoryStream.h>
+#include <base/StdioStream.h>
+#include <base/BinaryCStream.h>
 #include <Foundation/NSArchiver.h>
 #include <Foundation/NSException.h>
 
@@ -67,7 +67,8 @@ my_object_is_class(id object)
     {
       /* This code has not yet been ported to machines for which
 	 a pointer is not the same size as an int. */
-      assert(sizeof(void*) == sizeof(unsigned)); 
+      NSAssert(sizeof(void*) == sizeof(unsigned),
+	@"Pointer and int are different sizes"); 
 
       /* Initialize some defaults. */
       default_stream_class = [MemoryStream class];
@@ -275,7 +276,7 @@ my_object_is_class(id object)
 			NSIntMapValueCallBacks, 0);
 
   xref = NSCountMapTable (const_ptr_2_xref) + 1;
-  assert (! NSMapGet (const_ptr_2_xref, (void*)xref));
+  NSAssert (! NSMapGet (const_ptr_2_xref, (void*)xref), @"xref already in Map");
   NSMapInsert (const_ptr_2_xref, ptr, (void*)xref);
   return xref;
 }
@@ -299,7 +300,7 @@ my_object_is_class(id object)
       NSCreateMapTable (NSNonOwnedPointerOrNullMapKeyCallBacks,
 			NSIntMapValueCallBacks, 0);
   fref = ++fref_counter;
-  assert ( ! NSMapGet (object_2_fref, anObject));
+  NSAssert ( ! NSMapGet (object_2_fref, anObject), @"anObject already in Map");
   NSMapInsert (object_2_fref, anObject, (void*)fref);
   return fref;
 }
@@ -434,8 +435,8 @@ my_object_is_class(id object)
 	  /* It hasn't been encoded before; encode it. */
 	  int class_version = class_get_version (aClass);
 
-	  assert (class_name);
-	  assert (*class_name);
+	  NSAssert (class_name, @"Class doesn't have a name");
+	  NSAssert (*class_name, @"Class name is empty");
 
 	  [self encodeTag: CODER_CLASS];
 	  [self encodeValueOfCType: @encode(char*)
@@ -502,8 +503,7 @@ my_object_is_class(id object)
 	    [NSException raise: NSGenericException
 			 format: @"ObjC runtime didn't provide SEL name"];
 	  if (!sel_types || !*sel_types)
-	    [NSException raise: NSGenericException
-			 format: @"ObjC runtime didn't provide SEL type"];
+	    sel_types = NO_SEL_TYPES;
 
 	  [self _coderCreateReferenceForConstPtr: sel];
 	  [self encodeValueOfCType: @encode(char*) 
@@ -560,7 +560,7 @@ my_object_is_class(id object)
      encode here any forward-referenced objects that haven't been
      encoded yet.  No---the current behavior implements NeXT's
      -encodeConditionalObject: */
-  assert (interconnect_stack_height);
+  NSParameterAssert (interconnect_stack_height);
   interconnect_stack_height--;
 }
 
@@ -578,7 +578,7 @@ my_object_is_class(id object)
 }
 
 
-/* These next two methods are the designated coder methods called when
+/* These next three methods are the designated coder methods called when
    we've determined that the object has not already been
    encoded---we're not simply going to encode a cross-reference number
    to the object, we're actually going to encode an object (either a
@@ -611,17 +611,24 @@ my_object_is_class(id object)
   [encoded_object encodeWithCoder: (id)self];
 }
 
-/* This method overridden by ConnectedCoder */
+/* This method overridden by NSPortCoder */
 - (void) _doEncodeObject: anObj
 {
   [self _doEncodeBycopyObject:anObj];
 }
 
+/* This method overridden by NSPortCoder */
+- (void) _doEncodeByrefObject: anObj
+{
+  [self _doEncodeObject: anObj];
+}
+
 
 /* This is the designated object encoder */
 - (void) _encodeObject: anObj
    withName: (NSString*) name
    isBycopy: (BOOL) bycopy_flag
+   isByref: (BOOL) byref_flag
    isForwardReference: (BOOL) forward_ref_flag
 {
   [self encodeName:name];
@@ -681,6 +688,8 @@ my_object_is_class(id object)
 	  [self encodeIndent];
 	  if (bycopy_flag)
 	    [self _doEncodeBycopyObject:anObj];
+	  else if (byref_flag)
+	    [self _doEncodeByrefObject:anObj];
 	  else
 	    [self _doEncodeObject:anObj];	    
 	  [self encodeUnindent];
@@ -730,20 +739,42 @@ my_object_is_class(id object)
 - (void) encodeObject: anObj
    withName: (NSString*)name
 {
-  [self _encodeObject:anObj withName:name isBycopy:NO isForwardReference:NO];
+  [self _encodeObject:anObj
+	     withName:name
+	     isBycopy:NO
+	      isByref:NO
+   isForwardReference:NO];
 }
 
 
 - (void) encodeBycopyObject: anObj
    withName: (NSString*)name
 {
-  [self _encodeObject:anObj withName:name isBycopy:YES isForwardReference:NO];
+  [self _encodeObject:anObj
+	     withName:name
+	     isBycopy:YES
+	      isByref:NO
+   isForwardReference:NO];
+}
+
+- (void) encodeByrefObject: anObj
+   withName: (NSString*)name
+{
+  [self _encodeObject:anObj
+	     withName:name
+	     isBycopy:NO
+	      isByref:YES
+   isForwardReference:NO];
 }
 
 - (void) encodeObjectReference: anObj
    withName: (NSString*)name
 {
-  [self _encodeObject:anObj withName:name isBycopy:NO isForwardReference:YES];
+  [self _encodeObject:anObj
+	     withName:name
+	     isBycopy:NO
+	      isByref:NO
+   isForwardReference:YES];
 }
 
 
@@ -823,7 +854,7 @@ my_object_is_class(id object)
 
 #if 0
   if (classname_2_classname)
-    return NSMapGet (classname_2_classname, [trueName cStringNoCopy]);
+    return NSMapGet (classname_2_classname, [trueName cString]);
   return trueName;
 #endif
 }
