@@ -1,5 +1,5 @@
 /* NSArray - Array object to hold other objects.
-   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
+   Copyright (C) 1995, 1996, 1998 Free Software Foundation, Inc.
    
    Written by:  Andrew Kachites McCallum <mccallum@gnu.ai.mit.edu>
    From skeleton by:  Adam Fedor <fedor@boulder.colorado.edu>
@@ -27,8 +27,9 @@
    */
 
 #include <config.h>
-#include <gnustep/base/behavior.h>
+#include <base/behavior.h>
 #include <Foundation/NSArray.h>
+#include <Foundation/NSCoder.h>
 #include <Foundation/NSString.h>
 #include <Foundation/NSGArray.h>
 #include <limits.h>
@@ -36,6 +37,8 @@
 #include <Foundation/NSException.h>
 #include <Foundation/NSAutoreleasePool.h>
 
+#include <base/fast.x>
+
 @class NSArrayEnumerator;
 @class NSArrayEnumeratorReverse;
 
@@ -85,30 +88,6 @@ static Class NSMutableArray_concrete_class;
   return NSAllocateObject ([self _concreteClass], 0, z);
 }
 
-/* This is the designated initializer for NSArray. */
-- initWithObjects: (id*)objects count: (unsigned)count
-{
-  [self subclassResponsibility:_cmd];
-  return nil;
-}
-
-- (unsigned) count
-{
-  [self subclassResponsibility:_cmd];
-  return 0;
-}
-
-- objectAtIndex: (unsigned)index
-{
-  [self subclassResponsibility:_cmd];
-  return nil;
-}
-
-@end
-
-
-@implementation NSArrayNonCore
-
 + array
 {
   return [[[self alloc] init] 
@@ -134,6 +113,110 @@ static Class NSMutableArray_concrete_class;
 	  autorelease];
 }
 
+/* This is the designated initializer for NSArray. */
+- initWithObjects: (id*)objects count: (unsigned)count
+{
+  [self subclassResponsibility:_cmd];
+  return nil;
+}
+
+- (unsigned) count
+{
+  [self subclassResponsibility:_cmd];
+  return 0;
+}
+
+- objectAtIndex: (unsigned)index
+{
+  [self subclassResponsibility:_cmd];
+  return nil;
+}
+
+/* The NSCoding Protocol */
+
+- (void) encodeWithCoder: (NSCoder*)aCoder
+{
+    unsigned	count = [self count];
+
+    [aCoder encodeValueOfObjCType: @encode(unsigned)
+			       at: &count];
+
+    if (count > 0) {
+        unsigned	i;
+
+	for (i = 0; i < count; i++) {
+	    [aCoder encodeObject: [self objectAtIndex: i]];
+	}
+    }
+}
+
+- (id) initWithCoder: (NSCoder*)aCoder
+{
+    unsigned    count;
+
+    if ([aCoder systemVersion] == 0) {
+	unsigned dummy;
+	[aCoder decodeValueOfObjCType: @encode(unsigned) at: &dummy];
+	[aCoder decodeValueOfObjCType: @encode(unsigned) at: &dummy];
+    }
+    [aCoder decodeValueOfObjCType: @encode(unsigned) at: &count];
+    if (count > 0) {
+	id	contents[count];
+	int	i;
+
+	for (i = 0; i < count; i++) {
+	    contents[i] = [aCoder decodeObject];
+	}
+	return [self initWithObjects: contents count: count];
+    }
+    return [self initWithObjects: 0 count: 0];
+}
+
+/* The NSCopying Protocol */
+
+- copyWithZone: (NSZone*)zone
+{
+  /* a deep copy */
+  unsigned count = [self count];
+  id oldObjects[count];
+  id newObjects[count];
+  id newArray;
+  unsigned i;
+  BOOL needCopy = [self isKindOfClass: [NSMutableArray class]];
+
+  if (NSShouldRetainWithZone(self, zone) == NO)
+    needCopy = YES;
+  [self getObjects: oldObjects];
+  for (i = 0; i < count; i++)
+    {
+      newObjects[i] = [oldObjects[i] copyWithZone:zone];
+      if (newObjects[i] != oldObjects[i])
+	needCopy = YES;
+    }
+  if (needCopy)
+    newArray = [[[[self class] _concreteClass] allocWithZone:zone]
+	      initWithObjects:newObjects count:count];
+  else
+    newArray = [self retain];
+  for (i = 0; i < count; i++)
+    [newObjects[i] release];
+  return newArray;
+}
+
+/* The NSMutableCopying Protocol */
+
+- mutableCopyWithZone: (NSZone*)zone
+{
+  /* a shallow copy */
+  return [[[[self class] _mutableConcreteClass] allocWithZone:zone] 
+	  initWithArray:self];
+}
+
+@end
+
+
+@implementation NSArrayNonCore
+
 - (NSArray*) arrayByAddingObject: anObject
 {
   id na;
@@ -141,31 +224,31 @@ static Class NSMutableArray_concrete_class;
   id *objects;
  
   c = [self count];
-  OBJC_MALLOC (objects, id, c+1);
-  for (i = 0; i < c; i++)
-    objects[i] = [self objectAtIndex: i];
-  objects[c] = anObject;
-  na = [[NSArray alloc] initWithObjects: objects count: c+1];
-  OBJC_FREE (objects);
+  {
+    id	objects[c+1];
+
+    [self getObjects: objects];
+    objects[c] = anObject;
+    na = [[NSArray alloc] initWithObjects: objects count: c+1];
+  }
   return [na autorelease];
 }
 
 - (NSArray*) arrayByAddingObjectsFromArray: (NSArray*)anotherArray
 {
-  id na;
-  unsigned i, c, l;
-  id *objects;
+    id		na;
+    unsigned	c, l;
  
-  c = [self count];
-  l = [anotherArray count];
-  OBJC_MALLOC (objects, id, c+l);
-  for (i = 0; i < c; i++)
-    objects[i] = [self objectAtIndex: i];
-  for (i = c; i < c+l; i++)
-    objects[i] = [anotherArray objectAtIndex: i-c];
-  na = [[NSArray alloc] initWithObjects: objects count: c+l];
-  OBJC_FREE (objects);
-  return [na autorelease];
+    c = [self count];
+    l = [anotherArray count];
+    {
+	id	objects[c+l];
+
+	[self getObjects: objects];
+	[anotherArray getObjects: &objects[c]];
+	na = [NSArray arrayWithObjects: objects count: c+l];
+    }
+    return na;
 }
 
 - initWithObjects: firstObject rest: (va_list) ap
@@ -178,8 +261,8 @@ static Class NSMutableArray_concrete_class;
   auto		id			tmpId;
 
   /*	Do initial allocation.	*/
-  prevSize = 1;
-  curSize  = 2;
+  prevSize = 3;
+  curSize  = 5;
   OBJC_MALLOC(objsArray, id, curSize);
   tmpId = firstObject;
 
@@ -223,6 +306,25 @@ static Class NSMutableArray_concrete_class;
   return self;
 }
 
+- initWithContentsOfFile: (NSString*)file
+{
+  NSString 	*myString;
+
+  myString = [[NSString alloc] initWithContentsOfFile:file];
+  if (myString)
+    {
+      id result = [myString propertyList];
+      if ( [result isKindOfClass: [NSArray class]] )
+	{
+	  [self initWithArray: result];
+	  return self;
+	}
+    }
+  NSLog(@"Contents of file does not contain an array");
+  [self dealloc];
+  return nil;
+}
+
 + arrayWithObjects: firstObject, ...
 {
   va_list ap;
@@ -240,16 +342,16 @@ static Class NSMutableArray_concrete_class;
 
 - initWithArray: (NSArray*)array
 {
-  unsigned i, c;
-  id *objects;
+    unsigned c;
  
-  c = [array count];
-  OBJC_MALLOC(objects, id, c);
-  for (i = 0; i < c; i++)
-    objects[i] = [array objectAtIndex:i];
-  self = [self initWithObjects:objects count:c];
-  OBJC_FREE(objects);
-  return self;
+    c = [array count];
+    {
+	id	objects[c];
+
+	[array getObjects: objects];
+	self = [self initWithObjects: objects count: c];
+    }
+    return self;
 }
 
 - (void) getObjects: (id*)aBuffer
@@ -268,6 +370,11 @@ static Class NSMutableArray_concrete_class;
     aBuffer[j++] = [self objectAtIndex: i];
 }
 
+- (unsigned) hash
+{
+    return [self count];
+}
+
 - (unsigned) indexOfObjectIdenticalTo:anObject
 {
   unsigned i, c = [self count];
@@ -345,36 +452,35 @@ static Class NSMutableArray_concrete_class;
   return [self objectAtIndex: count-1];
 }
 
-- (void) makeObjectsPerform: (SEL)aSelector
-{
-  unsigned i = [self count];
-  while (i-- > 0)
-    [[self objectAtIndex:i] perform:aSelector];
-}
-
 - (void) makeObjectsPerformSelector: (SEL)aSelector
-{
-   [self makeObjectsPerform: aSelector];
-}
-
-- (void) makeObjectsPerform: (SEL)aSelector withObject:argument
 {
   unsigned i = [self count];
   while (i-- > 0)
-    [[self objectAtIndex:i] perform:aSelector withObject:argument];
+    [[self objectAtIndex:i] performSelector:aSelector];
+}
+
+- (void) makeObjectsPerform: (SEL)aSelector
+{
+   [self makeObjectsPerformSelector: aSelector];
 }
 
 - (void) makeObjectsPerformSelector: (SEL)aSelector withObject:argument
 {
-   [self makeObjectsPerform: aSelector withObject:argument];
+  unsigned i = [self count];
+  while (i-- > 0)
+    [[self objectAtIndex:i] performSelector:aSelector withObject:argument];
 }
 
+- (void) makeObjectsPerform: (SEL)aSelector withObject:argument
+{
+   [self makeObjectsPerformSelector: aSelector withObject:argument];
+}
 
 - (NSArray*) sortedArrayUsingSelector: (SEL)comparator
 {
   int compare(id elem1, id elem2, void* context)
     {
-      return (int)[elem1 perform:comparator withObject:elem2];
+      return (int)[elem1 performSelector:comparator withObject:elem2];
     }
 
     return [self sortedArrayUsingFunction:compare context:NULL];
@@ -465,16 +571,13 @@ static Class NSMutableArray_concrete_class;
   else
     j = range.location + range.length - 1;
 
-  // Copy the ids from the range into a temporary array
-  OBJC_MALLOC(objects, id, j-i+1);
-  for (k = i; k <= j; k++)
-    objects[k-i] = [self objectAtIndex:k];
-
-  // Create the new array
-  na = [[NSArray alloc] initWithObjects:objects count:j-i+1];
-  OBJC_FREE(objects);
-  return [na autorelease];
+  {
+    id	objects[j-i+1];
 
+    [self getObjects: objects range: NSMakeRange(i, j-i+1)];
+    na = [NSArray arrayWithObjects: objects count: j-i+1];
+  }
+  return na;
 }
 
 - (NSEnumerator*) objectEnumerator
@@ -502,141 +605,90 @@ static Class NSMutableArray_concrete_class;
 - (NSString*) descriptionWithLocale: (NSDictionary*)locale
 			     indent: (unsigned int)level
 {
-    NSMutableString	*result;
-    unsigned		size;
-    unsigned		indentSize;
-    unsigned		indentBase;
-    NSMutableString	*iBaseString;
-    NSMutableString	*iSizeString;
-    NSAutoreleasePool	*arp = [[NSAutoreleasePool alloc] init];
-    unsigned		count = [self count];
-    NSString		*plists[count];
-    unsigned		i;
+  NSMutableString	*result;
 
-    [self getObjects: plists];
-
-    /*
-     *	Indentation is at four space intervals using tab characters to
-     *	replace multiples of eight spaces.
-     *
-     *	We work out the sizes of the strings needed to perform indentation for
-     *	this level and build strings to make up the indentation.
-     */
-    indentBase = level << 2;
-    count = indentBase >> 3;
-    if ((indentBase % 8) == 0) {
-	indentBase = count;
-    }
-    else {
-	indentBase == count + 4;
-    }
-    iBaseString = [NSMutableString stringWithCapacity: indentBase];
-    for (i = 0; i < count; i++) {
-	[iBaseString appendString: @"\t"];
-    }
-    if (count != indentBase) {
-	[iBaseString appendString: @"    "];
-    }
-
-    level++;
-    indentSize = level << 2;
-    count = indentSize >> 3;
-    if ((indentSize % 8) == 0) {
-	indentSize = count;
-    }
-    else {
-	indentSize == count + 4;
-    }
-    iSizeString = [NSMutableString stringWithCapacity: indentSize];
-    for (i = 0; i < count; i++) {
-	[iSizeString appendString: @"\t"];
-    }
-    if (count != indentSize) {
-	[iSizeString appendString: @"    "];
-    }
-
-    /*
-     *	Basic size is - opening bracket, newline, closing bracket,
-     *	indentation for the closing bracket, and a nul terminator.
-     */
-    size = 4 + indentBase;
-
-    count = [self count];
-    for (i = 0; i < count; i++) {
-	id		item;
-
-	item = plists[i];
-	if ([item isKindOfClass: [NSString class]]) {
-	   item = [item descriptionForPropertyList];
-	}
-	else if ([item respondsToSelector:
-		@selector(descriptionWithLocale:indent:)]) {
-	   item = [item descriptionWithLocale: locale indent: level];
-	}
-	else if ([item respondsToSelector:
-		@selector(descriptionWithLocale:)]) {
-	   item = [item descriptionWithLocale: locale];
-	}
-	else {
-	   item = [item description];
-	}
-	plists[i] = item;
-
-	size += [item length] + indentSize;
-	if (i == count - 1) {
-	    size += 1;			/* newline	*/
-	}
-	else {
-	    size += 2;			/* ',' and newline	*/
-	}
-    }
-
-    result = [[NSMutableString alloc] initWithCapacity: size];
-    [result appendString: @"(\n"];
-    for (i = 0; i < count; i++) {
-	[result appendString: iSizeString];
-	[result appendString: plists[i]];
-	if (i == count - 1) {
-            [result appendString: @"\n"];
-	}
-	else {
-            [result appendString: @",\n"];
-	}
-    }
-    [result appendString: iBaseString];
-    [result appendString: @")"];
-
-    [arp release];
-
-    return [result autorelease];
+  result = [NSMutableString stringWithCapacity: 20*[self count]];
+  [self descriptionWithLocale: locale
+		       indent: level
+			   to: (id<GNUDescriptionDestination>)result];
+  return result;
 }
 
-/* The NSCopying Protocol */
+static NSString	*indentStrings[] = {
+  @"",
+  @"    ",
+  @"\t",
+  @"\t    ",
+  @"\t\t",
+  @"\t\t    ",
+  @"\t\t\t",
+  @"\t\t\t    ",
+  @"\t\t\t\t",
+  @"\t\t\t\t    ",
+  @"\t\t\t\t\t",
+  @"\t\t\t\t\t    ",
+  @"\t\t\t\t\t\t"
+};
 
-- (id) copy
+- (void) descriptionWithLocale: (NSDictionary*)locale
+		        indent: (unsigned int)level
+			    to: (id<GNUDescriptionDestination>)result
 {
-    return [self copyWithZone:NSDefaultMallocZone()];
-}
+  NSString		*iBaseString;
+  NSString		*iSizeString;
+  unsigned		count = [self count];
+  NSString		*plists[count];
+  unsigned		i;
+  SEL			appSel;
+  IMP			appImp;
 
-- copyWithZone: (NSZone*)zone
-{
-  /* a deep copy */
-  unsigned count = [self count];
-  id objects[count];
-  unsigned i;
+  appSel = @selector(appendString:);
+  appImp = [(NSObject*)result methodForSelector: appSel];
+
+  if (level < sizeof(indentStrings)/sizeof(NSString*))
+    iBaseString = indentStrings[level];
+  else
+    iBaseString = indentStrings[sizeof(indentStrings)/sizeof(NSString*)-1];
+  level++;
+  if (level < sizeof(indentStrings)/sizeof(NSString*))
+    iSizeString = indentStrings[level];
+  else
+    iSizeString = indentStrings[sizeof(indentStrings)/sizeof(NSString*)-1];
+
+  (*appImp)(result, appSel, @"(\n");
+
+  [self getObjects: plists];
   for (i = 0; i < count; i++)
-    objects[i] = [[self objectAtIndex:i] copyWithZone:zone];
-  return [[[[self class] _concreteClass] allocWithZone:zone] 
-	  initWithObjects:objects count:count];
-}
+    {
+      id	item = plists[i];
 
-/* The NSMutableCopying Protocol */
-
-- mutableCopyWithZone: (NSZone*)zone
-{
-  /* a shallow copy */
-  return [[[[self class] _mutableConcreteClass] allocWithZone:zone] 
-	  initWithArray:self];
+      (*appImp)(result, appSel, iSizeString);
+ 
+      if ([item respondsToSelector:
+	      @selector(descriptionWithLocale:indent:)])
+	{
+	  [item descriptionWithLocale: locale indent: level to: result];
+	}
+      else if ([item respondsToSelector:
+	      @selector(descriptionWithLocale:)])
+	{
+	  [item descriptionWithLocale: locale to: result];
+	}
+      else
+	{
+	  [item descriptionTo: result];
+	}
+      if (i == count - 1)
+	{
+	  (*appImp)(result, appSel, @"\n");
+	}
+      else
+	{
+	  (*appImp)(result, appSel, @",\n");
+	}
+    }
+  (*appImp)(result, appSel, iBaseString);
+  (*appImp)(result, appSel, @")");
 }
 
 @end
@@ -678,15 +730,15 @@ static Class NSMutableArray_concrete_class;
 - (void) replaceObjectsInRange: (NSRange)aRange
 	  withObjectsFromArray: (NSArray*)anArray
 {
-  unsigned	i;
+  id e, o;
 
-  if ([self count] <= aRange.location)
+  if ([self count] < (aRange.location + aRange.length))
     [NSException raise: NSRangeException
 		 format: @"Replacing objects beyond end of array."];
   [self removeObjectsInRange: aRange];
-  i = [anArray count];
-  while (i-- > 0)
-    [self insertObject: [anArray objectAtIndex: i] atIndex: aRange.location];
+  e = [anArray reverseObjectEnumerator];
+  while ((o = [e nextObject]))
+    [self insertObject: o atIndex: aRange.location];
 }
 
 - (void) replaceObjectsInRange: (NSRange)aRange
@@ -718,24 +770,6 @@ static Class NSMutableArray_concrete_class;
 	  autorelease];
 }
 
-- initWithContentsOfFile: (NSString*)file
-{
-  NSString 	*myString;
-
-  myString = [[NSString alloc] initWithContentsOfFile:file];
-  if (myString)
-    {
-      id result = [myString propertyList];
-      if ( [result isKindOfClass: [NSArray class]] )
-	{
-	  [self initWithArray: result];
-	  return self;
-	}
-    }
-  [self dealloc];
-  return nil;
-}
-
 - (BOOL)writeToFile:(NSString *)path atomically:(BOOL)useAuxiliaryFile
 {
   return [[self description] writeToFile:path atomically:useAuxiliaryFile];
@@ -744,7 +778,6 @@ static Class NSMutableArray_concrete_class;
 /* Override our superclass's designated initializer to go our's */
 - initWithObjects: (id*)objects count: (unsigned)count
 {
-  /* xxx Could be made more efficient by increasing capacity all at once. */
   unsigned i;
   self = [self initWithCapacity: count];
   for (i = 0; i < count; i++)
@@ -849,7 +882,6 @@ static Class NSMutableArray_concrete_class;
 
 - (void) addObjectsFromArray: (NSArray*)otherArray
 {
-  /* xxx Could be made more efficient by increasing capacity all at once. */
   unsigned i, c = [otherArray count];
   for (i = 0; i < c; i++)
     [self addObject: [otherArray objectAtIndex: i]];
@@ -864,12 +896,6 @@ static Class NSMutableArray_concrete_class;
 - (void) removeObjectsFromIndices: (unsigned*)indices 
 		       numIndices: (unsigned)count
 {
-  int compare_unsigned(const void *u1, const void *u2)
-    {
-      return *((int*)u1) - *((int*)u2);
-    }
-  /* xxx are we allowed to modify the contents of indices? */
-  qsort(indices, count, sizeof(unsigned), compare_unsigned);
   while (count--)
     [self removeObjectAtIndex:indices[count]];
 }
@@ -895,7 +921,7 @@ static Class NSMutableArray_concrete_class;
 {
   int compare(id elem1, id elem2, void* context)
     {
-      return (int)[elem1 perform:comparator withObject:elem2];
+      return (int)[elem1 performSelector:comparator withObject:elem2];
     }
 
     [self sortUsingFunction:compare context:NULL];
