@@ -31,12 +31,12 @@
    It's an instance of PortEncoder or PortDecoder. */
 
 #include <config.h>
-#include <gnustep/base/preface.h>
+#include <base/preface.h>
 #include <Foundation/DistributedObjects.h>
-#include <gnustep/base/TcpPort.h>
-#include <gnustep/base/mframe.h>
-#include <gnustep/base/Notification.h>
-#include <gnustep/base/MallocAddress.h>
+#include <base/TcpPort.h>
+#include <mframe.h>
+#include <base/NotificationDispatcher.h>
+#include <Foundation/NSData.h>
 #include <Foundation/NSRunLoop.h>
 #include <Foundation/NSArray.h>
 #include <Foundation/NSDictionary.h>
@@ -44,8 +44,9 @@
 #include <Foundation/NSString.h>
 #include <Foundation/NSDate.h>
 #include <Foundation/NSException.h>
+#include <Foundation/NSLock.h>
 #include <Foundation/NSThread.h>
-#include <assert.h>
+#include <Foundation/NSNotification.h>
 
 NSString* NSConnectionReplyMode = @"NSConnectionReplyMode";
 
@@ -61,6 +62,22 @@ NSString *NSConnectionRequestsSent = @"NSConnectionRequestsSent";
 NSString *NSConnectionLocalCount = @"NSConnectionLocalCount";
 NSString *NSConnectionProxyCount = @"NSConnectionProxyCount";
 
+@interface	NSDistantObject (NSConnection)
+- (BOOL) isVended;
+- (void) setVended;
+@end
+
+@implementation	NSDistantObject (NSConnection)
+- (BOOL) isVended
+{
+    return _isVended;
+}
+- (void) setVended
+{
+    _isVended = YES;
+}
+@end
+
 /*
  *	ConnectionLocalCounter is a trivial class to keep track of how
  *	many different connections a particular local object is vended
@@ -104,6 +121,56 @@ NSString *NSConnectionProxyCount = @"NSConnectionProxyCount";
 }
 @end
 
+
+
+/*
+ *	CachedLocalObject is a trivial class to keep track of how
+ *	many different connections a particular local object is vended
+ *	over.  This is required so that we know when to remove an object
+ *	from the global list when it is removed from the list of objects
+ *	vended on a particular connection.
+ */
+@interface	CachedLocalObject : NSObject
+{
+    id	obj;
+    int time;
+}
+- (BOOL)countdown;
+- (id) obj;
++ (CachedLocalObject*) itemWithObject: (id)o time: (int)t;
+@end
+
+@implementation	CachedLocalObject
+
++ (CachedLocalObject*) itemWithObject: (id)o time: (int)t
+{
+    CachedLocalObject	*item = [[self alloc] init];
+
+    item->obj = [o retain];
+    item->time = t;
+    return [item autorelease];
+}
+
+- (void) dealloc
+{
+    [obj release];
+    [super dealloc];
+}
+
+- (BOOL) countdown
+{
+    if (time-- > 0)
+	return YES;
+    return NO;
+}
+
+- (id) obj
+{
+    return obj;
+}
+
+@end
+
 
 @interface NSConnection (GettingCoderInterface)
 - (void) _handleRmc: rmc;
@@ -116,6 +183,7 @@ NSString *NSConnectionProxyCount = @"NSConnectionProxyCount";
 
 @interface NSConnection (Private)
 - _superInit;
++ setDebug: (int)val;
 @end
 
 #define proxiesHashGate refGate
@@ -167,6 +235,7 @@ static id default_encoding_class;
 static id default_decoding_class;
 static int default_reply_timeout;
 static int default_request_timeout;
+static NSTimer *timer;
 
 static int debug_connection = 0;
 
@@ -178,20 +247,21 @@ static int debug_connection = 0;
 /* We could write -hash and -isEqual implementations for NSConnection */
 static NSMutableArray *connection_array;
 static NSMutableArray *not_owned;
-static Lock *connection_array_gate;
+static NSLock *connection_array_gate;
 
 static NSMutableDictionary *root_object_dictionary;
-static Lock *root_object_dictionary_gate;
+static NSLock *root_object_dictionary_gate;
 
 static NSMapTable *receive_port_2_ancestor;
 
 static NSMapTable *all_connections_local_targets = NULL;
+static NSMapTable *all_connections_local_cached = NULL;
 
 /* rmc handling */
 static NSMutableArray *received_request_rmc_queue;
-static Lock *received_request_rmc_queue_gate;
+static NSLock *received_request_rmc_queue_gate;
 static NSMutableArray *received_reply_rmc_queue;
-static Lock *received_reply_rmc_queue_gate;
+static NSLock *received_reply_rmc_queue_gate;
 
 static int messages_received_count;
 
@@ -202,7 +272,11 @@ static int messages_received_count;
 
 + (NSArray*) allConnections
 {
-  return [connection_array copy];
+  int	count = [connection_array count];
+  id	cons[count];
+
+  [connection_array getObjects: cons];
+  return [NSArray arrayWithObjects: cons count: count];
 }
 
 + (NSConnection*) connectionWithRegisteredName: (NSString*)n
@@ -251,17 +325,20 @@ static int messages_received_count;
 {
   not_owned = [[NSMutableArray alloc] initWithCapacity:8];
   connection_array = [[NSMutableArray alloc] initWithCapacity:8];
-  connection_array_gate = [Lock new];
+  connection_array_gate = [NSLock new];
   /* xxx When NSHashTable's are working, change this. */
   all_connections_local_targets =
     NSCreateMapTable (NSNonOwnedPointerMapKeyCallBacks,
 		      NSObjectMapValueCallBacks, 0);
+  all_connections_local_cached =
+    NSCreateMapTable (NSNonOwnedPointerMapKeyCallBacks,
+		      NSObjectMapValueCallBacks, 0);
   received_request_rmc_queue = [[NSMutableArray alloc] initWithCapacity:32];
-  received_request_rmc_queue_gate = [Lock new];
+  received_request_rmc_queue_gate = [NSLock new];
   received_reply_rmc_queue = [[NSMutableArray alloc] initWithCapacity:32];
-  received_reply_rmc_queue_gate = [Lock new];
+  received_reply_rmc_queue_gate = [NSLock new];
   root_object_dictionary = [[NSMutableDictionary alloc] initWithCapacity:8];
-  root_object_dictionary_gate = [Lock new];
+  root_object_dictionary_gate = [NSLock new];
   receive_port_2_ancestor =
     NSCreateMapTable (NSNonOwnedPointerMapKeyCallBacks,
 		      NSNonOwnedPointerMapValueCallBacks, 0);
@@ -284,6 +361,12 @@ static int messages_received_count;
   return newConn;
 }
 
++ (id)currentConversation
+{
+  [self notImplemented: _cmd];
+  return self;
+}
+
 + (NSDistantObject*) rootProxyForConnectionWithRegisteredName: (NSString*)n
 						         host: (NSString*)h
 {
@@ -294,6 +377,25 @@ static int messages_received_count;
     return [self rootProxyAtPort: [p autorelease]];
 }
 
++ (void) _timeout: (NSTimer*)t
+{
+    NSArray	*cached_locals;
+    int	i;
+
+    cached_locals = NSAllMapTableValues(all_connections_local_cached);
+    for (i = [cached_locals count]; i > 0; i--) {
+	CachedLocalObject *item = [cached_locals objectAtIndex: i-1];
+
+	if ([item countdown] == NO) {
+	    NSMapRemove(all_connections_local_cached, [item obj]);
+	}
+    }
+    if ([cached_locals count] == 0) {
+	[t invalidate];
+	timer = nil;
+    }
+}
+
 - (void) addRequestMode: (NSString*)mode
 {
     if (![request_modes containsObject:mode]) {
@@ -306,7 +408,7 @@ static int messages_received_count;
 - (void) dealloc
 {
   if (debug_connection)
-    printf("deallocating 0x%x\n", (unsigned)self);
+    NSLog(@"deallocating 0x%x\n", (unsigned)self);
   [self invalidate];
 
   /* Remove rootObject from root_object_dictionary
@@ -347,6 +449,18 @@ static int messages_received_count;
     return independant_queueing;
 }
 
+- (void) enableMultipleThreads
+{
+  [self notImplemented: _cmd];
+}
+
+- (BOOL) multipleThreadsEnabled
+{
+  [self notImplemented: _cmd];
+  return NO;
+}
+
+
 - (id) init
 {
   id newPort = [[default_receive_port_class newForReceiving] autorelease];
@@ -424,10 +538,8 @@ static int messages_received_count;
 #endif
 
       if (debug_connection)
-	fprintf(stderr, "Invalidating connection 0x%x\n\t%s\n\t%s\n",
-		(unsigned)self,
-		[[receive_port description] cStringNoCopy],
-		[[send_port description] cStringNoCopy]);
+	NSLog(@"Invalidating connection 0x%x\n\t%@\n\t%@\n", (unsigned)self,
+		[receive_port description], [send_port description]);
 
       [NotificationDispatcher
 	postNotificationName: NSConnectionDidDieNotification
@@ -456,12 +568,21 @@ static int messages_received_count;
 	[super retain];
 	[connection_array_gate lock];
 	[connection_array removeObject: self];
+	[timer invalidate];
+	timer = nil;
+	NSResetMapTable(all_connections_local_cached);
 	[connection_array_gate unlock];
 	[super release];
     }
     [super release];
 }
 
+- (NSArray *) remoteObjects
+{
+  [self notImplemented: _cmd];
+  return nil;
+}
+
 - (void) removeRequestMode: (NSString*)mode
 {
     if ([request_modes containsObject:mode]) {
@@ -470,6 +591,11 @@ static int messages_received_count;
     }
 }
 
+- (void) removeRunLoop: (NSRunLoop *)runloop
+{
+  [self notImplemented: _cmd];
+}
+
 - (NSTimeInterval) replyTimeout
 {
   return reply_timeout;
@@ -705,15 +831,25 @@ static int messages_received_count;
    with that name. */
 
 + (NSConnection*) newRegisteringAtName: (NSString*)n withRootObject: anObj
+{
+  return [self newRegisteringAtName: n
+			     atPort: 0
+		     withRootObject: anObj];
+}
+
++ (NSConnection*) newRegisteringAtName: (NSString*)n
+				atPort: (int)p
+			withRootObject: anObj
 {
   id newPort;
   id newConn;
 
-  newPort = [default_receive_port_class newForReceivingFromRegisteredName: n];
-  newConn = [self newForInPort:[newPort autorelease]
-		       outPort:nil
-	    ancestorConnection:nil];
-  [self setRootObject:anObj forInPort:newPort];
+  newPort = [default_receive_port_class newForReceivingFromRegisteredName: n
+								 fromPort: p];
+  newConn = [self newForInPort: [newPort autorelease]
+		       outPort: nil
+	    ancestorConnection: nil];
+  [self setRootObject: anObj forInPort: newPort];
   return newConn;
 }
 
@@ -773,10 +909,8 @@ static int messages_received_count;
 
   newConn = [[NSConnection alloc] _superInit];
   if (debug_connection)
-    fprintf(stderr, "Created new connection 0x%x\n\t%s\n\t%s\n",
-	    (unsigned)newConn,
-	    [[ip description] cStringNoCopy],
-	    [[op description] cStringNoCopy]);
+    NSLog(@"Created new connection 0x%x\n\t%@\n\t%@\n",
+	    (unsigned)newConn, [ip description], [op description]);
   newConn->is_valid = 1;
   newConn->receive_port = ip;
   [ip retain];
@@ -973,6 +1107,10 @@ static int messages_received_count;
   return self;
 }
 
++ setDebug: (int)val
+{
+    debug_connection = val;
+}
 
 /* Creating new rmc's for encoding requests and replies */
 
@@ -1018,6 +1156,10 @@ static int messages_received_count;
 	case _C_ID:
 	  if (flags & _F_BYCOPY)
 	    [op encodeBycopyObject: *(id*)datum withName: ENCODED_ARGNAME];
+#ifdef	_F_BYREF
+	  else if (flags & _F_BYREF)
+	    [op encodeByrefObject: *(id*)datum withName: ENCODED_ARGNAME];
+#endif
 	  else
 	    [op encodeObject: *(id*)datum withName: ENCODED_ARGNAME];
 	  break;
@@ -1035,8 +1177,6 @@ static int messages_received_count;
 
     NSParameterAssert (is_valid);
     [[self retain] autorelease];
-    op = [self newSendingRequestRmc];
-    seq_num = [op sequenceNumber];
 
     /* get the method types from the selector */
 #if NeXT_runtime
@@ -1047,9 +1187,18 @@ static int messages_received_count;
 #else
     type = sel_get_type(sel);
 #endif
+    if (type == 0 || *type == '\0') {
+	type = [[object methodSignatureForSelector: sel] methodType];
+	if (type) {
+	    sel_register_typed_name(sel_get_name(sel), type);
+	}
+    }
     NSParameterAssert(type);
     NSParameterAssert(*type);
 
+    op = [self newSendingRequestRmc];
+    seq_num = [op sequenceNumber];
+
     /* Send the types that we're using, so that the performer knows
        exactly what qualifiers we're using.
        If all selectors included qualifiers, and if I could make
@@ -1066,6 +1215,8 @@ static int messages_received_count;
     out_parameters = mframe_dissect_call (argframe, type, encoder);
     /* Send the rmc */
     [op dismiss];
+    if (debug_connection > 1)
+      NSLog(@"Sent message to 0x%x\n", (unsigned)self);
     req_out_count++;	/* Sent a request.	*/
 
     /* Get the reply rmc, and decode it. */
@@ -1118,7 +1269,7 @@ static int messages_received_count;
 	     so we don't have a memory leak.  Request here that it be
 	     autorelease'ed. Also autorelease created objects. */
 	  if (*type == _C_CHARPTR)
-	    [MallocAddress autoreleaseMallocAddress: *(char**)datum];
+	    [NSData dataWithBytesNoCopy: *(void**)datum length: 1];
           else if (*type == _C_ID)
             [*(id*)datum autorelease];
 	}
@@ -1130,7 +1281,7 @@ static int messages_received_count;
 	 this is done is in this function DECODER().  IP will be nil
 	 if mframe_build_return() never called DECODER(), i.e. when
 	 we are just returning (void).*/
-      assert (ip == (id)-1 || ip == nil);
+      NSAssert(ip == (id)-1 || ip == nil, NSInternalInconsistencyException);
       rep_in_count++;	/* received a reply */
       return retframe;
     }
@@ -1162,7 +1313,7 @@ static int messages_received_count;
 	 so we don't have a memory leak.  Request here that it be
 	 autorelease'ed. Also autorelease created objects. */
       if (*type == _C_CHARPTR)
-	[MallocAddress autoreleaseMallocAddress: *(char**)datum];
+	[NSData dataWithBytesNoCopy: *(void**)datum length: 1];
       else if (*type == _C_ID)
         [*(id*)datum autorelease];
     }
@@ -1189,6 +1340,10 @@ static int messages_received_count;
 	case _C_ID:
 	  if (flags & _F_BYCOPY)
 	    [op encodeBycopyObject:*(id*)datum withName:ENCODED_RETNAME];
+#ifdef	_F_BYREF
+	  else if (flags & _F_BYREF)
+	    [op encodeByrefObject: *(id*)datum withName: ENCODED_ARGNAME];
+#endif
 	  else
 	    [op encodeObject:*(id*)datum withName:ENCODED_RETNAME];
 	  break;
@@ -1215,6 +1370,8 @@ static int messages_received_count;
 	    at:&forward_type
 	    withName:NULL];
 
+      if (debug_connection > 1)
+        NSLog(@"Handling message from 0x%x\n", (unsigned)self);
       req_in_count++;	/* Handling an incoming request. */
       mframe_do_call (forward_type, decoder, encoder);
       [op dismiss];
@@ -1260,12 +1417,12 @@ static int messages_received_count;
   NSParameterAssert([rmc connection] == self);
   [op encodeObject: rootObject withName: @"root object"];
   [op dismiss];
+  [rmc dismiss];
 }
 
 - (void) _service_release: rmc forConnection: receiving_connection
 {
     unsigned int	count;
-    unsigned int	target;
     unsigned int	pos;
 
     NSParameterAssert (is_valid);
@@ -1280,11 +1437,23 @@ static int messages_received_count;
 		   withName: NULL];
 
     for (pos = 0; pos < count; pos++) {
+	unsigned int	target;
+	char		vended;
+	NSDistantObject	*prox;
+
 	[rmc decodeValueOfCType: @encode(typeof(target))
 			     at: &target
 		       withName: NULL];
 
-	if ([self includesLocalObject:(void*)target]) {
+	[rmc decodeValueOfCType: @encode(typeof(char))
+			     at: &vended
+		       withName: NULL];
+
+	prox = [self includesLocalObject:(void*)target];
+	if (prox != nil) {
+	    if (vended) {
+	        [prox setVended];
+	    }
 	    [self removeLocalObject: (id)target];
 	}
     }
@@ -1307,8 +1476,8 @@ static int messages_received_count;
 			 at: &target
 		   withName: NULL];
 
-    if ([self includesLocalObject:(void*)target] == NO) {
-	if ([[self class] includesLocalObject:(void*)target] == YES) {
+    if ([self includesLocalObject:(void*)target] == nil) {
+	if ([[self class] includesLocalObject:(void*)target] != nil) {
 	    [NSDistantObject proxyWithLocal: (id)target connection: self];
 	}
     }
@@ -1403,6 +1572,7 @@ static int messages_received_count;
       at:&type
       withName:@"Requested Method Type for Target"];
   [op dismiss];
+  [rmc dismiss];
 }
 
 
@@ -1428,13 +1598,11 @@ static int messages_received_count;
       /* It won't take much time to handle this, so go ahead and service
 	 it, even if we are waiting for a reply. */
       [conn _service_rootObject: rmc];
-      [rmc dismiss];
       break;
     case METHODTYPE_REQUEST:
       /* It won't take much time to handle this, so go ahead and service
 	 it, even if we are waiting for a reply. */
       [conn _service_typeForSelector: rmc];
-      [rmc dismiss];
       break;
     case METHOD_REQUEST:
       /* We just got a new request; we need to decide whether to queue
@@ -1530,7 +1698,7 @@ static int messages_received_count;
 	  && [a_rmc sequenceNumber] == sn)
         {
 	  if (debug_connection)
-	    printf("Getting received reply from queue\n");
+	    NSLog(@"Getting received reply from queue\n");
           [received_reply_rmc_queue removeObjectAtIndex: i];
           the_rmc = a_rmc;
           break;
@@ -1617,6 +1785,9 @@ static int messages_received_count;
 	NSMapInsert(all_connections_local_targets, (void*)local, counter);
 	[counter release];
     }
+    if (debug_connection > 2)
+      NSLog(@"add local object (0x%x) to connection (0x%x) (ref %d)\n",
+		(unsigned)local, (unsigned) self, [counter value]);
     [proxiesHashGate unlock];
 }
 
@@ -1650,9 +1821,9 @@ static int messages_received_count;
 - (void) removeLocalObject: anObj
 {
     id	counter;
+    unsigned val = 0;
 
     [proxiesHashGate lock];
-    NSMapRemove (local_targets, (void*)anObj);
 
     /*
      *	If all references to a local proxy have gone - remove the
@@ -1661,15 +1832,40 @@ static int messages_received_count;
     counter = NSMapGet(all_connections_local_targets, (void*)anObj);
     if (counter) {
 	[counter decrement];
-	if ([counter value] == 0) {
+	if ((val = [counter value]) == 0) {
+	    NSDistantObject	*prox = NSMapGet(local_targets, (void*)anObj);
+
 	    NSMapRemove(all_connections_local_targets, (void*)anObj);
+	    /*
+	     *	If this proxy has been vended onwards by another process, we
+	     *	need to keep a reference to the local object around for a
+	     *	while in case that other process needs it.
+	     */
+	    if ([prox isVended]) {
+		id	item;
+		if (timer == nil) {
+		    timer = [NSTimer scheduledTimerWithTimeInterval: 1.0
+					   target: [NSConnection class]
+					 selector: @selector(_timeout:)
+					 userInfo: nil
+					  repeats: YES];
+		}
+		item = [CachedLocalObject itemWithObject: anObj time: 30];
+		NSMapInsert(all_connections_local_cached, anObj, item);
+	    }
 	}
     }
 
+    NSMapRemove (local_targets, (void*)anObj);
+
+    if (debug_connection > 2)
+	NSLog(@"remove local object (0x%x) to connection (0x%x) (ref %d)\n",
+		(unsigned)anObj, (unsigned) self, val);
+
     [proxiesHashGate unlock];
 }
 
-- (void) _release_targets: (unsigned int*)list count:(unsigned int)number
+- (void) _release_targets: (NSDistantObject**)list count:(unsigned int)number
 {
     NS_DURING
     {
@@ -1692,11 +1888,15 @@ static int messages_received_count;
 			  withName: NULL];
 
 	    for (i = 0; i < number; i++) {
-		unsigned int	target = list[i];
+		unsigned	target = (unsigned)[list[i] targetForProxy];
+		char		vended = [list[i] isVended];
 
 		[op encodeValueOfCType: @encode(typeof(target))
 				    at: &target
 			      withName: NULL];
+		[op encodeValueOfCType: @encode(char)
+				    at: &vended
+			      withName: NULL];
 	    }
 
 	    [op dismiss];
@@ -1705,8 +1905,7 @@ static int messages_received_count;
     NS_HANDLER
     {
       if (debug_connection)
-        fprintf (stderr, "failed to release targets - %s\n",
-	     [[localException name] cStringNoCopy]);
+        NSLog(@"failed to release targets - %@\n", [localException name]);
     }
     NS_ENDHANDLER
 }
@@ -1722,10 +1921,11 @@ static int messages_received_count;
 	if (receive_port && is_valid) {
 	    id		op;
 	    unsigned int 	i;
+	    int seq_num = [self _newMsgNumber];
 
 	    op = [[self encodingClass]
 		    newForWritingWithConnection: self
-				 sequenceNumber: [self _newMsgNumber]
+				 sequenceNumber: seq_num
 				     identifier: PROXY_RETAIN];
 
 	    [op encodeValueOfCType: @encode(typeof(target))
@@ -1738,8 +1938,7 @@ static int messages_received_count;
     NS_HANDLER
     {
       if (debug_connection)
-        fprintf (stderr, "failed to retain target - %s\n",
-	     [[localException name] cStringNoCopy]);
+        NSLog(@"failed to retain target - %@\n", [localException name]);
     }
     NS_ENDHANDLER
 }
@@ -1758,10 +1957,10 @@ static int messages_received_count;
      *	Tell the remote application that we have removed our proxy and
      *	it can release it's local object.
      */
-    [self _release_targets:&target count:1];
+    [self _release_targets:&aProxy count:1];
 }
 
-- (id <Collecting>) localObjects
+- (NSArray *) localObjects
 {
   id c;
 
@@ -1772,7 +1971,7 @@ static int messages_received_count;
   return c;
 }
 
-- (id <Collecting>) proxies
+- (NSArray *) proxies
 {
   id c;
 
@@ -1810,24 +2009,24 @@ static int messages_received_count;
   [proxiesHashGate unlock];
 }
 
-- (BOOL) includesProxyForTarget: (unsigned)target
+- (id) includesProxyForTarget: (void*)target
 {
-  BOOL ret;
+  NSDistantObject	*ret;
 
   /* Don't assert (is_valid); */
   [proxiesHashGate lock];
-  ret = NSMapGet (remote_proxies, (void*)target) ? YES : NO;
+  ret = NSMapGet (remote_proxies, (void*)target);
   [proxiesHashGate unlock];
   return ret;
 }
 
-- (BOOL) includesLocalObject: anObj
+- (id) includesLocalObject: anObj
 {
-  BOOL ret;
+  NSDistantObject* ret;
 
   /* Don't assert (is_valid); */
   [proxiesHashGate lock];
-  ret = NSMapGet (local_targets, (void*)anObj) ? YES : NO;
+  ret = NSMapGet(local_targets, (void*)anObj);
   [proxiesHashGate unlock];
   return ret;
 }
@@ -1838,14 +2037,17 @@ static int messages_received_count;
    for the Proxy to check the Proxy's connection only (using
    -includesLocalObject), because the proxy may have come from a
    triangle connection. */
-+ (BOOL) includesLocalObject: anObj
++ (id) includesLocalObject: anObj
 {
-  BOOL ret;
+  id ret;
 
   /* Don't assert (is_valid); */
   NSParameterAssert (all_connections_local_targets);
   [proxiesHashGate lock];
-  ret = NSMapGet (all_connections_local_targets, (void*)anObj) ? YES : NO;
+  ret = NSMapGet (all_connections_local_targets, (void*)anObj);
+  if (ret == nil) {
+    ret = NSMapGet (all_connections_local_cached, (void*)anObj);
+  }
   [proxiesHashGate unlock];
   return ret;
 }
@@ -1863,13 +2065,13 @@ static int messages_received_count;
       if (anObj)
 	{
 	  [root_object_dictionary_gate lock];
-	  [root_object_dictionary putObject: anObj atKey: aPort];
+	  [root_object_dictionary setObject: anObj forKey: aPort];
 	  [root_object_dictionary_gate unlock];
 	}
       else /* anObj == nil && oldRootObject != nil */
 	{
 	  [root_object_dictionary_gate lock];
-	  [root_object_dictionary removeObjectAtKey: aPort];
+	  [root_object_dictionary removeObjectForKey: aPort];
 	  [root_object_dictionary_gate unlock];
 	}
     }
@@ -1880,7 +2082,7 @@ static int messages_received_count;
   id ro;
 
   [root_object_dictionary_gate lock];
-  ro = [root_object_dictionary objectAtKey:aPort];
+  ro = [root_object_dictionary objectForKey:aPort];
   [root_object_dictionary_gate unlock];
   return ro;
 }
@@ -1992,9 +2194,9 @@ static int messages_received_count;
 	id port = [notification object];
 
 	if (debug_connection)
-	    fprintf (stderr, "Received port invalidation notification for "
-		"connection 0x%x\n\t%s\n", (unsigned)self,
-		[[port description] cStringNoCopy]);
+	    NSLog(@"Received port invalidation notification for "
+		@"connection 0x%x\n\t%@\n", (unsigned)self,
+		[port description]);
 
 	/* We shouldn't be getting any port invalidation notifications,
 	    except from our own ports; this is how we registered ourselves
