@@ -22,51 +22,154 @@
    */ 
 
 #include <config.h>
-#include <gnustep/base/preface.h>
+#include <base/preface.h>
 #include <stdarg.h>
 #include <Foundation/NSObject.h>
 #include <objc/Protocol.h>
 #include <objc/objc-api.h>
 #include <Foundation/NSMethodSignature.h>
+#include <base/Invocation.h>
 #include <Foundation/NSInvocation.h>
 #include <Foundation/NSAutoreleasePool.h>
 #include <Foundation/NSString.h>
-#include <gnustep/base/o_map.h>
+#include <base/o_map.h>
+#include <Foundation/NSArray.h>
 #include <Foundation/NSException.h>
+#include <Foundation/NSPortCoder.h>
+#include <Foundation/NSDistantObject.h>
 #include <Foundation/NSZone.h>
 #include <limits.h>
 
-
-/* Reference count management */
+#include <base/fast.x>
 
-/* Handles multi-threaded access.
-   ?? Doesn't handle exceptions. */
+fastCls	_fastCls;	/* Structure to cache classes.	*/
+fastImp	_fastImp;	/* Structure to cache methods.	*/
+
+@class	_FastMallocBuffer;
+static Class	fastMallocClass;
+static unsigned	fastMallocOffset;
+
+@class	NSDataMalloc;
+@class	NSMutableDataMalloc;
+
+void	_fastBuildCache()
+{
+  /*
+   *	Cache some classes for quick access later.
+   */
+
+  _fastCls._NSString = [NSString class];
+  _fastCls._NSGString = [NSGString class];
+  _fastCls._NSGMutableString = [NSGMutableString class];
+  _fastCls._NSGCString = [NSGCString class];
+  _fastCls._NSGMutableCString = [NSGMutableCString class];
+  _fastCls._NXConstantString = [NXConstantString class];
+  _fastCls._NSDataMalloc = [NSDataMalloc class];
+  _fastCls._NSMutableDataMalloc = [NSMutableDataMalloc class];
+
+  /*
+   *	Cache some method implementations for quick access later.
+   */
+
+  _fastImp._NSString_hash = (unsigned (*)())[_fastCls._NSString
+	    instanceMethodForSelector: @selector(hash)];
+  _fastImp._NSString_isEqualToString_ = (BOOL (*)())[_fastCls._NSString
+	    instanceMethodForSelector: @selector(isEqualToString:)];
+  _fastImp._NSGString_isEqual_ = (BOOL (*)())[_fastCls._NSGString
+	    instanceMethodForSelector: @selector(isEqual:)];
+  _fastImp._NSGCString_isEqual_ = (BOOL (*)())[_fastCls._NSGCString
+	    instanceMethodForSelector: @selector(isEqual:)];
+}
+
+
+/*
+ *	Reference count and memory management
+ *
+ *	If REFCNT_LOCAL is defined, reference counts for object are stored
+ *	with the object, otherwise they are stored in a global map table
+ *	that has to be protected by mutexes in a multithreraded environment.
+ *	You therefore want REFCNT_LOCAL defined for best performance.
+ *
+ *	If CACHE_ZONE is defined, the zone in which an object has been
+ *	allocated is stored with the object - this makes lookup of the
+ *	correct zone to free memory very fast.
+ */
+
+#define	REFCNT_LOCAL	1
+#define	CACHE_ZONE	1
+
+#if	defined(REFCNT_LOCAL) || defined(CACHE_ZONE)
+
+/*
+ *	Define a structure to hold information that is held locally
+ *	(before the start) in each object.
+ */ 
+
+typedef struct obj_layout_unpadded {
+#if	defined(REFCNT_LOCAL)
+    unsigned	retained;
+#endif
+#if	defined(CACHE_ZONE)
+    NSZone	*zone;
+#endif
+} unp;
+#define	UNP sizeof(unp)
+
+/*
+ *	Now do the REAL version - using the other version to determine
+ *	what padding (if any) is required to get the alignment of the
+ *	structure correct.
+ */
+#define	ALIGN __alignof__(double)
+struct obj_layout {
+#if	defined(REFCNT_LOCAL)
+    unsigned	retained;
+#endif
+#if	defined(CACHE_ZONE)
+    NSZone	*zone;
+#endif
+    char	padding[ALIGN - ((UNP % ALIGN) ? (UNP % ALIGN) : ALIGN)];
+};
+typedef	struct obj_layout *obj;
+
+#endif	/* defined(REFCNT_LOCAL) || defined(CACHE_ZONE) */
+
+
+/*
+ *	Now do conditional compilation of reference count functions
+ *	depending on whether we are using local or global counting.
+ */
+#if	defined(REFCNT_LOCAL)
+void
+NSIncrementExtraRefCount(id anObject)
+{
+  ((obj)anObject)[-1].retained++;
+}
+
+#define	NSIncrementExtraRefCount(X) \
+	((obj)(X))[-1].retained++
+
+BOOL
+NSDecrementExtraRefCountWasZero(id anObject)
+{
+  if (((obj)anObject)[-1].retained-- == 0)
+    return YES;
+  else
+    return NO;
+}
+
+#define	NSDecrementExtraRefCountWasZero(X) \
+	(((obj)(X))[-1].retained-- == 0 ? YES : NO)
+
+#define	extraRefCount(X)	(((obj)(X))[-1].retained)
+
+#else
 
 /* The maptable of retain counts on objects */
 static o_map_t *retain_counts = NULL;
 /* The mutex lock to protect multi-threaded use of `retain_counts' */
 static objc_mutex_t retain_counts_gate = NULL;
 
-/* The mutex lock to protect RETAIN_COUNTS. */
-static objc_mutex_t retain_counts_gate;
-
-/* The Class responsible for handling autorelease's.  This does not
-   need mutex protection, since it is simply a pointer that gets read
-   and set. */
-static id autorelease_class = nil;
-
-/* When this is `YES', every call to release/autorelease, checks to
-   make sure isn't being set up to release itself too many times.
-   This does not need mutex protection. */
-static BOOL double_release_check_enabled = NO;
-
-BOOL
-NSShouldRetainWithZone (NSObject *anObject, NSZone *requestedZone)
-{
-  return (!requestedZone || requestedZone == NSDefaultMallocZone()
-	  || [anObject zone] == requestedZone);
-}
-
 void
 NSIncrementExtraRefCount (id anObject)
 {
@@ -96,7 +199,7 @@ NSDecrementExtraRefCountWasZero (id anObject)
       objc_mutex_unlock (retain_counts_gate);
       return YES;
     }
-  assert ((int)(node->value) > 0);
+  NSAssert((int)(node->value) > 0, NSInternalInconsistencyException);
   if (!--((int)(node->value)))
     o_map_remove_node (node);
   objc_mutex_unlock (retain_counts_gate);
@@ -116,6 +219,149 @@ extraRefCount (id anObject)
   return ret;	/* ExtraRefCount + 1	*/
 }
 
+#endif	/* defined(REFCNT_LOCAL) */
+
+
+/*
+ *	Now do conditional compilation of memory allocation functions
+ *	depending on what information (if any) we are storing before
+ *	the start of each object.
+ */
+#if	defined(REFCNT_LOCAL) || defined(CACHE_ZONE)
+
+#if defined(CACHE_ZONE)
+
+inline NSZone *
+fastZone(NSObject *object)
+{
+    if (fastClass(object) == _fastCls._NXConstantString)
+	return NSDefaultMallocZone();
+    return ((obj)object)[-1].zone;
+}
+
+#else	/* defined(CACHE_ZONE)	*/
+
+inline NSZone *
+fastZone(NSObject *object)
+{
+    if (fastClass(object) == _fastCls._NXConstantString)
+	return NSDefaultMallocZone();
+    return NSZoneFromPointer(&((obj)object)[-1]);
+}
+
+#endif	/* defined(CACHE_ZONE)	*/
+
+inline NSObject *
+NSAllocateObject (Class aClass, unsigned extraBytes, NSZone *zone)
+{
+  id new = nil;
+  int size = aClass->instance_size + extraBytes + sizeof(struct obj_layout);
+  if (CLS_ISCLASS (aClass))
+    {
+      if (zone == 0)
+	zone = NSDefaultMallocZone();
+      new = NSZoneMalloc(zone, size);
+    }
+  if (new != nil)
+    {
+      memset (new, 0, size);
+#if	defined(CACHE_ZONE)
+      ((obj)new)->zone = zone;
+#endif
+      new = (id)&((obj)new)[1];
+      new->class_pointer = aClass;
+#ifndef	NDEBUG
+      GSDebugAllocationAdd(aClass);
+#endif
+    }
+  return new;
+}
+
+inline void
+NSDeallocateObject(NSObject *anObject)
+{
+  if ((anObject!=nil) && CLS_ISCLASS(((id)anObject)->class_pointer))
+    {
+      obj	o = &((obj)anObject)[-1];
+      NSZone	*z = fastZone(anObject);
+
+#ifndef	NDEBUG
+      GSDebugAllocationRemove(((id)anObject)->class_pointer);
+#endif
+      ((id)anObject)->class_pointer = (void*) 0xdeadface;
+      NSZoneFree(z, o);
+    }
+  return;
+}
+
+#else
+
+inline NSZone *
+fastZone(NSObject *object)
+{
+    if (fastClass(object) == _fastCls._NXConstantString)
+	return NSDefaultMallocZone();
+    return NSZoneFromPointer(object);
+}
+
+inline NSObject *
+NSAllocateObject (Class aClass, unsigned extraBytes, NSZone *zone)
+{
+  id new = nil;
+  int size = aClass->instance_size + extraBytes;
+  if (CLS_ISCLASS (aClass))
+    new = NSZoneMalloc (zone, size);
+  if (new != nil)
+    {
+      memset (new, 0, size);
+      new->class_pointer = aClass;
+#ifndef	NDEBUG
+      GSDebugAllocationAdd(aClass);
+#endif
+    }
+  return new;
+}
+
+inline void
+NSDeallocateObject(NSObject *anObject)
+{
+  if ((anObject!=nil) && CLS_ISCLASS(((id)anObject)->class_pointer))
+    {
+      NSZone	*z = [anObject zone];
+
+#ifndef	NDEBUG
+      GSDebugAllocationRemove(((id)anObject)->class_pointer);
+#endif
+      ((id)anObject)->class_pointer = (void*) 0xdeadface;
+      NSZoneFree(z, anObject);
+    }
+  return;
+}
+
+#endif	/* defined(REFCNT_LOCAL) || defined(CACHE_ZONE) */
+
+BOOL
+NSShouldRetainWithZone (NSObject *anObject, NSZone *requestedZone)
+{
+    return (!requestedZone || requestedZone == NSDefaultMallocZone()
+	  || fastZone(anObject) == requestedZone);
+}
+
+
+
+
+/* The Class responsible for handling autorelease's.  This does not
+   need mutex protection, since it is simply a pointer that gets read
+   and set. */
+static id autorelease_class = nil;
+static SEL autorelease_sel = @selector(addObject:);
+static IMP autorelease_imp = 0;
+
+/* When this is `YES', every call to release/autorelease, checks to
+   make sure isn't being set up to release itself too many times.
+   This does not need mutex protection. */
+static BOOL double_release_check_enabled = NO;
+
 
 @implementation NSObject
 
@@ -125,11 +371,16 @@ extraRefCount (id anObject)
     {
       // Create the global lock
       gnustep_global_lock = [[NSRecursiveLock alloc] init];
-
+#if !defined(REFCNT_LOCAL)
       retain_counts = o_map_with_callbacks (o_callbacks_for_non_owned_void_p,
 					    o_callbacks_for_int);
       retain_counts_gate = objc_mutex_allocate ();
+#endif
       autorelease_class = [NSAutoreleasePool class];
+      autorelease_imp = [autorelease_class methodForSelector: autorelease_sel];
+      fastMallocClass = [_FastMallocBuffer class];
+      fastMallocOffset = fastMallocClass->instance_size % ALIGN;
+      _fastBuildCache();
     }
   return;
 }
@@ -144,19 +395,19 @@ extraRefCount (id anObject)
   return NSAllocateObject (self, 0, z);
 }
 
++ (id) copyWithZone: (NSZone*)z
+{
+  return self;
+}
+
 + (id) new
 {
   return [[self alloc] init];
 }
 
-- copyWithZone:(NSZone *)zone
-{
-  return NSCopyObject (self, 0, zone);
-}
-
 - (id) copy
 {
-  return [self copyWithZone: NSDefaultMallocZone()];
+  return [(id)self copyWithZone: NULL];
 }
 
 - (void) dealloc
@@ -176,14 +427,9 @@ extraRefCount (id anObject)
   return self;
 }
 
-- mutableCopyWithZone:(NSZone *)zone
-{
-  return [self copyWithZone:zone];
-}
-
 - (id) mutableCopy
 {
-  return [self mutableCopyWithZone: NSDefaultMallocZone()];
+  return [(id)self mutableCopyWithZone: NULL];
 }
 
 + (Class) superclass
@@ -198,7 +444,11 @@ extraRefCount (id anObject)
 
 + (BOOL) instancesRespondToSelector: (SEL)aSelector
 {
+#if 0
   return (class_get_instance_method(self, aSelector) != METHOD_NULL);
+#else
+  return __objc_responds_to((id)&self, aSelector);
+#endif
 }
 
 + (BOOL) conformsToProtocol: (Protocol*)aProtocol
@@ -230,40 +480,77 @@ extraRefCount (id anObject)
 
 + (IMP) instanceMethodForSelector: (SEL)aSelector
 {
-  return method_get_imp(class_get_instance_method(self, aSelector));
-}
-  
-- (IMP) methodForSelector: (SEL)aSelector
-{
-  return (method_get_imp(object_is_instance(self)
-                         ?class_get_instance_method(self->isa, aSelector)
-                         :class_get_class_method(self->isa, aSelector)));
+  /*
+   *	Since 'self' is an class, get_imp() will get the instance method.
+   */
+  return get_imp((Class)self, aSelector);
 }
 
+- (IMP) methodForSelector: (SEL)aSelector
+{
+  /*
+   *	If 'self' is an instance, fastClass() will get the class,
+   *	and get_imp() will get the instance method.
+   *	If 'self' is a class, fastClass() will get the meta-class,
+   *	and get_imp() will get the class method.
+   */
+  return get_imp(fastClass(self), aSelector);
+}
+
++ (NSMethodSignature*) instanceMethodSignatureForSelector: (SEL)aSelector
+{
+    struct objc_method* mth = class_get_instance_method(self, aSelector);
+    return mth ? [NSMethodSignature signatureWithObjCTypes:mth->method_types]
+		: nil;
+}
+  
 - (NSMethodSignature*) methodSignatureForSelector: (SEL)aSelector
 {
     struct objc_method* mth =
 	    (object_is_instance(self) ?
 		  class_get_instance_method(self->isa, aSelector)
 		: class_get_class_method(self->isa, aSelector));
-
     return mth ? [NSMethodSignature signatureWithObjCTypes:mth->method_types]
 		: nil;
 }
 
 - (NSString*) description
 {
-  return [NSString stringWithCString: object_get_class_name(self)];
+  return [NSString stringWithFormat: @"<%s %lx>",
+	object_get_class_name(self), (unsigned long)self];
 }
 
 + (NSString*) description
 {
-  return [NSString stringWithCString: class_get_class_name(self)];
+  return [NSString stringWithFormat: @"<%s>", object_get_class_name(self)];
+}
+
+- (void) descriptionTo: (id<GNUDescriptionDestination>)output
+{
+  [output appendString: [self description]];
+}
+
+- (void) descriptionWithLocale: (NSDictionary*)aLocale
+			    to: (id<GNUDescriptionDestination>)output
+{
+  [output appendString: [(id)self descriptionWithLocale: aLocale]];
+}
+
+- (void) descriptionWithLocale: (NSDictionary*)aLocale
+			indent: (unsigned)level
+			    to: (id<GNUDescriptionDestination>)output
+{
+  [output appendString: [(id)self descriptionWithLocale: aLocale indent: level]];
 }
 
 + (void) poseAsClass: (Class)aClassObject
 {
   class_pose_as(self, aClassObject);
+  /*
+   *	We may have replaced a class in the cache, or may have replaced one
+   *	which had cached methods, so we must rebuild the cache.
+   */
+  _fastBuildCache();
 }
 
 - (void) doesNotRecognizeSelector: (SEL)aSelector
@@ -275,21 +562,12 @@ extraRefCount (id anObject)
 
 - (retval_t) forward:(SEL)aSel :(arglist_t)argFrame
 {
-#if 1
-  [self doesNotRecognizeSelector:aSel];
-  return NULL;
-#else
-  void *retFrame;
-  NSMethodSignature *ms = [self methodSignatureForSelector:aSel];
-  NSInvocation *inv = [NSInvocation invocationWithMethodSignature:ms
-				    frame:argFrame];
-  /* is this right? */
-  retFrame = (void*) alloca([ms methodReturnLength]);
+  NSInvocation *inv;
+
+  inv = [[[NSInvocation alloc] initWithArgframe: argFrame
+				       selector: aSel] autorelease];
   [self forwardInvocation:inv];
-  [inv getReturnValue:retFrame];
-  /* where do ms and inv get free'd? */
-  return retFrame;
-#endif
+  return [inv returnFrame: argFrame];
 }
 
 - (void) forwardInvocation: (NSInvocation*)anInvocation
@@ -303,16 +581,45 @@ extraRefCount (id anObject)
   return self;
 }
 
+- (Class) classForArchiver
+{
+  return [self classForCoder];
+}
+
 - (Class) classForCoder
 {
   return [self class];
 }
 
+- (Class) classForPortCoder
+{
+    return [self classForCoder];
+}
+
+- (id) replacementObjectForArchiver: (NSArchiver*)anArchiver
+{
+  return [self replacementObjectForCoder: (NSCoder*)anArchiver];
+}
+
 - (id) replacementObjectForCoder: (NSCoder*)anEncoder
 {
   return self;
 }
 
+- (id) replacementObjectForPortCoder: (NSPortCoder*)aCoder
+{
+    if ([aCoder isBycopy]) {
+	return self;
+    }
+    else if ([self isKindOfClass: [NSDistantObject class]]) {
+	return self;
+    }
+    else {
+	return [NSDistantObject proxyWithLocal: self
+				    connection: [aCoder connection]];
+    }
+}
+
 
 /* NSObject protocol */
 
@@ -330,7 +637,7 @@ extraRefCount (id anObject)
 	  @"%d release(s) versus %d retain(s)", release_count, retain_count];
     }
 
-  [autorelease_class addObject:self];
+  (*autorelease_imp)(autorelease_class, autorelease_sel, self);
   return self;
 }
 
@@ -339,6 +646,11 @@ extraRefCount (id anObject)
   return self;
 }
 
++ (Class) class
+{
+  return self;
+}
+
 - (Class) class
 {
   return object_get_class(self);
@@ -354,6 +666,20 @@ extraRefCount (id anObject)
   return (self == anObject);
 }
 
++ (BOOL) isKindOfClass: (Class)aClass
+{
+  Class class;
+
+  for (class = self; 
+       class != Nil;
+       class = class_get_super_class (class))
+    {
+      if (class == aClass)
+	return YES;
+    }
+  return NO;
+}
+
 - (BOOL) isKindOfClass: (Class)aClass
 {
   Class class;
@@ -368,6 +694,11 @@ extraRefCount (id anObject)
   return NO;
 }
 
++ (BOOL) isMemberOfClass: (Class)aClass
+{
+  return self == aClass;
+}
+
 - (BOOL) isMemberOfClass: (Class)aClass
 {
   return self->isa==aClass;
@@ -378,42 +709,68 @@ extraRefCount (id anObject)
   return NO;
 }
 
-- perform: (SEL)aSelector
+- performSelector: (SEL)aSelector
 {
-  IMP msg = objc_msg_lookup(self, aSelector);
+  IMP msg;
+
+  if (aSelector == 0)
+    {
+      [NSException raise: NSInvalidArgumentException
+		  format: @"nul selector passed to %s", sel_get_name(_cmd)];
+      return nil;
+    }
+    
+  msg = get_imp(fastClass(self), aSelector);
   if (!msg)
     {
-      [NSException
-	raise: NSGenericException
-	format: @"invalid selector passed to %s", sel_get_name(_cmd)];
+      [NSException raise: NSGenericException
+		  format: @"invalid selector passed to %s", sel_get_name(_cmd)];
       return nil;
     }
   return (*msg)(self, aSelector);
 }
 
-- perform: (SEL)aSelector withObject: anObject
+- performSelector: (SEL)aSelector withObject: anObject
 {
-  IMP msg = objc_msg_lookup(self, aSelector);
-  if (!msg)
+  IMP msg;
+
+  if (aSelector == 0)
     {
-      [NSException
-	raise: NSGenericException
-	format: @"invalid selector passed to %s", sel_get_name(_cmd)];
+      [NSException raise: NSInvalidArgumentException
+		  format: @"nul selector passed to %s", sel_get_name(_cmd)];
       return nil;
     }
+    
+  msg = get_imp(fastClass(self), aSelector);
+  if (!msg)
+    {
+      [NSException raise: NSGenericException
+		  format: @"invalid selector passed to %s", sel_get_name(_cmd)];
+      return nil;
+    }
+
   return (*msg)(self, aSelector, anObject);
 }
 
-- perform: (SEL)aSelector withObject: object1 withObject: object2
+- performSelector: (SEL)aSelector withObject: object1 withObject: object2
 {
-  IMP msg = objc_msg_lookup(self, aSelector);
-  if (!msg)
+  IMP msg;
+
+  if (aSelector == 0)
     {
-      [NSException
-	raise: NSGenericException
-	format: @"invalid selector passed to %s", sel_get_name(_cmd)];
+      [NSException raise: NSInvalidArgumentException
+		  format: @"nul selector passed to %s", sel_get_name(_cmd)];
       return nil;
     }
+  
+  msg = get_imp(fastClass(self), aSelector);
+  if (!msg)
+    {
+      [NSException raise: NSGenericException
+		  format: @"invalid selector passed to %s", sel_get_name(_cmd)];
+      return nil;
+    }
+
   return (*msg)(self, aSelector, object1, object2);
 }
 
@@ -441,9 +798,14 @@ extraRefCount (id anObject)
 
 - (BOOL) respondsToSelector: (SEL)aSelector
 {
-  return ((object_is_instance(self)
-           ?class_get_instance_method(self->isa, aSelector)
-           :class_get_class_method(self->isa, aSelector))!=METHOD_NULL);
+#if 0
+  if (fastIsInstance(self))
+    return (class_get_instance_method(fastClass(self), aSelector)!=METHOD_NULL);
+  else
+    return (class_get_class_method(fastClass(self), aSelector)!=METHOD_NULL);
+#else
+  return __objc_responds_to(self, aSelector);
+#endif
 }
 
 - retain
@@ -474,7 +836,7 @@ extraRefCount (id anObject)
 
 - (NSZone *)zone
 {
-  return NSZoneFromPointer(self);
+    return fastZone(self);
 }
 
 - (void) encodeWithCoder: (NSCoder*)aCoder
@@ -487,6 +849,21 @@ extraRefCount (id anObject)
   return self;
 }
 
++ (int)version
+{
+  return class_get_version(self);
+}
+
++ setVersion:(int)aVersion
+{
+  if (aVersion < 0)
+    [NSException raise: NSInvalidArgumentException
+	        format: @"%s +setVersion: may not set a negative version",
+			object_get_class_name(self)];
+  class_set_version(self, aVersion);
+  return self;
+}
+
 @end
 
 
@@ -552,7 +929,7 @@ extraRefCount (id anObject)
   return objc_msg_sendv(self, aSel, argFrame);
 }
 
-+ (IMP)instanceMethodFor:(SEL)aSel
++ (IMP) instanceMethodFor:(SEL)aSel
 {
   return [self instanceMethodForSelector:aSel];
 }
@@ -565,7 +942,7 @@ extraRefCount (id anObject)
 		: nil;
 }
 
-- (IMP)methodFor:(SEL)aSel
+- (IMP) methodFor:(SEL)aSel
 {
   return [self methodForSelector:aSel];
 }
@@ -576,17 +953,6 @@ extraRefCount (id anObject)
   return self;
 }
 
-+ (int)version
-{
-  return class_get_version(self);
-}
-
-+ setVersion:(int)aVersion
-{
-  class_set_version(self, aVersion);
-  return self;
-}
-
 - notImplemented:(SEL)aSel
 {
   [NSException
@@ -605,24 +971,26 @@ extraRefCount (id anObject)
 
 - perform: (SEL)sel with: anObject
 {
-  return [self perform:sel withObject:anObject];
+  return [self performSelector:sel withObject:anObject];
 }
 
 - perform: (SEL)sel with: anObject with: anotherObject
 {
-  return [self perform:sel withObject:anObject withObject:anotherObject];
+  return [self performSelector:sel withObject:anObject 
+	       withObject:anotherObject];
 }
 
 @end
 
 
-@implementation NSObject (GNU)
+@implementation NSObject (GNUstep)
 
 /* GNU Object class compatibility */
 
 + (void) setAutoreleaseClass: (Class)aClass
 {
   autorelease_class = aClass;
+  autorelease_imp = [self instanceMethodForSelector: autorelease_sel];
 }
 
 + (Class) autoreleaseClass
@@ -747,49 +1115,27 @@ extraRefCount (id anObject)
 
 @end
 
-
-@implementation NSObject (OPENSTEP)
-
-/* OPENSTEP Object class extensions
-   as distinquished from the OpenStep specification. */
-
-- performSelector: (SEL)aSelector
-{
-  IMP msg = objc_msg_lookup(self, aSelector);
-  if (!msg)
-    {
-      [NSException
-	raise: NSGenericException
-	format: @"invalid selector passed to %s", sel_get_name(_cmd)];
-      return nil;
-    }
-  return (*msg)(self, aSelector);
-}
-
-- performSelector: (SEL)aSelector withObject: anObject
-{
-  IMP msg = objc_msg_lookup(self, aSelector);
-  if (!msg)
-    {
-      [NSException
-	raise: NSGenericException
-	format: @"invalid selector passed to %s", sel_get_name(_cmd)];
-      return nil;
-    }
-  return (*msg)(self, aSelector, anObject);
-}
-
-- performSelector: (SEL)aSelector withObject: object1 withObject: object2
-{
-  IMP msg = objc_msg_lookup(self, aSelector);
-  if (!msg)
-    {
-      [NSException
-	raise: NSGenericException
-	format: @"invalid selector passed to %s", sel_get_name(_cmd)];
-      return nil;
-    }
-  return (*msg)(self, aSelector, object1, object2);
-}
-
+/*
+ *	Stuff for temporary memory management.
+ */
+@interface	_FastMallocBuffer : NSObject
 @end
+
+@implementation	_FastMallocBuffer
+@end
+
+/*
+ *	Function for giving us the fastest possible allocation of memory to
+ *	be used for temporary storage.
+ */
+void *
+_fastMallocBuffer(unsigned size)
+{
+  _FastMallocBuffer	*o;
+
+  o = (_FastMallocBuffer*)NSAllocateObject(fastMallocClass,
+	size + fastMallocOffset, NSDefaultMallocZone());
+  (*autorelease_imp)(autorelease_class, autorelease_sel, o);
+  return ((void*)&o[1])+fastMallocOffset;
+}
+
