@@ -51,9 +51,8 @@
    */
 
 #include <config.h>
-#include <gnustep/base/preface.h>
-#include <gnustep/base/Bag.h>
-#include <gnustep/base/Heap.h>
+#include <base/preface.h>
+#include <base/Heap.h>
 #include <Foundation/NSMapTable.h>
 #include <Foundation/NSDate.h>
 #include <Foundation/NSValue.h>
@@ -63,9 +62,11 @@
 #include <Foundation/NSNotificationQueue.h>
 #include <Foundation/NSRunLoop.h>
 #include <Foundation/NSThread.h>
+#include <Foundation/NSDebug.h>
 
 #include <sys/types.h>
 #ifndef __WIN32__
+#include <time.h>
 #include <sys/time.h>
 #endif /* !__WIN32__ */
 #include <limits.h>
@@ -107,170 +108,121 @@ static int debug_run_loop = 0;
  *	runloop must ask the 'receiver' (or its delegate) to supply a date
  *	using the '[-limitDateForMode:]' message.
  *
+ *	NB.  This class is private to NSRunLoop and must not be subclassed.
  */
  
 @interface RunLoopWatcher: NSObject
 {
-    BOOL		invalidated;
-    void*		data;
-    id			receiver;
-    RunLoopEventType	type;
-    NSDate*		limit;
-    unsigned 		count;
+@public
+  BOOL			invalidated;
+  void			*data;
+  id			receiver;
+  RunLoopEventType	type;
+  NSDate		*limit;
+  unsigned 		count;
 }
-- (void) eventFor: (void*)info mode: (NSString*)mode;
-- (void*) getData;
-- (NSDate*) getLimit;
-- (id) getReceiver;
-- (RunLoopEventType) getType;
-- (BOOL) decrement;
-- (void) increment;
+- (void) eventFor: (void*)info
+	     mode: (NSString*)mode;
 - initWithType: (RunLoopEventType)type
       receiver: (id)anObj
           data: (void*)data;
 - (void) invalidate;
 - (BOOL) isValid;
-- (void) setData:(void*)item;
-- (void) setLimit:(NSDate*)when;
-- (void) setReceiver:anObj;
 @end
 
 @implementation	RunLoopWatcher
 
 - (void) dealloc
 {
-    [self invalidate];
-    [limit release];
-    [receiver release];
-    [super dealloc];
+  [self invalidate];
+  [limit release];
+  [receiver release];
+  [super dealloc];
 }
 
-- (BOOL) decrement
+- (void) eventFor: (void*)info
+	     mode: (NSString*)mode
 {
-    if (count > 0) {
-	count--;
-	if (count > 0) {
-	    return YES;
+  if ([self isValid] == NO)
+    {
+      return;
+    }
+
+  if ([receiver respondsToSelector:
+		@selector(receivedEvent:type:extra:forMode:)])
+    {
+      [receiver receivedEvent: data type: type extra: info forMode: mode];
+    }
+  else
+    {
+      switch (type)
+	{
+	  case ET_RDESC:
+	  case ET_RPORT:
+	    [receiver readyForReadingOnFileDescriptor: (int)info];
+	    break;
+
+	  case ET_WDESC:
+	    [receiver readyForWritingOnFileDescriptor: (int)info];
+	    break;
 	}
     }
-    return NO;
-}
-
-- (void) eventFor: (void*)info mode: (NSString*)mode
-{
-    if ([self isValid] == NO) {
-	return;
-    }
-
-    if ([receiver respondsToSelector:
-		@selector(receivedEvent:type:extra:forMode:)]) {
-	[receiver receivedEvent:data type:type extra:info forMode:mode];
-    }
-    else {
-	switch (type) {
-	    case ET_RDESC:
-	    case ET_RPORT:
-		[receiver readyForReadingOnFileDescriptor: (int)info];
-		break;
-
-	    case ET_WDESC:
-		[receiver readyForWritingOnFileDescriptor: (int)info];
-		break;
-	}
-    }
-}
-
-- (void*) getData
-{
-    return data;
-}
-
-- (NSDate*) getLimit
-{
-    return limit;
-}
-
-- (id) getReceiver
-{
-    return receiver;
-}
-
-- (RunLoopEventType) getType
-{
-    return type;
-}
-
-- (void) increment
-{
-    count++;
 }
 
 - initWithType: (RunLoopEventType)aType
       receiver: (id)anObj
           data: (void*)item
 {
-    self = [super init];
-    if (self) {
-	invalidated = NO;
-	switch (aType) {
-	    case ET_RDESC:	type = aType;	break;
-	    case ET_WDESC:	type = aType;	break;
-	    case ET_RPORT:	type = aType;	break;
-	    default:
-		[NSException raise:NSInvalidArgumentException
-		     format:@"NSRunLoop - unknown event type"];
-	}
-	[self setReceiver:anObj];
-	[self setData:item];
-	[self setLimit:nil];
-	count = 0;
+  invalidated = NO;
+  switch (aType)
+    {
+      case ET_RDESC:	type = aType;	break;
+      case ET_WDESC:	type = aType;	break;
+      case ET_RPORT:	type = aType;	break;
+      default:
+	[NSException raise: NSInvalidArgumentException
+		    format: @"NSRunLoop - unknown event type"];
     }
-    return self;
+  receiver = [anObj retain];
+  data = item;
+  return self;
 }
 
 - (void) invalidate
 {
-    invalidated = YES;
+  invalidated = YES;
 }
 
 - (BOOL) isValid
 {
-    if (invalidated == YES) {
-	return NO;
+  if (invalidated == YES)
+    {
+      return NO;
     }
-    if ([receiver respondsToSelector:@selector(isValid)] &&
-	[receiver isValid] == NO) {
-	[self invalidate];
-	return NO;
+  if ([receiver respondsToSelector: @selector(isValid)] &&
+      [receiver isValid] == NO)
+    {
+      [self invalidate];
+      return NO;
     }
-    return YES;
-}
-
-- (void) setData: (void*)item
-{
-    data = item;
-}
-
-- (void) setLimit: (NSDate*)when
-{
-    NSDate*	d = [when retain];
-
-    [limit release];
-    limit = d;
-}
-
-- (void) setReceiver: anObject
-{
-    id	obj = receiver;
-
-    receiver = [anObject retain];
-
-    [obj release];
+  return YES;
 }
 
 @end
 
 
+
+@interface NSRunLoop (TimedPerformers)
+- (NSMutableArray*) _timedPerformers;
+@end
+
+@implementation	NSRunLoop (TimedPerformers)
+- (NSMutableArray*) _timedPerformers
+{
+  return _timedPerformers;
+}
+@end
+
 /*
  *	The RunLoopPerformer class is used to hold information about
  *	messages which are due to be sent to objects once a particular
@@ -278,77 +230,183 @@ static int debug_run_loop = 0;
  */
 @interface RunLoopPerformer: NSObject
 {
-    SEL		selector;
-    id		target;
-    id		argument;
-    unsigned	order;
-    NSArray*	modes;
+  SEL		selector;
+  id		target;
+  id		argument;
+  unsigned	order;
+  NSArray	*modes;
+  NSTimer	*timer;
 }
 
 - (void) fire;
 - initWithSelector: (SEL)aSelector
-	    target: target
-          argument: argument
+	    target: (id)target
+          argument: (id)argument
              order: (unsigned int)order
              modes: (NSArray*)modes;
 - (BOOL) matchesSelector: (SEL)aSelector
-		  target: aTarget
-		argument: anArgument;
+		  target: (id)aTarget
+		argument: (id)anArgument;
 - (NSArray*) modes;
 - (unsigned int) order;
+- (void) setTimer: (NSTimer*)timer;
 @end
 
 @implementation RunLoopPerformer
 
 - (void) dealloc
 {
-    [target release];
-    [argument release];
-    [modes release];
-    [super dealloc];
+  [timer invalidate];
+  [target release];
+  [argument release];
+  [modes release];
+  [super dealloc];
 }
 
-- (void)fire
+- (void) fire
 {
-    [target perform:selector withObject:argument];
+  if (timer != nil)
+    {
+      timer = nil;
+      [[self retain] autorelease];
+      [[[NSRunLoop currentInstance] _timedPerformers]
+		removeObjectIdenticalTo: self];
+    }
+  [target performSelector: selector withObject: argument];
 }
 
 - initWithSelector: (SEL)aSelector
-	    target: aTarget
-          argument: anArgument
+	    target: (id)aTarget
+          argument: (id)anArgument
              order: (unsigned int)theOrder
              modes: (NSArray*)theModes
 {
-    self = [super init];
-    if (self) {
-	selector = aSelector;
-	target = [aTarget retain];
-	argument = [anArgument retain];
-	order = theOrder;
-	modes = [theModes copy];
+  self = [super init];
+  if (self)
+    {
+      selector = aSelector;
+      target = [aTarget retain];
+      argument = [anArgument retain];
+      order = theOrder;
+      modes = [theModes copy];
     }
-    return self;
+  return self;
 }
 
 - (BOOL) matchesSelector: (SEL)aSelector
-		  target: aTarget
-		argument: anArgument
+		  target: (id)aTarget
+		argument: (id)anArgument
 {
-    if (selector == aSelector)
-	if (target == aTarget)
-	    if ([argument isEqual:anArgument])
-		return YES;
-    return NO;
+  if (selector == aSelector)
+    {
+      if (target == aTarget)
+	{
+	  if ([argument isEqual:anArgument])
+	    {
+	      return YES;
+	    }
+	}
+    }
+  return NO;
 }
 
 - (NSArray*) modes
 {
-    return modes;
+  return modes;
 }
 
 - (unsigned int) order
 {
-    return order;
+  return order;
+}
+
+- (void) setTimer: (NSTimer*)t
+{
+  timer = t;
+}
+@end
+
+@implementation NSObject (TimedPerformers)
+
++ (void) cancelPreviousPerformRequestsWithTarget: (id)target
+					selector: (SEL)aSelector
+					  object: (id)arg
+{
+  NSMutableArray	*array;
+  int			i;
+
+  [target retain];
+  [arg retain];
+  array = [[NSRunLoop currentInstance] _timedPerformers];
+  for (i = [array count]; i > 0; i--)
+    {
+      if ([[array objectAtIndex: i-1] matchesSelector: aSelector
+					       target: target
+					     argument: arg])
+	{
+	  [array removeObjectAtIndex: i-1];
+	}
+    }
+  [arg release];
+  [target release];
+}
+
+- (void) performSelector: (SEL)aSelector
+	      withObject: (id)argument
+	      afterDelay: (NSTimeInterval)seconds
+{
+  NSMutableArray	*array;
+  RunLoopPerformer	*item;
+
+  array = [[NSRunLoop currentInstance] _timedPerformers];
+  item = [[RunLoopPerformer alloc] initWithSelector: aSelector
+					     target: self
+					   argument: argument
+					      order: 0
+					      modes: nil];
+  [array addObject: item];
+  [item setTimer: [NSTimer scheduledTimerWithTimeInterval: seconds
+						   target: item
+						 selector: @selector(fire)
+						 userInfo: nil
+						  repeats: NO]];
+  [item release];
+}
+
+- (void) performSelector: (SEL)aSelector
+	      withObject: (id)argument
+	      afterDelay: (NSTimeInterval)seconds
+		 inModes: (NSArray*)modes
+{
+  NSRunLoop		*loop;
+  NSMutableArray	*array;
+  RunLoopPerformer	*item;
+  NSTimer		*timer;
+  int			i;
+
+  if (modes == nil || [modes count] == 0)
+    {
+      return;
+    }
+  loop = [NSRunLoop currentInstance];
+  array = [loop _timedPerformers];
+  item = [[RunLoopPerformer alloc] initWithSelector: aSelector
+					     target: self
+					   argument: argument
+					      order: 0
+					      modes: nil];
+  [array addObject: item];
+  timer = [NSTimer timerWithTimeInterval: seconds
+				  target: item
+				selector: @selector(fire)
+				userInfo: nil
+				 repeats: NO];
+  [item setTimer: timer];
+  [item release];
+  for (i = 0; i < [modes count]; i++)
+    {
+      [loop addTimer: timer forMode: [modes objectAtIndex: i]];
+    }
 }
 
 @end
@@ -362,6 +420,7 @@ static int debug_run_loop = 0;
 - (RunLoopWatcher*) _getWatcher: (void*)data
 			   type: (RunLoopEventType)type
 		        forMode: (NSString*)mode;
+- (NSMutableArray*) _performers;
 - (void) _removeWatcher: (void*)data
 		   type: (RunLoopEventType)type
 		forMode: (NSString*)mode;
@@ -374,73 +433,97 @@ static int debug_run_loop = 0;
    limit-date order. */
 - (void) _addWatcher: (RunLoopWatcher*) item forMode: (NSString*)mode
 {
-    NSMutableArray	*watchers;
-    id			obj;
-    NSDate		*limit;
-    int			count;
+  NSMutableArray	*watchers;
+  id			obj;
+  NSDate		*limit;
+  int			count;
 
-    watchers = NSMapGet (_mode_2_watchers, mode);
-    if (watchers == nil) {
-	watchers = [NSMutableArray new];
-	NSMapInsert (_mode_2_watchers, mode, watchers);
-	[watchers release];
-	count = 0;
+  watchers = NSMapGet (_mode_2_watchers, mode);
+  if (watchers == nil)
+    {
+      watchers = [NSMutableArray new];
+      NSMapInsert (_mode_2_watchers, mode, watchers);
+      [watchers release];
+      count = 0;
     }
-    else {
-	count = [watchers count];
+  else
+    {
+      count = [watchers count];
     }
 
-    /* If the receiver or its delegate (if any) respond to
-       'limitDateForMode:' then we ask them for the limit date for
-	this watcher.
-     */
-    obj = [item getReceiver];
-    if ([obj respondsToSelector: @selector(limitDateForMode:)]) {
-	[item setLimit: [obj limitDateForMode:mode]];
+  /*
+   *	If the receiver or its delegate (if any) respond to
+   *	'limitDateForMode:' then we ask them for the limit date for
+   *	this watcher.
+   */
+  obj = item->receiver;
+  if ([obj respondsToSelector: @selector(limitDateForMode:)])
+    {
+      NSDate	*d = [obj limitDateForMode: mode];
+
+      ASSIGN(item->limit, d);
     }
-    else if ([obj respondsToSelector: @selector(delegate)]) {
-	obj = [obj delegate];
-	if ([obj respondsToSelector: @selector(limitDateForMode:)]) {
-	    [item setLimit: [obj limitDateForMode:mode]];
+  else if ([obj respondsToSelector: @selector(delegate)])
+    {
+      obj = [obj delegate];
+      if ([obj respondsToSelector: @selector(limitDateForMode:)])
+	{
+	  NSDate	*d = [obj limitDateForMode: mode];
+
+	  ASSIGN(item->limit, d);
 	}
     }
-    limit = [item getLimit];
+  limit = item->limit;
 
-    /* Make sure that the items in the watchers list are ordered. */
-    if (limit == nil || count == 0) {
-	[watchers addObject:item];
+  /*
+   *	Make sure that the items in the watchers list are ordered.
+   */
+  if (limit == nil || count == 0)
+    {
+      [watchers addObject:item];
     }
-    else {
-	int	i;
+  else
+    {
+      int	i;
 
-	for (i = 0; i < count; i++) {
-	    NSDate*	when = [[watchers objectAtIndex:i] getLimit];
+      for (i = 0; i < count; i++)
+	{
+	  RunLoopWatcher	*watcher = [watchers objectAtIndex: i];
+	  NSDate		*when = watcher->limit;
 
-	    if (when == nil || [limit earlierDate:when] == when) {
-		[watchers insertObject:item atIndex:i];
-		break;
+	  if (when == nil || [limit earlierDate:when] == when)
+	    {
+	      [watchers insertObject:item atIndex:i];
+	      break;
 	    }
 	}
-	if (i == count) {
-	    [watchers addObject:item];
+      if (i == count)
+	{
+	  [watchers addObject:item];
 	}
     }
 }
 
 - (void) _checkPerformers
 {
-    RunLoopPerformer	*item;
-    int			count = [_performers count];
-    int			i;
+  RunLoopPerformer	*item;
+  NSArray		*array = [NSArray arrayWithArray: _performers];
+  int			count = [array count];
+  unsigned		pos;
+  int			i;
 
-    for (i = count; i > 0; i++) {
-	item = (RunLoopPerformer*)[_performers objectAtIndex:(i-1)];
+  for (i = 0; i < count; i++)
+    {
+      item = (RunLoopPerformer*)[array objectAtIndex: i];
 
-	if ([[item modes] containsObject: _current_mode]) {
-	    [item retain];
-	    [_performers removeObjectAtIndex:(i-1)];
-	    [item fire];
-	    [item release];
+      pos = [_performers indexOfObjectIdenticalTo: item];
+      if (pos != NSNotFound)
+	{
+	  if ([[item modes] containsObject: _current_mode])
+	    {
+	      [_performers removeObjectAtIndex: pos];
+	      [item fire];
+	    }
 	}
     }
 }
@@ -450,37 +533,37 @@ static int debug_run_loop = 0;
 
 + currentInstance
 {
-    return [self currentRunLoop];
+  return [self currentRunLoop];
 }
 
 + (NSString*) currentMode
 {
-    return [[NSRunLoop currentRunLoop] currentMode];
+  return [[NSRunLoop currentRunLoop] currentMode];
 }
 
 + (void) run
 {
-    [[NSRunLoop currentRunLoop] run];
+  [[NSRunLoop currentRunLoop] run];
 }
 
 + (void) runUntilDate: date
 {
-    [[NSRunLoop currentRunLoop] runUntilDate: date];
+  [[NSRunLoop currentRunLoop] runUntilDate: date];
 }
 
 + (void) runUntilDate: date forMode: (NSString*)mode
 {
-    [[NSRunLoop currentRunLoop] runUntilDate: date forMode: mode];
+  [[NSRunLoop currentRunLoop] runUntilDate: date forMode: mode];
 }
 
 + (BOOL) runOnceBeforeDate: date 
 {
-    return [[NSRunLoop currentRunLoop] runOnceBeforeDate: date];
+  return [[NSRunLoop currentRunLoop] runOnceBeforeDate: date];
 }
 
 + (BOOL) runOnceBeforeDate: date forMode: (NSString*)mode
 {
-    return [[NSRunLoop currentRunLoop] runOnceBeforeDate: date forMode: mode];
+  return [[NSRunLoop currentRunLoop] runOnceBeforeDate: date forMode: mode];
 }
 
 - (void) addEvent: (void*)data
@@ -488,30 +571,32 @@ static int debug_run_loop = 0;
           watcher: (id<RunLoopEvents>)watcher
           forMode: (NSString*)mode
 {
-    RunLoopWatcher	*info;
+  RunLoopWatcher	*info;
 
-    if (mode == nil)
-	mode = _current_mode;
-
-    info = [self _getWatcher: data type: type forMode: mode];
-
-    if (info && [info getReceiver] == watcher) {
-	/* Increment usage count for this watcher. */
-	[info increment];
+  if (mode == nil)
+    {
+      mode = _current_mode;
     }
-    else {
-	/* Remove any existing handler for another watcher. */
-	[self _removeWatcher: data type: type forMode: mode];
 
-	/* Create new object to hold information. */
-	info = [[RunLoopWatcher alloc] initWithType: type
-					   receiver: watcher
-					       data: data];
-	/* Add the object to the array for the mode and keep count. */
-	[self _addWatcher:info forMode:mode];
-	[info increment];
+  info = [self _getWatcher: data type: type forMode: mode];
 
-	[info release];		/* Now held in array.	*/
+  if (info && info->receiver == (id)watcher)
+    {
+      /* Increment usage count for this watcher. */
+      info->count++;
+    }
+  else
+    {
+      /* Remove any existing handler for another watcher. */
+      [self _removeWatcher: data type: type forMode: mode];
+
+      /* Create new object to hold information. */
+      info = [[RunLoopWatcher alloc] initWithType: type
+					 receiver: watcher
+					     data: data];
+      /* Add the object to the array for the mode. */
+      [self _addWatcher:info forMode:mode];
+      [info release];		/* Now held in array.	*/
     }
 }
 
@@ -519,10 +604,10 @@ static int debug_run_loop = 0;
 		    object: (id <FdListening>)listener
 		   forMode: (NSString*)mode
 {
-    return [self addEvent: (void*)fd
-		     type: ET_RDESC
-		  watcher: (id<RunLoopEvents>)listener
-		  forMode: mode];
+  return [self addEvent: (void*)fd
+		   type: ET_RDESC
+		watcher: (id<RunLoopEvents>)listener
+		forMode: mode];
 }
 
   /* Add our new handler information to the array. */
@@ -530,10 +615,10 @@ static int debug_run_loop = 0;
 		     object: (id <FdSpeaking>)speaker
 		    forMode: (NSString*)mode
 {
-    return [self addEvent: (void*)fd
-		     type: ET_WDESC
-		  watcher: (id<RunLoopEvents>)speaker
-		  forMode: mode];
+  return [self addEvent: (void*)fd
+		   type: ET_WDESC
+		watcher: (id<RunLoopEvents>)speaker
+		forMode: mode];
 }
 
 - (void) removeEvent: (void*)data
@@ -541,45 +626,56 @@ static int debug_run_loop = 0;
              forMode: (NSString*)mode
 		 all: (BOOL)removeAll
 {
-    if (mode == nil)
-  	mode = _current_mode;
-  
-    if (removeAll) {
-	[self _removeWatcher: data type: type forMode: mode];
+  if (mode == nil)
+    {
+      mode = _current_mode;
     }
-    else {
-	RunLoopWatcher	*info;
+  if (removeAll)
+    {
+      [self _removeWatcher: data type: type forMode: mode];
+    }
+  else
+    {
+      RunLoopWatcher	*info;
 
-	info = [self _getWatcher: data type: type forMode: mode];
+      info = [self _getWatcher: data type: type forMode: mode];
   
-	if (info && [info decrement] == NO) {
-	    [self _removeWatcher: data type: type forMode: mode];
-  	}
+      if (info)
+	{
+	  if (info->count == 0)
+	    {
+	      [self _removeWatcher: data type: type forMode: mode];
+  	    }
+	  else
+	    {
+	      info->count--;
+	    }
+	}
     }
 }
 
 - (void) removeReadDescriptor: (int)fd 
 		      forMode: (NSString*)mode
 {
-    return [self removeEvent:(void*)fd type: ET_RDESC forMode:mode all:NO];
+  return [self removeEvent:(void*)fd type: ET_RDESC forMode:mode all:NO];
 }
 
 - (void) removeWriteDescriptor: (int)fd
 		       forMode: (NSString*)mode
 {
-    return [self removeEvent:(void*)fd type: ET_WDESC forMode:mode all:NO];
+  return [self removeEvent:(void*)fd type: ET_WDESC forMode:mode all:NO];
 }
 
 - (BOOL) runOnceBeforeDate: date
 {
-    return [self runOnceBeforeDate: date forMode: _current_mode];
+  return [self runOnceBeforeDate: date forMode: _current_mode];
 }
 
 /* Running the run loop once through for timers and input listening. */
 
 - (BOOL) runOnceBeforeDate: date forMode: (NSString*)mode
 {
-    return [self runMode:mode beforeDate:date];
+  return [self runMode:mode beforeDate:date];
 }
 
 - (void) runUntilDate: date forMode: (NSString*)mode
@@ -608,17 +704,16 @@ static int debug_run_loop = 0;
 
 + currentRunLoop
 {
+  static NSString	*key = @"NSRunLoopThreadKey";
   NSRunLoop*	r;
   NSThread*	t;
-  NSString*	key;
 
   t = [NSThread currentThread];
-  key = @"NSRunLoopThreadKey";
-  r = [[t threadDictionary] objectForKey:key];
+  r = [[t threadDictionary] objectForKey: key];
   if (r == nil)
     {
       r = [NSRunLoop new];
-      [[t threadDictionary] setObject:r forKey:key];
+      [[t threadDictionary] setObject: r forKey: key];
       [r release];
     }
   return r;
@@ -639,16 +734,18 @@ static int debug_run_loop = 0;
 				     NSObjectMapValueCallBacks, 0);
   _mode_2_watchers = NSCreateMapTable (NSObjectMapKeyCallBacks,
 					   NSObjectMapValueCallBacks, 0);
-  _performers = [[NSMutableArray arrayWithCapacity:8] retain];
+  _performers = [[NSMutableArray alloc] initWithCapacity:8];
+  _timedPerformers = [[NSMutableArray alloc] initWithCapacity:8];
   return self;
 }
 
 - (void) dealloc
 {
-    NSFreeMapTable(_mode_2_timers);
-    NSFreeMapTable(_mode_2_watchers);
-    [_performers release];
-    [super dealloc];
+  NSFreeMapTable(_mode_2_timers);
+  NSFreeMapTable(_mode_2_watchers);
+  [_performers release];
+  [_timedPerformers release];
+  [super dealloc];
 }
 
 - (NSString*) currentMode
@@ -681,204 +778,238 @@ static int debug_run_loop = 0;
 
 - limitDateForMode: (NSString*)mode
 {
-    id			saved_mode;
-    Heap		*timers;
-    NSTimer		*min_timer = nil;
-    RunLoopWatcher	*min_watcher = nil;
-    NSArray		*watchers;
-    NSDate		*when;
+  id			saved_mode;
+  Heap			*timers;
+  NSTimer		*min_timer = nil;
+  RunLoopWatcher	*min_watcher = nil;
+  NSMutableArray	*watchers;
+  NSDate		*when;
 
-    saved_mode = _current_mode;
-    _current_mode = mode;
+  saved_mode = _current_mode;
+  _current_mode = mode;
 
-    timers = NSMapGet(_mode_2_timers, mode);
-    if (timers) {
-	while ((min_timer = [timers minObject]) != nil) {
-	    if (![min_timer isValid]) {
-		[timers removeFirstObject];
-		min_timer = nil;
-		continue;
+  timers = NSMapGet(_mode_2_timers, mode);
+  if (timers)
+    {
+      while ((min_timer = [timers minObject]) != nil)
+	{
+	  if (![min_timer isValid])
+	    {
+	      [timers removeFirstObject];
+	      min_timer = nil;
+	      continue;
 	    }
 
-	    if ([[min_timer fireDate] timeIntervalSinceNow] > 0) {
-		break;
+	  if ([[min_timer fireDate] timeIntervalSinceNow] > 0)
+	    {
+	      break;
 	    }
 
-	    [min_timer retain];
-	    [timers removeFirstObject];
-	    /* Firing will also increment its fireDate, if it is repeating. */
-	    [min_timer fire];
-	    if ([min_timer isValid]) {
-		[timers addObject: min_timer];
+	  [min_timer retain];
+	  [timers removeFirstObject];
+	  /* Firing will also increment its fireDate, if it is repeating. */
+	  [min_timer fire];
+	  if ([min_timer isValid])
+	    {
+	      [timers addObject: min_timer];
 	    }
-	    [min_timer release];
-	    min_timer = nil;
-	    [NSNotificationQueue runLoopASAP];	/* Post notifications. */
+	  [min_timer release];
+	  min_timer = nil;
+	  [NSNotificationQueue runLoopASAP];	/* Post notifications. */
 	}
     }
 
-    /* Is this right? At the moment we invalidate and discard watchers
-       whose limit-dates have passed. */
-    watchers = NSMapGet(_mode_2_watchers, mode);
-    if (watchers) {
-	while ([watchers count] > 0) {
-	    min_watcher = (RunLoopWatcher*)[watchers objectAtIndex:0];
+  /* Is this right? At the moment we invalidate and discard watchers
+     whose limit-dates have passed. */
+  watchers = NSMapGet(_mode_2_watchers, mode);
+  if (watchers)
+    {
+      while ([watchers count] > 0)
+	{
+	  min_watcher = (RunLoopWatcher*)[watchers objectAtIndex:0];
 
-	    if (![min_watcher isValid]) {
-		[watchers removeObjectAtIndex:0];
-		min_watcher = nil;
-		continue;
+	  if (![min_watcher isValid])
+	    {
+	      [watchers removeObjectAtIndex: 0];
+	      min_watcher = nil;
+	      continue;
 	    }
 
-	    when = [min_watcher getLimit];
-	    if (when == nil || [when timeIntervalSinceNow] > 0) {
-		break;
+	  when = min_watcher->limit;
+	  if (when == nil || [when timeIntervalSinceNow] > 0)
+	    {
+	      break;
 	    }
-	    else {
-		id	obj;
-		NSDate	*nxt = nil;
+	  else
+	    {
+	      id	obj;
+	      NSDate	*nxt = nil;
 
-		/*
- 		 *	If the receiver or its delegate wants to know about
-		 *	timeouts - inform it and give it a chance to set a
-		 *	revised limit date.
-		 */
-		obj = [min_watcher getReceiver];
-		if ([obj respondsToSelector:
-			@selector(timedOutEvent:type:forMode:)]) {
-		    nxt = [obj timedOutEvent:[min_watcher getData]
-				        type:[min_watcher getType]
-				     forMode:_current_mode];
+	      /*
+	       *	If the receiver or its delegate wants to know about
+	       *	timeouts - inform it and give it a chance to set a
+	       *	revised limit date.
+	       */
+	      obj = min_watcher->receiver;
+	      if ([obj respondsToSelector:
+		      @selector(timedOutEvent:type:forMode:)])
+		{
+		  nxt = [obj timedOutEvent: min_watcher->data
+				      type: min_watcher->type
+				   forMode: _current_mode];
 		}
-		else if ([obj respondsToSelector:@selector(delegate)]) {
-		    obj = [obj delegate];
-		    if ([obj respondsToSelector:
-			    @selector(timedOutEvent:type:forMode:)]) {
-			nxt = [obj timedOutEvent:[min_watcher getData]
-					    type:[min_watcher getType]
-					 forMode:_current_mode];
+	      else if ([obj respondsToSelector:@selector(delegate)])
+		{
+		  obj = [obj delegate];
+		  if ([obj respondsToSelector:
+			    @selector(timedOutEvent:type:forMode:)])
+		    {
+		      nxt = [obj timedOutEvent: min_watcher->data
+					  type: min_watcher->type
+				       forMode: _current_mode];
 		    }
 		}
-		if (nxt && [nxt timeIntervalSinceNow] > 0.0) {
-		    /*
-		     *	If the watcher has been given a revised limit date -
-		     *	re-insert it into the queue in the correct place.
-		     */
-		    [min_watcher retain];
-		    [min_watcher setLimit:nxt];
-		    [watchers removeObjectAtIndex:0];
-		    [self _addWatcher:min_watcher forMode:mode];
-		    [min_watcher release];
+	      if (nxt && [nxt timeIntervalSinceNow] > 0.0)
+		{
+		  /*
+		   *	If the watcher has been given a revised limit date -
+		   *	re-insert it into the queue in the correct place.
+		   */
+		  [min_watcher retain];
+		  ASSIGN(min_watcher->limit, nxt);
+		  [watchers removeObjectAtIndex: 0];
+		  [self _addWatcher: min_watcher forMode: mode];
+		  [min_watcher release];
 		}
-		else {
-		    /*
-		     *	If the watcher is now useless - invalidate it and
-		     *	remove it from the queue so that we don't need to
-		     *	check it again.
-		     */
-		    [min_watcher invalidate];
-		    [watchers removeObjectAtIndex:0];
+	      else
+		{
+		  /*
+		   *	If the watcher is now useless - invalidate it and
+		   *	remove it from the queue so that we don't need to
+		   *	check it again.
+		   */
+		  [min_watcher invalidate];
+		  [watchers removeObjectAtIndex:0];
 		}
-		min_watcher = nil;
+	      min_watcher = nil;
 	    }
 	}
     }
 
-    /*
-     *	If there are timers - set limit date to the earliest of them.
-     */
-    if (min_timer) {
-	when = [min_timer fireDate];
+  /*
+   *	If there are timers - set limit date to the earliest of them.
+   */
+  if (min_timer)
+    {
+      when = [min_timer fireDate];
     }
-    else {
-	when = nil;
+  else
+    {
+      when = nil;
     }
 
-    /*
-     *	If there are watchers, set the limit date to that of the earliest
-     *	watcher (or leave it as the date of the earliest timer if that is
-     *	before the watchers limit).
-     *	NB. A watcher without a limit date watches forever - so it's limit
-     *	is effectively some time in the distant future.
-     */
-    if (min_watcher) {
-	NSDate*	lim;
+  /*
+   *	If there are watchers, set the limit date to that of the earliest
+   *	watcher (or leave it as the date of the earliest timer if that is
+   *	before the watchers limit).
+   *	NB. A watcher without a limit date watches forever - so it's limit
+   *	is effectively some time in the distant future.
+   */
+  if (min_watcher)
+    {
+      NSDate*	lim;
 
-	if ([min_watcher getLimit] == nil) {	/* No limit for watcher	*/
-	    lim = [NSDate distantFuture];	/* - watches forever.	*/
+      if (min_watcher->limit == nil)		/* No limit for watcher	*/
+	{
+	  lim = [NSDate distantFuture];		/* - watches forever.	*/
 	}
-	else {
-	    lim = [min_watcher getLimit];
+      else
+	{
+	  lim = min_watcher->limit;
 	}
 
-	if (when == nil) {
-	    when = lim;
+      if (when == nil)
+	{
+	  when = lim;
 	}
-	else {
-	    when = [when earlierDate:lim];
+      else
+	{
+	  when = [when earlierDate:lim];
 	}
     }
 
-    /*
-     *	'when' will only be nil if there are neither timers nor watchers
-     *	outstanding.
-     */
-    if (when && debug_run_loop) {
-	printf ("\tNSRunLoop limit date %f\n",
+  /*
+   *	'when' will only be nil if there are neither timers nor watchers
+   *	outstanding.
+   */
+  if (when && debug_run_loop)
+    {
+      printf ("\tNSRunLoop limit date %f\n",
 	    [when timeIntervalSinceReferenceDate]);
     }
-    _current_mode = saved_mode;
+  _current_mode = saved_mode;
 
-    return when;
+  return when;
 }
 
 - (RunLoopWatcher*) _getWatcher: (void*)data
 			   type: (RunLoopEventType)type
 		        forMode: (NSString*)mode
 {
-    NSArray		*watchers;
-    RunLoopWatcher	*info;
-    int			count;
+  NSArray		*watchers;
+  RunLoopWatcher	*info;
+  int			count;
 
-    if (mode == nil)
-	mode = _current_mode;
-
-    watchers = NSMapGet (_mode_2_watchers, mode);
-    if (watchers == nil) {
-	return nil;
+  if (mode == nil)
+    {
+      mode = _current_mode;
     }
-    for (count = 0; count < [watchers count]; count++) {
-	info = [watchers objectAtIndex: count];
 
-	if ([info getType] == type) {
-	    if ([info getData] == data) {
-		return info;
+  watchers = NSMapGet (_mode_2_watchers, mode);
+  if (watchers == nil)
+    {
+      return nil;
+    }
+  for (count = 0; count < [watchers count]; count++)
+    {
+      info = [watchers objectAtIndex: count];
+
+      if (info->type == type)
+	{
+	  if (info->data == data)
+	    {
+	      return info;
 	    }
 	}
     }
-    return nil;
+  return nil;
 }
 
 - (void) _removeWatcher: (void*)data
                    type: (RunLoopEventType)type
                 forMode: (NSString*)mode
 {
-    NSMutableArray	*watchers;
+  NSMutableArray	*watchers;
 
-    if (mode == nil )
-	mode = _current_mode;
+  if (mode == nil)
+    {
+      mode = _current_mode;
+    }
 
-    watchers = NSMapGet (_mode_2_watchers, mode);
-    if (watchers) {
-	int	i;
+  watchers = NSMapGet (_mode_2_watchers, mode);
+  if (watchers)
+    {
+      int	i;
 
-	for (i = [watchers count]; i > 0; i--) {
-	    RunLoopWatcher*	info;
+      for (i = [watchers count]; i > 0; i--)
+	{
+	  RunLoopWatcher*	info;
 
-	    info = (RunLoopWatcher*)[watchers objectAtIndex:(i-1)];
-	    if ([info getType] == type && [info getData] == data) {
-		[info invalidate];
-		[watchers removeObject: info];
+	  info = (RunLoopWatcher*)[watchers objectAtIndex:(i-1)];
+	  if (info->type == type && info->data == data)
+	    {
+	      [info invalidate];
+	      [watchers removeObject: info];
 	    }
 	}
     }
@@ -907,7 +1038,7 @@ static int debug_run_loop = 0;
   id saved_mode;
   int num_inputs = 0;
 
-  assert (mode);
+  NSAssert(mode, NSInvalidArgumentException);
   saved_mode = _current_mode;
   _current_mode = mode;
 
@@ -967,7 +1098,7 @@ static int debug_run_loop = 0;
 
   /* Do the pre-listening set-up for the file descriptors of this mode. */
   {
-      NSArray	*watchers;
+      NSMutableArray	*watchers;
 
       watchers = NSMapGet (_mode_2_watchers, mode);
       if (watchers) {
@@ -978,19 +1109,19 @@ static int debug_run_loop = 0;
 	      int fd;
 
 	      if ([info isValid] == NO) {
-		[watchers removeObjectAtIndex:(i-1)];
+		[watchers removeObjectAtIndex: (i-1)];
 		continue;
               }
-	      switch ([info getType]) {
+	      switch (info->type) {
 		case ET_WDESC:
-	          fd = (int)[info getData];
+	          fd = (int)info->data;
 	          FD_SET (fd, &write_fds);
 	          NSMapInsert (wfd_2_object, (void*)fd, info);
 	          num_inputs++;
 		  break;
 
 		case ET_RDESC:
-	          fd = (int)[info getData];
+	          fd = (int)info->data;
 	          FD_SET (fd, &fds);
 	          NSMapInsert (rfd_2_object, (void*)fd, info);
 	          num_inputs++;
@@ -998,7 +1129,7 @@ static int debug_run_loop = 0;
 
 		case ET_RPORT:
 		  {
-		    id	port = [info getReceiver];
+		    id	port = info->receiver;
 		    int port_fd_count = 128; // xxx #define this constant
 		    int port_fd_array[port_fd_count];
 
@@ -1046,13 +1177,20 @@ static int debug_run_loop = 0;
 
   if (select_return < 0)
     {
-      /* Some exceptional condition happened. */
-      /* xxx We can do something with exception_fds, instead of
-	 aborting here. */
-      perror ("[TcpInPort receivePacketWithTimeout:] select()");
-      abort ();
+      if (errno == EINTR)
+	{
+	  select_return = 0;
+	}
+      else
+	{
+	  /* Some exceptional condition happened. */
+	  /* xxx We can do something with exception_fds, instead of
+	     aborting here. */
+	  perror ("[NSRunLoop acceptInputForMode:beforeDate:] select()");
+	  abort ();
+	}
     }
-  else if (select_return == 0)
+  if (select_return == 0)
     {
       NSFreeMapTable (rfd_2_object);
       NSFreeMapTable (wfd_2_object);
@@ -1069,14 +1207,14 @@ static int debug_run_loop = 0;
       if (FD_ISSET (fd_index, &write_fds))
         {
 	  id watcher = (id) NSMapGet (wfd_2_object, (void*)fd_index);
-	  assert (watcher);
+	  NSAssert(watcher, NSInternalInconsistencyException);
 	  [watcher eventFor:(void*)fd_index mode:_current_mode];
           [NSNotificationQueue runLoopASAP];
         }
       if (FD_ISSET (fd_index, &read_fds))
         {
 	  id watcher = (id) NSMapGet (rfd_2_object, (void*)fd_index);
-	  assert (watcher);
+	  NSAssert(watcher, NSInternalInconsistencyException);
 	  [watcher eventFor:(void*)fd_index mode:_current_mode];
           [NSNotificationQueue runLoopASAP];
         }
@@ -1092,33 +1230,37 @@ static int debug_run_loop = 0;
 
 - (BOOL) runMode: (NSString*)mode beforeDate: date
 {
-    id	d;
+  id	d;
 
-    /* If date has already passed, simply return. */
-    if ([date timeIntervalSinceNow] < 0) {
-	if (debug_run_loop) {
-	    printf ("\tNSRunLoop run mode with date already past\n");
+  /* If date has already passed, simply return. */
+  if ([date timeIntervalSinceNow] < 0)
+    {
+      if (debug_run_loop)
+	{
+	  printf ("\tNSRunLoop run mode with date already past\n");
 	}
-	return NO;
+      return NO;
     }
 
-    /* Find out how long we can wait before first limit date. */
-    d = [self limitDateForMode: mode];
-    if (d == nil) {
-	if (debug_run_loop) {
-	    printf ("\tNSRunLoop run mode with nothing to do\n");
+  /* Find out how long we can wait before first limit date. */
+  d = [self limitDateForMode: mode];
+  if (d == nil)
+    {
+      if (debug_run_loop)
+	{
+	  printf ("\tNSRunLoop run mode with nothing to do\n");
 	}
-	return NO;
+      return NO;
     }
 
-    /* Use the earlier of the two dates we have. */
-    d = [[d earlierDate:date] retain];
+  /* Use the earlier of the two dates we have. */
+  d = [[d earlierDate:date] retain];
 
-    /* Wait, listening to our input sources. */
-    [self acceptInputForMode: mode beforeDate: d];
+  /* Wait, listening to our input sources. */
+  [self acceptInputForMode: mode beforeDate: d];
 
-    [d release];
-    return YES;
+  [d release];
+  return YES;
 }
 
 - (void) run
@@ -1145,27 +1287,33 @@ id NSDefaultRunLoopMode = @"NSDefaultRunLoopMode";
 - (void) addPort: (NSPort*)port
          forMode: (NSString*)mode
 {
-    return [self addEvent: (void*)port
-		     type: ET_RPORT
-		  watcher: (id<RunLoopEvents>)port
-		  forMode: (NSString*)mode];
+  return [self addEvent: (void*)port
+		   type: ET_RPORT
+		watcher: (id<RunLoopEvents>)port
+		forMode: (NSString*)mode];
 }
 
 - (void) cancelPerformSelector: (SEL)aSelector
 			target: target
 		      argument: argument
 {
-    RunLoopPerformer	*item;
-    int			count = [_performers count];
-    int			i;
+  RunLoopPerformer	*item;
+  int			count = [_performers count];
+  int			i;
 
-    for (i = count; i > 0; i++) {
-	item = (RunLoopPerformer*)[_performers objectAtIndex:(i-1)];
+  [target retain];
+  [argument retain];
+  for (i = count; i > 0; i--)
+    {
+      item = (RunLoopPerformer*)[_performers objectAtIndex:(i-1)];
 
-	if ([item matchesSelector:aSelector target:target argument:argument]) {
-	    [_performers removeObjectAtIndex:(i-1)];
+      if ([item matchesSelector:aSelector target:target argument:argument])
+	{
+	  [_performers removeObjectAtIndex:(i-1)];
 	}
     }
+  [argument release];
+  [target release];
 }
 
 - (void) configureAsServer
@@ -1179,38 +1327,43 @@ id NSDefaultRunLoopMode = @"NSDefaultRunLoopMode";
 		   order: (unsigned int)order
 		   modes: (NSArray*)modes
 {
-    RunLoopPerformer	*item;
-    int			count = [_performers count];
+  RunLoopPerformer	*item;
+  int			count = [_performers count];
 
-    item = [[RunLoopPerformer alloc] initWithSelector: aSelector
-					       target: target
-          				     argument: argument
-             					order: order
-						modes: modes];
-    /* Add new item to list - reverse ordering */
-    if (count == 0) {
-	[_performers addObject:item];
+  item = [[RunLoopPerformer alloc] initWithSelector: aSelector
+					     target: target
+					   argument: argument
+					      order: order
+					      modes: modes];
+  /* Add new item to list - reverse ordering */
+  if (count == 0)
+    {
+      [_performers addObject:item];
     }
-    else {
-	int	i;
+  else
+    {
+      int	i;
 
-	for (i = 0; i < count; i++) {
-	    if ([[_performers objectAtIndex:i] order] <= order) {
-		[_performers insertObject:item atIndex:i];
-		break;
+      for (i = 0; i < count; i++)
+	{
+	  if ([[_performers objectAtIndex:i] order] <= order)
+	    {
+	      [_performers insertObject:item atIndex:i];
+	      break;
 	    }
 	}
-	if (i == count) {
-	    [_performers addObject:item];
+      if (i == count)
+	{
+	  [_performers addObject:item];
 	}
     }
-    [item release];
+  [item release];
 }
 
 - (void) removePort: (NSPort*)port
             forMode: (NSString*)mode
 {
-    return [self removeEvent:(void*)port type: ET_RPORT forMode:mode all:NO];
+  return [self removeEvent:(void*)port type: ET_RPORT forMode:mode all:NO];
 }
 
 @end
