@@ -1,5 +1,5 @@
 /* Implementation of GNUSTEP string class
-   Copyright (C) 1995, 1996, 1997 Free Software Foundation, Inc.
+   Copyright (C) 1995, 1996, 1997, 1998 Free Software Foundation, Inc.
    
    Written by:  Andrew Kachites McCallum <mccallum@gnu.ai.mit.edu>
    Date: January 1995
@@ -7,6 +7,9 @@
    Unicode implementation by Stevo Crvenkovski <stevo@btinternet.com>
    Date: February 1997
 
+   Optimisations by Richard Frith-Macdonald <richard@brainstorm.co.uk>
+   Date: October 1998
+
    This file is part of the GNUstep Base Library.
 
    This library is free software; you can redistribute it and/or
@@ -37,8 +40,8 @@
 */
 
 #include <config.h>
-#include <gnustep/base/preface.h>
-#include <gnustep/base/Coding.h>
+#include <base/preface.h>
+#include <base/Coding.h>
 #include <Foundation/NSString.h>
 #include <Foundation/NSArray.h>
 #include <Foundation/NSCharacterSet.h>
@@ -46,10 +49,14 @@
 #include <Foundation/NSValue.h>
 #include <Foundation/NSDictionary.h>
 #include <Foundation/NSUserDefaults.h>
-#include <gnustep/base/IndexedCollection.h>
+#include <Foundation/NSFileManager.h>
+#include <Foundation/NSPortCoder.h>
+#include <Foundation/NSPathUtilities.h>
+
+#include <base/IndexedCollection.h>
 #include <Foundation/NSData.h>
 #include <Foundation/NSBundle.h>
-#include <gnustep/base/IndexedCollectionPrivate.h>
+#include <base/IndexedCollectionPrivate.h>
 #include <limits.h>
 #include <string.h>		// for strstr()
 #include <sys/stat.h>
@@ -58,12 +65,16 @@
 #include <fcntl.h>
 #include <stdio.h>
 
-#include <gnustep/base/MallocAddress.h>
-#include <gnustep/base/behavior.h>
+#include <base/behavior.h>
+
+#include <base/NSGSequence.h>
+#include <base/Unicode.h>
+#include <base/GetDefEncoding.h>
+#include <base/NSGString.h>
+#include <base/NSGCString.h>
+
+#include <base/fast.x>
 
-#include <gnustep/base/NSGSequence.h>
-#include <gnustep/base/Unicode.h>
-#include <gnustep/base/GetDefEncoding.h>
 
 // Uncomment when implemented
     static NSStringEncoding _availableEncodings[] = {
@@ -210,7 +221,7 @@ handle_printf_atsign (FILE *stream,
 #endif
   len = fprintf(stream, "%*s",
 		(info->left ? - info->width : info->width),
-		[[string_object description] cStringNoCopy]);
+		[[string_object description] cString]);
   return len;
 }
 #endif /* HAVE_REGISTER_PRINTF_FUNCTION */
@@ -268,14 +279,14 @@ handle_printf_atsign (FILE *stream,
 
 + (NSString*) stringWithCString: (const char*) byteString
 {
-  return [[[self alloc] initWithCString:byteString]
+  return [[[NSString_c_concrete_class alloc] initWithCString:byteString]
 	  autorelease];
 }
 
 + (NSString*) stringWithCString: (const char*)byteString
    length: (unsigned int)length
 {
-  return [[[self alloc]
+  return [[[NSString_c_concrete_class alloc]
 	   initWithCString:byteString length:length]
 	  autorelease];
 }
@@ -311,28 +322,58 @@ handle_printf_atsign (FILE *stream,
 
 /* This is the designated initializer for Unicode Strings. */
 - (id) initWithCharactersNoCopy: (unichar*)chars
-   length: (unsigned int)length
-   freeWhenDone: (BOOL)flag
+			 length: (unsigned int)length
+		       fromZone: (NSZone*)zone
 {
   [self subclassResponsibility:_cmd];
   return self;
 }
 
-- (id) initWithCharacters: (const unichar*)chars
+- (id) initWithCharactersNoCopy: (unichar*)chars
    length: (unsigned int)length
+   freeWhenDone: (BOOL)flag
 {
-  unichar *s;
-  OBJC_MALLOC(s, unichar, length+1);
-  if (chars)
-    memcpy(s, chars,2*length);
-  s[length] = (unichar)0;
-  return [self initWithCharactersNoCopy:s length:length freeWhenDone:YES];
+  if (flag)
+    return [self initWithCharactersNoCopy: chars
+				   length: length
+				 fromZone: NSZoneFromPointer(chars)];
+  else
+    return [self initWithCharactersNoCopy: chars
+				   length: length
+				 fromZone: 0];
+  return self;
+}
+
+- (id) initWithCharacters: (const unichar*)chars
+		   length: (unsigned int)length
+{
+    NSZone	*z = [self zone];
+    unichar	*s = NSZoneMalloc(z, sizeof(unichar)*length);
+
+    if (chars)
+	memcpy(s, chars, sizeof(unichar)*length);
+
+    return [self initWithCharactersNoCopy:s length:length fromZone:z];
+}
+
+- (id) initWithCStringNoCopy: (char*)byteString
+		      length: (unsigned int)length
+		freeWhenDone: (BOOL)flag
+{
+  if (flag)
+    return [self initWithCStringNoCopy: byteString
+				length: length
+			      fromZone: NSZoneFromPointer(byteString)];
+  else
+    return [self initWithCStringNoCopy: byteString
+				length: length
+			      fromZone: 0];
 }
 
 /* This is the designated initializer for CStrings. */
 - (id) initWithCStringNoCopy: (char*)byteString
-   length: (unsigned int)length
-   freeWhenDone: (BOOL)flag
+		      length: (unsigned int)length
+		    fromZone: (NSZone*)zone
 {
   [self subclassResponsibility:_cmd];
   return self;
@@ -340,12 +381,12 @@ handle_printf_atsign (FILE *stream,
 
 - (id) initWithCString: (const char*)byteString  length: (unsigned int)length
 {
-  char *s;
-  OBJC_MALLOC(s, char, length+1);
-  if (byteString)
-    memcpy(s, byteString, length);
-  s[length] = '\0';
-  return [self initWithCStringNoCopy:s length:length freeWhenDone:YES];
+    NSZone	*z = [self zone];
+    char	*s = NSZoneMalloc(z, length);
+
+    if (byteString)
+	memcpy(s, byteString, length);
+    return [self initWithCStringNoCopy:s length:length fromZone:z];
 }
 
 - (id) initWithCString: (const char*)byteString
@@ -356,14 +397,14 @@ handle_printf_atsign (FILE *stream,
 
 - (id) initWithString: (NSString*)string
 {
-  unichar *s;
-  unsigned length = [string length];
-  OBJC_MALLOC(s, unichar, length+1);
-  [string getCharacters:s];
-  s[length] = (unichar)0;
-  return [self initWithCharactersNoCopy: s
-				 length: length
-			   freeWhenDone: YES];
+    NSZone	*z = [self zone];
+    unsigned	length = [string length];
+    unichar	*s = NSZoneMalloc(z, sizeof(unichar)*length);
+
+    [string getCharacters:s];
+    return [self initWithCharactersNoCopy: s
+				   length: length
+				 fromZone: z];
 }
 
 - (id) initWithFormat: (NSString*)format,...
@@ -380,7 +421,7 @@ handle_printf_atsign (FILE *stream,
    arguments: (va_list)arg_list
 {
 #if HAVE_VSPRINTF
-  const char *format_cp = [format cStringNoCopy];
+  const char *format_cp = [format cString];
   int format_len = strlen (format_cp);
   /* xxx horrible disgusting BUFFER_EXTRA arbitrary limit; fix this! */
   #define BUFFER_EXTRA 1024*500
@@ -455,7 +496,7 @@ handle_printf_atsign (FILE *stream,
 	    format_to_go = spec_pos+1;
 	  }
 	/* Get a C-string (char*) from the String object, and print it. */
-	cstring = [[(id) va_arg (arg_list, id) description] cStringNoCopy];
+	cstring = [[(id) va_arg (arg_list, id) description] cString];
 	if (!cstring)
 	  cstring = "<null string>";
 	strcat (buf+printed_len, cstring);
@@ -500,42 +541,39 @@ handle_printf_atsign (FILE *stream,
 - (id) initWithData: (NSData*)data
    encoding: (NSStringEncoding)encoding
 {
-  if((encoding==[NSString defaultCStringEncoding])
-  || (encoding==NSASCIIStringEncoding))
-  {
-    char *s;
+  if ((encoding==[NSString defaultCStringEncoding])
+    || (encoding==NSASCIIStringEncoding))
+    {
+      NSZone *z = fastZone(self);
+      int len=[data length];
+      char *s = NSZoneMalloc(z, len+1);
 
-    int len=[data length];
-    OBJC_MALLOC(s, char, len+1);
-    [data getBytes:s];
-    s[len]=0;
-      return [self initWithCStringNoCopy:s length:len freeWhenDone:YES];
+      [data getBytes:s];
+      return [self initWithCStringNoCopy:s length:len fromZone:z];
     }
   else
-  {
-    unichar *u;
-    int count;
-
-    int len=[data length];
-    const unsigned char *b=[data bytes];
-    OBJC_MALLOC(u, unichar, len+1);
-
-    if(encoding==NSUnicodeStringEncoding)
     {
-      if((b[0]==0xFE)&(b[1]==0xFF))
-        for(count=2;count<(len-1);count+=2)
-          u[count/2 - 1]=256*b[count]+b[count+1];
-      else
-        for(count=2;count<(len-1);count+=2)
-          u[count/2 -1]=256*b[count+1]+b[count];
-      count = count/2 -1;
-    }
-    else
-      count = encode_strtoustr(u,b,len,encoding);
+      NSZone *z = fastZone(self);
+      int len=[data length];
+      unichar *u = NSZoneMalloc(z, sizeof(unichar)*(len+1));
+      int count;
+      const unsigned char *b=[data bytes];
 
-    u[count]=(unichar)0;
-     return [self initWithCharactersNoCopy:u length:count freeWhenDone:YES];
-  }
+      if(encoding==NSUnicodeStringEncoding)
+        {
+	  if((b[0]==0xFE)&(b[1]==0xFF))
+	    for(count=2;count<(len-1);count+=2)
+	      u[count/2 - 1]=256*b[count]+b[count+1];
+	  else
+	    for(count=2;count<(len-1);count+=2)
+	      u[count/2 -1]=256*b[count+1]+b[count];
+	  count = count/2 -1;
+	}
+      else
+	count = encode_strtoustr(u,b,len,encoding);
+
+      return [self initWithCharactersNoCopy:u length:count fromZone:z];
+    }
   return self;
 }
 
@@ -557,7 +595,7 @@ handle_printf_atsign (FILE *stream,
 
 - (id) init
 {
- [super init];
+ self = [super init];
  return self;
 }
 
@@ -610,17 +648,16 @@ handle_printf_atsign (FILE *stream,
 
 - (NSString*) stringByAppendingString: (NSString*)aString
 {
+  NSZone *z = fastZone(self);
   unsigned len = [self length];
   unsigned otherLength = [aString length];
-  unichar *s;
+  unichar *s = NSZoneMalloc(z, (len+otherLength)*sizeof(unichar));
   NSString *tmp;
-  OBJC_MALLOC(s, unichar, len+otherLength+1);
+
   [self getCharacters:s];
   [aString getCharacters:s+len];
-  s[len + otherLength]=(unichar) 0;
-  tmp = [[[self class] alloc] initWithCharactersNoCopy: s
-						length: len+otherLength
-					  freeWhenDone: YES];
+  tmp = [[[self class] allocWithZone:z] initWithCharactersNoCopy: s
+		    length: len+otherLength fromZone: z];
   return [tmp autorelease];
 }
 
@@ -663,6 +700,7 @@ handle_printf_atsign (FILE *stream,
 
 - (NSString*) substringFromRange: (NSRange)aRange
 {
+  NSZone *z;
   unichar *buf;
   id ret;
 
@@ -672,11 +710,12 @@ handle_printf_atsign (FILE *stream,
     [NSException raise: NSRangeException format:@"Invalid location+length."];
   if (aRange.length == 0)
     return @"";
-  OBJC_MALLOC(buf, unichar, aRange.length+1);
+  z = fastZone(self);
+  buf = NSZoneMalloc(z, sizeof(unichar)*aRange.length);
   [self getCharacters:buf range:aRange];
   ret = [[[self class] alloc] initWithCharactersNoCopy: buf
 						length: aRange.length
-					  freeWhenDone: YES];
+					      fromZone: z];
   return [ret autorelease];
 }
 
@@ -711,25 +750,28 @@ handle_printf_atsign (FILE *stream,
 
 /* xxx FIXME */
 - (NSRange) rangeOfCharacterFromSet: (NSCharacterSet*)aSet
-    options: (unsigned int)mask
-    range: (NSRange)aRange
+			    options: (unsigned int)mask
+			      range: (NSRange)aRange
 {
   int i, start, stop, step;
   NSRange range;
 
-  /* xxx check to make sure aRange is within self; raise NSStringBoundsError */
-  assert(NSMaxRange(aRange) <= [self length]);
+  i = [self length];
+  if (aRange.location > i)
+    [NSException raise: NSRangeException format:@"Invalid location."];
+  if (aRange.length > (i - aRange.location))
+    [NSException raise: NSRangeException format:@"Invalid location+length."];
 
   if ((mask & NSBackwardsSearch) == NSBackwardsSearch)
     {
-      start = NSMaxRange(aRange); stop = aRange.location; step = -1;
+      start = NSMaxRange(aRange)-1; stop = aRange.location-1; step = -1;
     }
   else
     {
       start = aRange.location; stop = NSMaxRange(aRange); step = 1;
     }
   range.length = 0;
-  for (i = start; i < stop; i+=step)
+  for (i = start; i != stop; i+=step)
     {
       unichar letter = [self characterAtIndex:i];
       if ([aSet characterIsMember:letter])
@@ -1304,6 +1346,8 @@ handle_printf_atsign (FILE *stream,
   if (aRange.length > ([self length] - aRange.location))
     [NSException raise: NSRangeException format:@"Invalid location+length."];
 
+  if (aRange.length == 0)
+    return NSOrderedSame;
   if((([self length] - aRange.location == 0) && (![aString length])))
     return NSOrderedSame;
   if(![self length])
@@ -1325,7 +1369,7 @@ if (mask & NSLiteralSearch)
   [aString getCharacters:s2];
   s2[s2len] = (unichar)0;
   end = s1len+1;
-  if (s2len > s1len)
+  if (s2len < s1len)
     end = s2len+1;
 
   if (mask & NSCaseInsensitiveSearch)
@@ -1346,7 +1390,12 @@ if (mask & NSLiteralSearch)
 	  if (s1[i] > s2[i]) return NSOrderedDescending;
 	}
     }
-  return NSOrderedSame;
+  if (s1len > s2len)
+    return NSOrderedDescending;
+  else if (s1len < s2len)
+    return NSOrderedAscending;
+  else
+    return NSOrderedSame;
 }  /* if NSLiteralSearch */
 else
 {
@@ -1362,9 +1411,9 @@ else
   while(myCount < end)
   {
     if(strCount>=[aString length])
-      return NSOrderedAscending;
-    if(myCount>=[self length])
       return NSOrderedDescending;
+    if(myCount>=[self length])
+      return NSOrderedAscending;
     myRange = [self rangeOfComposedCharacterSequenceAtIndex:  myCount];
     myCount += myRange.length;
     strRange = [aString rangeOfComposedCharacterSequenceAtIndex:  strCount];
@@ -1379,7 +1428,7 @@ else
       return result;
     } /* while */
   if(strCount<[aString length])
-    return NSOrderedDescending;
+    return NSOrderedAscending;
   return NSOrderedSame;
  }  /* else */
 }
@@ -1395,27 +1444,39 @@ else
 {
   NSRange range;
   range = [self rangeOfString:aString options:NSBackwardsSearch];
-  return (range.location == ([self length] - [aString length])) ? YES : NO;
+  return (range.length > 0 && range.location == ([self length] - [aString length])) ? YES : NO;
 }
 
 - (BOOL) isEqual: (id)anObject
 {
-  if (anObject == self)
-    return YES;
-  if ([anObject isKindOf:[NSString class]])
-    return [self isEqualToString:anObject];
-  return NO;
+    if (anObject == self) {
+	return YES;
+    }
+    if (anObject != nil) {
+	Class c = fastClassOfInstance(anObject);
+
+	if (c != nil) {
+	    if (fastClassIsKindOfClass(c, _fastCls._NSString)) {
+		return [self isEqualToString: anObject];
+	    }
+	}
+    }
+    return NO;
 }
 
 - (BOOL) isEqualToString: (NSString*)aString
 {
   id mySeq, strSeq;
   NSRange myRange, strRange;
-  unsigned int myLength = [self length];
-  unsigned int strLength = [aString length];
+  unsigned int myLength;
+  unsigned int strLength;
   unsigned int myIndex = 0;
   unsigned int strIndex = 0;
 
+  if ([self hash] != [aString hash])
+    return NO;
+  myLength = [self length];
+  strLength = [aString length];
   if((!myLength) && (!strLength))
     return YES;
   if(!myLength)
@@ -1479,13 +1540,12 @@ else
   {
     if(len > NSHashStringLength)
       len = NSHashStringLength;
-    OBJC_MALLOC(source, unichar, len*MAXDEC + 1);
+    source = alloca(sizeof(unichar)*(len*MAXDEC+1));
     [self getCharacters: source range:NSMakeRange(0,len)];
     source[len]=(unichar)0;
 
 // decompose
-
-    OBJC_MALLOC(target, unichar, len*MAXDEC+1);
+    target = alloca(sizeof(unichar)*(len*MAXDEC+1));
     spoint = source;
     tpoint = target;
     do
@@ -1507,7 +1567,6 @@ else
       tpoint = target;
       spoint = source;
     } while(notdone);
-    OBJC_FREE(target);
 
 // order
 
@@ -1544,16 +1603,34 @@ else
   } while(notdone);
 
   p = source;
+
   while (*p && char_count++ < NSHashStringLength)
     {
+#if 1
+      ret = (ret << 5) + ret + *p++;
+#else
+#if 1
+      ret = (ret << 4) + *p++;
+      ctr = ret & 0xf0000000L;
+      if (ctr != 0)
+         ret ^= ctr ^ (ctr >> 24);
+#else
       ret ^= *p++ << ctr;
-      ctr = (ctr + 1) % sizeof (void*);
+      ctr = (ctr + 4) % 20;
+#endif
+#endif
     }
-  OBJC_FREE(source);
+
+  /*
+   *	The hash caching in our concrete strin classes uses zero to denote
+   *	an empty cache value, so we MUST NOT return a hash of zero.
+   */
+  if (ret == 0)
+    ret = 0xffffffff;
   return ret;
   }
   else
-  return 0;
+    return 0xfffffffe;	/* Hash for an empty string.	*/
 }
 
 // Getting a Shared Prefix
@@ -1564,11 +1641,11 @@ else
  if(mask & NSLiteralSearch)
  {
   int prefix_len = 0;
-  unichar *s1;
-  unichar *s2;
   unichar *u,*w;
-  OBJC_MALLOC(s1, unichar,[self length] +1);
-  OBJC_MALLOC(s2, unichar,[aString length] +1);
+  unichar a1[[self length]+1];
+  unichar *s1 = a1;
+  unichar a2[[aString length]+1];
+  unichar *s2 = a2;
   u=s1;
   [self getCharacters:s1];
   s1[[self length]] = (unichar)0;
@@ -1592,7 +1669,6 @@ else
       s2++;
       prefix_len++;
     }
-    OBJC_FREE(w);
     return [NSString stringWithCharacters: u length: prefix_len];
  }
  else
@@ -1798,13 +1874,14 @@ else
 // but this will work in most cases
 - (NSString*) capitalizedString
 {
+  NSZone *z = fastZone(self);
   unichar *s;
   int count=0;
   BOOL found=YES;
   int len=[self length];
   id white = [NSCharacterSet whitespaceAndNewlineCharacterSet];
 
-  OBJC_MALLOC(s, unichar,len +1);
+  s = NSZoneMalloc(z, sizeof(unichar)*(len+1));
   [self getCharacters:s];
   s[len] = (unichar)0;
   while(count<len)
@@ -1831,36 +1908,35 @@ else
     };
     found=NO;
   };
-  s[count] = (unichar)0;
-  return [[[NSString alloc] initWithCharactersNoCopy:s length:len freeWhenDone:YES] autorelease];
+  return [[[NSString alloc] initWithCharactersNoCopy:s length:len fromZone:z] autorelease];
 }
 
 - (NSString*) lowercaseString
 {
+  NSZone *z = fastZone(self);
   unichar *s;
   int count;
   int len=[self length];
-  OBJC_MALLOC(s, unichar,len +1);
+  s = NSZoneMalloc(z, sizeof(unichar)*(len+1));
   for(count=0;count<len;count++)
     s[count]=uni_tolower([self characterAtIndex:count]);
-  s[len] = (unichar)0;
   return [[[[self class] alloc] initWithCharactersNoCopy: s
 						  length: len
-					    freeWhenDone: YES] autorelease];
+					        fromZone: z] autorelease];
 }
 
 - (NSString*) uppercaseString;
 {
+  NSZone *z = fastZone(self);
   unichar *s;
   int count;
   int len=[self length];
-  OBJC_MALLOC(s, unichar,len +1);
+  s = NSZoneMalloc(z, sizeof(unichar)*(len+1));
   for(count=0;count<len;count++)
     s[count]=uni_toupper([self characterAtIndex:count]);
-  s[len] = (unichar)0;
   return [[[[self class] alloc] initWithCharactersNoCopy: s
 						  length: len
-					    freeWhenDone: YES] autorelease];
+					        fromZone: z] autorelease];
 }
 
 // Storing the String
@@ -1908,8 +1984,12 @@ else
 {
   int len, count;
 
-  /* xxx check to make sure aRange is within self; raise NSStringBoundsError */
-  assert(NSMaxRange(aRange) <= [self cStringLength]);
+  len = [self cStringLength];
+  if (aRange.location > len)
+    [NSException raise: NSRangeException format:@"Invalid location."];
+  if (aRange.length > (len - aRange.location))
+    [NSException raise: NSRangeException format:@"Invalid location+length."];
+
   if (maxLength < aRange.length)
     {
       len = maxLength;
@@ -1942,19 +2022,26 @@ else
 
 // xxx Sould we use NSScanner here ?
 
+- (BOOL) boolValue
+{
+    if ([self caseInsensitiveCompare: @"YES"] == NSOrderedSame) 
+	return YES;
+    return [self intValue] != 0 ? YES : NO;
+}
+
 - (double) doubleValue
 {
-  return atof([self cStringNoCopy]);
+  return atof([self cString]);
 }
 
 - (float) floatValue
 {
-  return (float) atof([self cStringNoCopy]);
+  return (float) atof([self cString]);
 }
 
 - (int) intValue
 {
-  return atoi([self cStringNoCopy]);
+  return atoi([self cString]);
 }
 
 // Working With Encodings
@@ -2074,14 +2161,58 @@ else
 
 // Manipulating File System Paths
 
-
 - (unsigned int) completePathIntoString: (NSString**)outputName
-   caseSensitive: (BOOL)flag
-   matchesIntoArray: (NSArray**)outputArray
-   filterTypes: (NSArray*)filterTypes
+			  caseSensitive: (BOOL)flag
+		       matchesIntoArray: (NSArray**)outputArray
+			    filterTypes: (NSArray*)filterTypes
 {
-  [self notImplemented:_cmd];
-  return 0;
+  NSString * base_path = [self stringByDeletingLastPathComponent];
+  NSString * last_compo = [self lastPathComponent];
+  NSString * tmp_path;
+  NSDirectoryEnumerator * e;
+  NSMutableArray	*op;
+  int match_count = 0;
+
+  if (outputArray != 0)
+    op = (NSMutableArray*)[NSMutableArray array];
+
+  if (outputName != NULL) *outputName = nil;
+
+  if ([base_path length] == 0) base_path = @".";
+
+  e = [[NSFileManager defaultManager] enumeratorAtPath: base_path];
+  while (tmp_path = [e nextObject], tmp_path)
+    {
+      /* Prefix matching */
+      if (YES == flag)
+	{ /* Case sensitive */
+	  if (NO == [tmp_path hasPrefix: last_compo])
+	    continue ;
+	}
+      else
+	{
+	  if (NO == [[tmp_path uppercaseString]
+		      hasPrefix: [last_compo uppercaseString]])
+	    continue;
+	}
+
+      /* Extensions filtering */
+      if (filterTypes &&
+	  (NO == [filterTypes containsObject: [tmp_path pathExtension]]))
+	continue ;
+
+      /* Found a completion */
+      match_count++;
+      if (outputArray != NULL)
+	[*op addObject: tmp_path];
+      
+      if ((outputName != NULL) && 
+	((*outputName == nil) || (([*outputName length] < [tmp_path length]))))
+	*outputName = tmp_path;
+    }
+  if (outputArray != NULL)
+    *outputArray = [[op copy] autorelease];
+  return match_count;
 }
 
 /* Return a string for passing to OS calls to handle file system objects. */
@@ -2092,7 +2223,7 @@ else
 
 - (BOOL)getFileSystemRepresentation: (char*)buffer maxLength: (unsigned int)size
 {
-  const char* ptr = [self cStringNoCopy];
+  const char* ptr = [self cString];
   if (strlen(ptr) > size)
     return NO;
   strcpy(buffer, ptr);
@@ -2323,8 +2454,62 @@ else
 
 - (NSString*) stringByResolvingSymlinksInPath
 {
-  [self notImplemented:_cmd];
+#if defined(__WIN32__) || defined(_WIN32)
   return self;
+#else 
+  NSString *first_half = self, * second_half = @"";
+
+  const char * tmp_cpath;
+  const int MAX_PATH_LEN = 1024;
+  char tmp_buf[MAX_PATH_LEN];
+
+  int syscall_result;
+  struct stat tmp_stat;  
+
+  while(1)
+    {
+      tmp_cpath = [first_half cString];
+
+      syscall_result = lstat(tmp_cpath, &tmp_stat);
+      if (0 != syscall_result)  return self ;
+      
+      if ((tmp_stat.st_mode & S_IFLNK) &&
+	  ((syscall_result = readlink(tmp_cpath, tmp_buf, MAX_PATH_LEN)) != -1))
+	{
+	  /* 
+	   * first half is a path to a symbolic link.
+	   */
+	  tmp_buf[syscall_result] = '\0'; // Make a C string
+	  second_half 		  = [[NSString stringWithCString: tmp_buf]
+				      stringByAppendingPathComponent: second_half];
+	  first_half = [first_half stringByDeletingLastPathComponent];
+	}
+      else
+	{
+	  /* 
+	   * second_half is an absolute path 
+	   */
+	  if ([second_half hasPrefix: @"/"]) 
+	    return [second_half stringByResolvingSymlinksInPath];
+
+	  /* 
+	   * first half is NOT a path to a symbolic link 
+	   */
+	  second_half = [[first_half lastPathComponent]
+			  stringByAppendingPathComponent: second_half];
+	  first_half = [first_half stringByDeletingLastPathComponent];
+	}
+
+      /* BREAK CONDITION */
+      if ([first_half length] == 0) break;
+      else if ([first_half isEqual: @"/"])
+	{
+	  second_half = [@"/" stringByAppendingPathComponent: second_half];
+	  break;
+	}
+    }
+  return second_half;
+#endif  /* (__WIN32__) || (_WIN32) */  
 }
 
 - (NSString*) stringByStandardizingPath
@@ -2381,12 +2566,14 @@ else
 {
   #define MAXDEC 18
 
+  NSZone *z = fastZone(self);
   unichar *u, *upoint;
   NSRange r;
   id seq,ret;
   int len = [self length];
   int count = 0;
-  OBJC_MALLOC(u, unichar, len*MAXDEC+1);
+
+  u = NSZoneMalloc(z, sizeof(unichar)*(len*MAXDEC+1));
   upoint = u;
 
   while(count < len)
@@ -2401,17 +2588,16 @@ else
 
   ret = [[[[self class] alloc] initWithCharactersNoCopy: u
 						 length: uslen(u)
-					   freeWhenDone: YES] autorelease];
+					       fromZone: z] autorelease];
   return ret;
 }
 
-// #ifndef STRICT_OPENSTEP
 + (NSString*) pathWithComponents: (NSArray*)components
 {
-    NSString	*s = [self string];
+    NSString    *s = [components objectAtIndex: 0];
     int		i;
 
-    for (i = 0; i < [components count]; i++) {
+    for (i = 1; i < [components count]; i++) {
 	s = [s stringByAppendingPathComponent: [components objectAtIndex: i]];
     }
     return s;
@@ -2494,65 +2680,68 @@ else
     d = [self dataUsingEncoding: NSUnicodeStringEncoding];
   return [d writeToFile: filename atomically: useAuxiliaryFile];
 }
-// #endif
 
-// #ifndef NO_GNUSTEP
-// This method should be removed
-- (const char *) cStringNoCopy
+- (void) descriptionTo: (id<GNUDescriptionDestination>)output
 {
-  [self subclassResponsibility: _cmd];
-  return NULL;
-}
+  static NSCharacterSet	*quotables = nil;
 
-- (NSString*) descriptionForPropertyList
-{
-    static NSCharacterSet	*quotables = nil;
-
-    if (quotables == nil) {
-	quotables = [[NSCharacterSet alphanumericCharacterSet] invertedSet];
-	[quotables retain];
+  if ([self length] == 0)
+    {
+      [output appendString: @"\"\""];
+      return;
     }
 
-    if ([self length] == 0 ||
-	[self rangeOfCharacterFromSet: quotables].length > 0) {
-	NSString	*result;
-	const char	*cstring = [self cString];
-	const char	*from;
-	char		*buf;
-	char		*ptr;
-	int		len = 0;
+  if (quotables == nil)
+    {
+      quotables = [[NSCharacterSet alphanumericCharacterSet] invertedSet];
+      [quotables retain];
+    }
 
-	for (from = cstring; *from; from++) {
-	    switch (*from) {
-		case '\a':
-		case '\b':
-		case '\t':
-		case '\r':
-		case '\n':
-		case '\v':
-		case '\f':
-		case '\\':
-		case '\'' :
-		case '"' :
-		    len += 2;
-		    break;
+  if ([self rangeOfCharacterFromSet: quotables].length > 0)
+    {
+      const char	*cstring = [self cString];
+      const char	*from;
+      int		len = 0;
 
-		default:
-		    if (isprint(*from) || *from == ' ') {
-			len++;
-		    }
-		    else {
-			len += 4;
-		    }
-		    break;
+      for (from = cstring; *from; from++)
+	{
+	  switch (*from)
+	    {
+	      case '\a':
+	      case '\b':
+	      case '\t':
+	      case '\r':
+	      case '\n':
+	      case '\v':
+	      case '\f':
+	      case '\\':
+	      case '\'' :
+	      case '"' :
+		len += 2;
+		break;
+
+	      default:
+		if (isprint(*from) || *from == ' ')
+		  {
+		    len++;
+		  }
+		else
+		  {
+		    len += 4;
+		  }
+		break;
 	    }
 	}
 
-	buf = objc_malloc(len + 3);
-	ptr = buf;
+      {
+	char	buf[len+3];
+	char	*ptr = buf;
+
 	*ptr++ = '"';
-	for (from = cstring; *from; from++) {
-	    switch (*from) {
+	for (from = cstring; *from; from++)
+	  {
+	    switch (*from)
+	      {
 		case '\a':	*ptr++ = '\\'; *ptr++ = 'a';  break;
 		case '\b':	*ptr++ = '\\'; *ptr++ = 'b';  break;
 		case '\t':	*ptr++ = '\\'; *ptr++ = 't';  break;
@@ -2565,46 +2754,40 @@ else
 		case '"' :	*ptr++ = '\\'; *ptr++ = '"';  break;
 
 		default:
-		    if (isprint(*from) || *from == ' ') {
-			*ptr++ = *from;
+		  if (isprint(*from) || *from == ' ')
+		    {
+		      *ptr++ = *from;
 		    }
-		    else {
-			sprintf(ptr, "\\%03o", *(unsigned char*)from);
-			ptr = &ptr[4];
+		  else
+		    {
+		      sprintf(ptr, "\\%03o", *(unsigned char*)from);
+		      ptr = &ptr[4];
 		    }
-		    break;
-	    }
-	}
+		  break;
+	      }
+	  }
 	*ptr++ = '"';
 	*ptr = '\0';
-	result = [NSString stringWithCString: buf];
-	objc_free(buf);
-	return result;
+	[output appendString: [NSString stringWithCString: buf]];
+      }
+    }
+  else
+    {
+      [output appendString: self];
     }
-    return self;
 }
-// #endif /* NO_GNUSTEP */
 
 
 /* NSCopying Protocol */
 
 - copyWithZone: (NSZone*)zone
 {
-  return [[[self class] allocWithZone:zone] initWithString:self];
-}
-
-/* xxx Temporarily put this NSObject-like implementation here, so
-   we don't get String's Collection implementation. 
-   When we separate Core from NonCore methods, this problem will
-   go away. */
-- copy
-{
-  return [self copyWithZone: NSDefaultMallocZone ()];
-}
-
-- mutableCopy
-{
-  return [self mutableCopyWithZone: NSDefaultMallocZone ()];
+  if ([self isKindOfClass: [NSMutableString class]] ||
+	NSShouldRetainWithZone(self, zone) == NO)
+    return [[[[self class] _concreteClass] allocWithZone:zone]
+	initWithString:self];
+  else
+    return [self retain];
 }
 
 - mutableCopyWithZone: (NSZone*)zone
@@ -2617,12 +2800,35 @@ else
 
 - (void) encodeWithCoder: anEncoder
 {
-  [super encodeWithCoder:anEncoder];
+    [self subclassResponsibility:_cmd];
 }
 
 - initWithCoder: aDecoder
 {
-  return [super initWithCoder:aDecoder];
+    [self subclassResponsibility:_cmd];
+    return self;
+}
+
+- (Class) classForArchiver
+{
+  return [self class];
+}
+
+- (Class) classForCoder
+{
+  return [self class];
+}
+
+- (Class) classForPortCoder
+{
+  return [self class];
+}
+
+- (id) replacementObjectForPortCoder: (NSPortCoder*)aCoder
+{
+  if ([aCoder isByref] == NO)
+    return self;
+  return [super replacementObjectForPortCoder: aCoder];
 }
 
 @end
@@ -2637,12 +2843,6 @@ else
   return [super allocWithZone:z];
 }
 
-/* xxx This method may be removed in future. */
-- (void) setCString: (const char *)byteString length: (unsigned)length
-{
-  [self subclassResponsibility:_cmd];
-}
-
 // Creating Temporary Strings
 
 + (NSMutableString*) stringWithCapacity:(unsigned)capacity
@@ -2662,15 +2862,17 @@ else
 
 + (NSString*) stringWithCString: (const char*)byteString
 {
-  return [self stringWithCString:byteString length:strlen(byteString)];
+  return [[[NSMutableString_c_concrete_class alloc]
+		initWithCString:byteString]
+		autorelease];
 }
 
-+ (NSString*) stringWithCString: (const char*)bytes
-   length:(unsigned)length
++ (NSString*) stringWithCString: (const char*)byteString
+   length: (unsigned int)length
 {
-  id n = [[self alloc] initWithCapacity:length];
-  [n setCString:bytes length:length];
-  return n;
+  return [[[NSMutableString_c_concrete_class alloc]
+		initWithCString:byteString length:length]
+		autorelease];
 }
 
 /* xxx Change this when we have non-CString classes */
@@ -2739,63 +2941,3 @@ else
 
 @end
 
-@implementation NXConstantString
-
-/*
- *	NXConstantString overrides [-dealloc] so that it is never deallocated.
- *	If we pass an NXConstantString to another process it will never get
- *	deallocated in the other process - causing a memory leak.  So we tell
- *	the DO system to use the super class instead.
- */
-- (Class)classForPortCoder
-{
-  return [self superclass];
-}
-
-- (void)dealloc
-{
-}
-
-- (const char*) cString
-{
-  return _contents_chars;
-}
-
-- retain
-{
-  return self;
-}
-
-- (oneway void) release
-{
-  return;
-}
-
-- autorelease
-{
-  return self;
-}
-
-- copyWithZone: (NSZone*)z
-{
-  return self;
-}
-
-- (NSStringEncoding) fastestEncoding
-{
-  return NSASCIIStringEncoding;
-}
-
-- (NSStringEncoding) smallestEncoding
-{
-  return NSASCIIStringEncoding;
-}
-
-- (unichar) characterAtIndex: (unsigned int)index
-{
-  /* xxx This should raise an NSException. */
-  CHECK_INDEX_RANGE_ERROR(index, _count);
-  return (unichar)_contents_chars[index];
-}
-
-@end
