@@ -42,7 +42,7 @@
    FIXME?: use leap seconds? */
 
 #include <config.h>
-#include <gnustep/base/preface.h>
+#include <base/preface.h>
 #include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -327,14 +327,25 @@ decode (const void *ptr)
 - (void)encodeWithCoder: aCoder
 {
   [super encodeWithCoder: aCoder];
-  [aCoder encodeObject: name];
+  if (self == localTimeZone)
+    [aCoder encodeObject: @"NSLocalTimeZone"];
+  else
+    [aCoder encodeObject: name];
+}
+
+- (id) awakeAfterUsingCoder: aCoder
+{
+  if ([name isEqual: @"NSLocalTimeZone"]) {
+    return localTimeZone;
+  }
+  return [NSTimeZone timeZoneWithName: name];
 }
 
 - initWithDecoder: aDecoder
 {
-  /* FIXME?: is this right? */
   self = [super initWithCoder: aDecoder];
-  return (self = (id)[NSTimeZone timeZoneWithName: [aDecoder decodeObject]]);
+  [aDecoder decodeValueOfObjCType: @encode(id) at: &name];
+  return self;
 }
 
 - (NSTimeZoneDetail*)timeZoneDetailForDate: (NSDate*)date
@@ -414,7 +425,7 @@ decode (const void *ptr)
 - initWithCoder: aDecoder
 {
   self = [super initWithCoder: aDecoder];
-  name = [aDecoder decodeObject];
+  [aDecoder decodeValueOfObjCType: @encode(id) at: &name];
   offset = [name intValue];
   return self;
 }
@@ -468,7 +479,7 @@ decode (const void *ptr)
 - initWithCoder: aDecoder
 {
   self = [super initWithCoder: aDecoder];
-  abbrev = [aDecoder decodeObject];
+  [aDecoder decodeValueOfObjCType: @encode(id) at: &abbrev];
   [aDecoder decodeValueOfObjCType: @encode(int) at: &offset];
   [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &is_dst];
   return self;
@@ -533,7 +544,7 @@ decode (const void *ptr)
 
 	 if (f)
 	   {
-	     fp = fopen([f cStringNoCopy], "r");
+	     fp = fopen([f cString], "r");
 	     if (fp != NULL)
 	       {
                  if (fscanf(fp, "%79s", zone_name) == 1)
@@ -634,8 +645,8 @@ decode (const void *ptr)
   /* Make sure that only time zone files are accessed.
      FIXME: Make this more robust. */
   if ([aTimeZoneName length] == 0
-      || ([aTimeZoneName cStringNoCopy])[0] == '/'
-      || strchr([aTimeZoneName cStringNoCopy], '.') != NULL)
+      || ([aTimeZoneName cString])[0] == '/'
+      || strchr([aTimeZoneName cString], '.') != NULL)
     {
       NSLog(@"Disallowed time zone name `%@'.", aTimeZoneName);
       [zone_mutex unlock];
@@ -647,7 +658,7 @@ decode (const void *ptr)
 
     /* Open file. */
     fileName = [NSTimeZone getTimeZoneFile: aTimeZoneName];
-    file = fopen([fileName cStringNoCopy], "rb");
+    file = fopen([fileName cString], "rb");
     if (file == NULL)
       [NSException raise: fileException format: errMess];
 
@@ -768,7 +779,7 @@ decode (const void *ptr)
   /* Read dictionary from file. */
   abbreviationDictionary = [[NSMutableDictionary alloc] init];
   fileName = [NSTimeZone getAbbreviationFile];
-  file = fopen([fileName cStringNoCopy], "r");
+  file = fopen([fileName cString], "r");
   if (file == NULL)
     [NSException
       raise: NSInternalInconsistencyException
@@ -815,7 +826,7 @@ decode (const void *ptr)
     temp_array[i] = [[NSMutableArray alloc] init];
 
   fileName = [NSTimeZone getRegionsFile];
-  file = fopen([fileName cStringNoCopy], "r");
+  file = fopen([fileName cString], "r");
   if (file == NULL)
     [NSException
       raise: NSInternalInconsistencyException
