@@ -23,7 +23,7 @@
 */ 
 
 #include <config.h>
-#include <gnustep/base/preface.h>
+#include <base/preface.h>
 #if 0 
 /* My Linux doesn't have <libc.h>.  Why is this necessary? 
    What is a work-around that will work for all?  -mccallum*/
@@ -87,14 +87,121 @@ static NSMutableString   *processName = nil;
     convenience; other instances may also be created.
     */
 {
-  if (sharedDefaults)
+  static BOOL	beenHere = NO;	/* Flag to prevent infinite recursion */
+
+  if (beenHere)
     return sharedDefaults;
-	
+  beenHere = YES;
   // Create new sharedDefaults (NOTE: Not added to the autorelease pool!)
   sharedDefaults = [[self alloc] init];
 	
   [sharedDefaults __createStandardSearchList];
-	
+
+  if (sharedDefaults)
+    {
+      NSUserDefaults	*defs;
+      NSDictionary	*registrationDefaults;
+      NSArray		*ampm;
+      NSArray		*long_day;
+      NSArray		*long_month;
+      NSArray		*short_day;
+      NSArray		*short_month;
+      NSArray		*earlyt;
+      NSArray		*latert;
+      NSArray		*hour_names;
+      NSArray		*ymw_names;
+
+      defs = [NSUserDefaults standardUserDefaults];
+      ampm = [NSArray arrayWithObjects: @"AM", @"PM", nil];
+      short_month = [NSArray arrayWithObjects:
+						@"Jan",
+						@"Feb",
+						@"Mar",
+						@"Apr",
+						@"May",
+						@"Jun",
+						@"Jul",
+						@"Aug",
+						@"Sep",
+						@"Oct",
+						@"Nov",
+						@"Dec",
+						nil];
+      long_month = [NSArray arrayWithObjects:
+					      @"January",
+					      @"February",
+					      @"March",
+					      @"April",
+					      @"May",
+					      @"June",
+					      @"July",
+					      @"August",
+					      @"September",
+					      @"October",
+					      @"November",
+					      @"December",
+					      nil];
+      short_day = [NSArray arrayWithObjects:
+					      @"Sun",
+					      @"Mon",
+					      @"Tue",
+					      @"Wed",
+					      @"Thu",
+					      @"Fri",
+					      @"Sat",
+					      nil];
+      long_day = [NSArray arrayWithObjects:
+					    @"Sunday",
+					    @"Monday",
+					    @"Tuesday",
+					    @"Wednesday",
+					    @"Thursday",
+					    @"Friday",
+					    @"Saturday",
+					    nil];
+      earlyt = [NSArray arrayWithObjects:
+					  @"prior",
+					  @"last",
+					  @"past",
+					  @"ago",
+					  nil];
+      latert = [NSArray arrayWithObjects:
+					  @"next",
+					  nil];
+      ymw_names = [NSArray arrayWithObjects:
+					      @"year",
+					      @"month",
+					      @"week",
+					      nil];
+      hour_names = [NSArray arrayWithObjects:
+	  [NSArray arrayWithObjects: @"0", @"midnight", nil],
+	  [NSArray arrayWithObjects: @"12", @"noon", @"lunch", nil],
+	  [NSArray arrayWithObjects: @"10", @"morning", nil],
+	  [NSArray arrayWithObjects: @"14", @"afternoon", nil],
+	  [NSArray arrayWithObjects: @"19", @"dinner", nil],
+	  nil];
+      registrationDefaults = [NSDictionary dictionaryWithObjectsAndKeys:
+		ampm, NSAMPMDesignation,
+		long_month, NSMonthNameArray,
+		long_day, NSWeekDayNameArray,
+		short_month, NSShortMonthNameArray,
+		short_day, NSShortWeekDayNameArray,
+		@"DMYH", NSDateTimeOrdering,
+		@"tomorrow", NSNextDayDesignations,
+		@"nextday", NSNextNextDayDesignations,
+		@"yesterday", NSPriorDayDesignations,
+		@"today", NSThisDayDesignations,
+		earlyt, NSEarlierTimeDesignations,
+		latert, NSLaterTimeDesignations,
+		hour_names, NSHourNameDesignations,
+		ymw_names, NSYearMonthWeekDesignations,
+		nil];
+      [sharedDefaults registerDefaults: registrationDefaults];
+    }
+  else
+    {
+      NSLog(@"WARNING - unable to create shared user defaults!\n");
+    }
   return sharedDefaults;
 }
 
@@ -267,6 +374,7 @@ static NSMutableString   *processName = nil;
   [persDomains release];
   [tempDomains release];
   [changedDomains release];
+  [dictionaryRep release];
   [super dealloc];
 }
 
@@ -367,7 +475,19 @@ static NSMutableString   *processName = nil;
 	
   if (obj)
     {
-      [[persDomains objectForKey:processName] removeObjectForKey:defaultName];
+      id	obj = [persDomains objectForKey: processName];
+      NSMutableDictionary *dict;
+
+      if ([obj isKindOfClass: [NSMutableDictionary class]] == YES)
+	{
+	  dict = obj;
+	}
+      else
+	{
+	  dict = [obj mutableCopy];
+	  [persDomains setObject: dict forKey: processName];
+	}
+      [dict removeObjectForKey:defaultName];
       [self __changePersistentDomain:processName];
     }
   return;
@@ -401,8 +521,19 @@ static NSMutableString   *processName = nil;
 {
   if (value && defaultName && ([defaultName length] > 0))
     {
-      [[persDomains objectForKey:processName]
-	setObject:value forKey:defaultName];
+      id	obj = [persDomains objectForKey: processName];
+      NSMutableDictionary *dict;
+
+      if ([obj isKindOfClass: [NSMutableDictionary class]] == YES)
+	{
+	  dict = obj;
+	}
+      else
+	{
+	  dict = [obj mutableCopy];
+	  [persDomains setObject: dict forKey: processName];
+	}
+      [dict setObject:value forKey:defaultName];
       [self __changePersistentDomain:processName];
     }
   return;
@@ -444,6 +575,7 @@ static NSMutableString   *processName = nil;
 
 - (void)setSearchList:(NSArray*)newList
 {
+  DESTROY(dictionaryRep);
   [searchList release];
   searchList = [newList mutableCopy];
 }
@@ -453,7 +585,7 @@ static NSMutableString   *processName = nil;
  *************************************************************************/
 - (NSDictionary *)persistentDomainForName:(NSString *)domainName
 {
-  return [persDomains objectForKey:domainName];
+  return [[persDomains objectForKey:domainName] copy];
 }
 
 - (NSArray *)persistentDomainNames
@@ -505,6 +637,8 @@ static NSMutableString   *processName = nil;
     else
       return NO;
 	
+  DESTROY(dictionaryRep);
+
   // Read the persistent data from the stored database
   if ([[NSFileManager defaultManager] fileExistsAtPath: defaultsDatabase])
     newDict = [[NSMutableDictionary allocWithZone:[self zone]]
@@ -560,8 +694,8 @@ static NSMutableString   *processName = nil;
  *************************************************************************/
 - (void)removeVolatileDomainForName:(NSString *)domainName
 {
+  DESTROY(dictionaryRep);
   [tempDomains removeObjectForKey:domainName];
-  return;
 }
 
 - (void)setVolatileDomain:(NSDictionary *)domain 
@@ -576,6 +710,7 @@ static NSMutableString   *processName = nil;
 		   domainName];
       return;
     }
+  DESTROY(dictionaryRep);
   [tempDomains setObject:domain forKey:domainName];
   return;
 }
@@ -593,28 +728,41 @@ static NSMutableString   *processName = nil;
 /*************************************************************************
  *** Making Advanced Use of Defaults
  *************************************************************************/
-- (NSDictionary *)dictionaryRepresentation
+- (NSDictionary *) dictionaryRepresentation
 {
-  NSEnumerator *enumerator = [searchList reverseObjectEnumerator];
-  NSMutableDictionary *dictRep =
-    [NSMutableDictionary dictionaryWithCapacity:10];
-  id obj;
-  id dict;
-	
-  while ((obj = [enumerator nextObject]))
+  if (dictionaryRep == nil)
     {
-      if ( (dict = [persDomains objectForKey:obj])
-	   || (dict = [tempDomains objectForKey:obj]) )
-	[dictRep addEntriesFromDictionary:dict];
+      NSEnumerator		*enumerator;
+      NSMutableDictionary	*dictRep;
+      id obj;
+      id dict;
+	
+      enumerator = [searchList reverseObjectEnumerator];
+      dictRep = [NSMutableDictionary allocWithZone: NSDefaultMallocZone()];
+      dictRep = [dictRep initWithCapacity: 512];
+      while ((obj = [enumerator nextObject]))
+	{
+	  if ( (dict = [persDomains objectForKey: obj])
+	       || (dict = [tempDomains objectForKey: obj]) )
+	    [dictRep addEntriesFromDictionary: dict];
+	}
+      dictionaryRep = [dictRep copy];
+      [dictRep release];
     }
-  // $$$ Should we return NSDictionary here ?
-  return dictRep;
+  return dictionaryRep;
 }
 
-- (void)registerDefaults:(NSDictionary *)dictionary
+- (void) registerDefaults: (NSDictionary*)newVals
 {
-  [tempDomains setObject:dictionary forKey:NSRegistrationDomain];
-  return;
+  NSMutableDictionary	*regDefs;
+
+  regDefs = [tempDomains objectForKey: NSRegistrationDomain];
+  if (regDefs == nil)
+    {
+      regDefs = [NSMutableDictionary dictionaryWithCapacity: [newVals count]];
+    }
+  DESTROY(dictionaryRep);
+  [regDefs addEntriesFromDictionary: newVals];
 }
 
 /*************************************************************************
@@ -673,13 +821,13 @@ static NSMutableString   *processName = nil;
 	val = [enumerator nextObject];
 	if (!val)
 	  {            // No more args
-	    [argDict setObject:nil forKey:key];
+	    [argDict setObject:@"" forKey:key];		// arg is empty.
 	    done = YES;
 	    continue;
 	  }
 	else if ([val hasPrefix:@"-"])
 	  {  // Yet another argument
-	    [argDict setObject:nil forKey:key];
+	    [argDict setObject:@"" forKey:key];		// arg is empty.
 	    key = val;
 	    continue;
 	  }
@@ -698,12 +846,13 @@ static NSMutableString   *processName = nil;
 {
   NSEnumerator *enumerator = nil;
   id obj;
-	
+
+  DESTROY(dictionaryRep);
   if (!changedDomains)
     {
       changedDomains = [[NSMutableArray arrayWithCapacity:5] retain];
       [[NSNotificationCenter defaultCenter] 
-	postNotificationName:NSUserDefaultsChanged object:nil];
+	postNotificationName:NSUserDefaultsDidChangeNotification object:nil];
     }
 	
   if (!tickingTimer)
