@@ -34,15 +34,16 @@
    */
 
 #include <config.h>
-#include <gnustep/base/preface.h>
-#include <gnustep/base/TcpPort.h>
-#include <gnustep/base/Array.h>
-#include <gnustep/base/Notification.h>
-#include <gnustep/base/NSException.h>
+#include <base/preface.h>
+#include <base/TcpPort.h>
+#include <base/Array.h>
+#include <base/NotificationDispatcher.h>
+#include <base/NSException.h>
 #include <Foundation/NSRunLoop.h>
-#include <gnustep/base/Invocation.h>
+#include <base/Invocation.h>
 #include <Foundation/NSData.h>
 #include <Foundation/NSDate.h>
+#include <Foundation/NSPortNameServer.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <signal.h>
@@ -68,7 +69,6 @@
 #define NBLK_OPT     FNDELAY
 #endif
 
-#define	GDOMAP	1	/* 1 = Use name server.	*/
 #define	stringify_it(X)	#X
 #define	make_gdomap_cmd(X)	stringify_it(X) "/Tools/"GNUSTEP_TARGET_DIR"/gdomap &"
 #define	make_gdomap_err(X)	"check that " stringify_it(X) "/Tools/"GNUSTEP_TARGET_DIR"/gdomap is running and owned by root."
@@ -76,6 +76,7 @@
 #endif /* !__WIN32__ */
 #include <string.h>		/* for memset() and strchr() */
 #ifndef __WIN32__
+#include <time.h>
 #include <sys/time.h>
 #include <sys/resource.h>
 #include <sys/errno.h>
@@ -96,6 +97,17 @@ static int debug_tcp_port = 0;
 @end
 
 
+@interface NSPort (Debug)
++ (void) setDebug: (int)val;
+@end
+
+@implementation NSPort (Debug)
++ (void) setDebug: (int)val
+{
+    debug_tcp_port = val;
+}
+@end
+
 
 /* Private interfaces */
 
@@ -149,12 +161,6 @@ static int debug_tcp_port = 0;
 
 
 
-#ifdef	GDOMAP
-/*
- *	Code to contact distributed objects name server.
- */
-#include	"../Tools/gdomap.h"
-
 extern	int	errno;	/* For systems where it is not in the include	*/
 
 /*
@@ -305,15 +311,7 @@ tryWrite(int desc, int tim, unsigned char* dat, int len)
       return(-1);		/* Error in select.	*/
     }
     else if (len > 0) {
-      void	(*ifun)();
-
-      /*
-       *	Should be able to write this short a message immediately, but
-       *	if the connection is lost we will get a signal we must trap.
-       */
-      ifun = signal(SIGPIPE, (void(*)(int))SIG_IGN);
       rval = write(desc, &dat[pos], len - pos);
-      signal(SIGPIPE, ifun);
 
       if (rval <= 0) {
 	if (errno != EWOULDBLOCK) {
@@ -333,337 +331,6 @@ tryWrite(int desc, int tim, unsigned char* dat, int len)
   }
 }
 
-/*
- *	Name -		tryHost()
- *	Purpose -	Perform a name server operation with a given
- *			request packet to a server at specified address.
- *			On error - return non-zero with reason in 'errno'
- */
-static int
-tryHost(unsigned char op, unsigned char len, const unsigned char* name,
-struct sockaddr_in* addr, unsigned short* p, unsigned char **v)
-{
-    int desc = socket(AF_INET, SOCK_STREAM, 0);
-    int	e = 0;
-    unsigned long	port = *p;
-    gdo_req		msg;
-    struct sockaddr_in sin;
-
-    *p = 0;
-    if (desc < 0) {
-	return(1);	/* Couldn't create socket.	*/
-    }
-
-    if ((e = fcntl(desc, F_GETFL, 0)) >= 0) {
-	e |= NBLK_OPT;
-	if (fcntl(desc, F_SETFL, e) < 0) {
-	    e = errno;
-	    close(desc);
-	    errno = e;
-	    return(2);	/* Couldn't set non-blocking.	*/
-	}
-    }
-    else {
-	e = errno;
-	close(desc);
-	errno = e;
-	return(2);	/* Couldn't set non-blocking.	*/
-    }
-
-    memcpy(&sin, addr, sizeof(sin));
-    if (connect(desc, (struct sockaddr*)&sin, sizeof(sin)) != 0) {
-	if (errno == EINPROGRESS) {
-	    e = tryWrite(desc, 10, 0, 0);
-	    if (e == -2) {
-		e = errno;
-		close(desc);
-		errno = e;
-		return(3);	/* Connect timed out.	*/
-	    }
-	    else if (e == -1) {
-		e = errno;
-		close(desc);
-		errno = e;
-		return(3);	/* Select failed.	*/
-	    }
-	}
-	else {
-	    e = errno;
-	    close(desc);
-	    errno = e;
-	    return(3);		/* Failed connect.	*/
-	}
-    }
-
-    memset((char*)&msg, '\0', GDO_REQ_SIZE);
-    msg.rtype = op;
-    msg.nsize = len;
-    msg.ptype = GDO_TCP_GDO;
-    if (op != GDO_REGISTER) {
-	port = 0;
-    }
-    msg.port = htonl(port);
-    memcpy(msg.name, name, len);
-
-    e = tryWrite(desc, 10, (unsigned char*)&msg, GDO_REQ_SIZE);
-    if (e != GDO_REQ_SIZE) {
-	e = errno;
-	close(desc);
-	errno = e;
-	return(4);
-    }
-    e = tryRead(desc, 3, (unsigned char*)&port, 4);
-    if (e != 4) {
-	e = errno;
-	close(desc);
-	errno = e;
-	return(5);	/* Read timed out.	*/
-    }
-    port = ntohl(port);
-
-/*
- *	Special case for GDO_SERVERS - allocate buffer and read list.
- */
-    if (op == GDO_SERVERS) {
-	int	len = port * sizeof(struct in_addr);
-	unsigned char*	b;
-
-	b = (unsigned char*)objc_malloc(len);
-	if (tryRead(desc, 3, b, len) != len) {
-	    objc_free(b);
-	    e = errno;
-	    close(desc);
-	    errno = e;
-	    return(5);
-	}
-	*v = b;
-    }
-
-    *p = (unsigned short)port;
-    close(desc);
-    errno = 0;
-    return(0);
-}
-
-/*
- *	Name -		nameFail()
- *	Purpose -	If given a failure status from tryHost()
- *			raise an appropriate exception.
- */
-static void
-nameFail(int why)
-{
-    switch (why) {
-	case 0:	break;
-	case 1:
-	    [NSException raise: NSInternalInconsistencyException
-		format: @"failed to contact name server - socket - %s - %s",
-		strerror(errno),
-	        make_gdomap_err(GNUSTEP_INSTALL_PREFIX)];
-	case 2:
-	    [NSException raise: NSInternalInconsistencyException
-		format: @"failed to contact name server - socket - %s - %s",
-		strerror(errno),
-	        make_gdomap_err(GNUSTEP_INSTALL_PREFIX)];
-	case 3:
-	    [NSException raise: NSInternalInconsistencyException
-		format: @"failed to contact name server - socket - %s - %s",
-		strerror(errno),
-	        make_gdomap_err(GNUSTEP_INSTALL_PREFIX)];
-	case 4:
-	    [NSException raise: NSInternalInconsistencyException
-		format: @"failed to contact name server - socket - %s - %s",
-		strerror(errno),
-	        make_gdomap_err(GNUSTEP_INSTALL_PREFIX)];
-    }
-}
-
-/*
- *	Name -		nameServer()
- *	Purpose -	Perform name server lookup or registration.
- *			Return success/failure status and set up an
- *			address structure for use in bind or connect.
- *	Restrictions -	0xffff byte name limit
- *			Uses old style host lookup - only handles the
- *			primary network interface for each host!
- */
-static int
-nameServer(const char* name, const char* host, int op, struct sockaddr_in* addr, int pnum, int max)
-{
-    struct sockaddr_in	sin;
-    struct servent*	sp;
-    struct hostent*	hp;
-    unsigned short	p = htons(GDOMAP_PORT);
-    unsigned short	port = 0;
-    int			len = strlen(name);
-    int			multi = 0;
-    int			found = 0;
-    int			rval;
-    char local_hostname[MAXHOSTNAMELEN];
-
-    if (len == 0) {
-        [NSException raise: NSInternalInconsistencyException
-		format: @"no name specified"];
-    }
-    if (len > 255) {
-        [NSException raise: NSInternalInconsistencyException
-		format: @"name length to large (>255 characters)"];
-    }
-
-#if	GDOMAP_PORT_OVERRIDE
-    p = htons(GDOMAP_PORT_OVERRIDE);
-#else
-    /*
-     *	Ensure we have port number to connect to name server.
-     *	The TCP service name 'gdomap' overrides the default port.
-     */
-    if ((sp = getservbyname("gdomap", "tcp")) != 0) {
-	p = sp->s_port;		/* Network byte order.	*/
-    }
-#endif
-
-    /*
-     *	The host name '*' matches any host on the local network.
-     */
-    if (host && host[0] == '*' && host[1] == '\0') {
-	multi = 1;
-    }
-    /*
-     *	If no host name is given, we use the name of the local host.
-     *	NB. This should always be the case for operations other than lookup.
-     */
-    if (multi || host == 0 || *host == '\0') {
-        char *first_dot;
-
-        if (gethostname(local_hostname, sizeof(local_hostname)) < 0) {
-	    [NSException raise: NSInternalInconsistencyException
-		format: @"gethostname() failed: %s", strerror(errno)];
-	}
-        first_dot = strchr(local_hostname, '.');
-        if (first_dot) {
-	    *first_dot = '\0';
-	}
-	host = local_hostname;
-    }
-    if ((hp = gethostbyname(host)) == 0) {
-	[NSException raise: NSInternalInconsistencyException
-		format: @"get host address for %s", host];
-    }
-    if (hp->h_addrtype != AF_INET) {
-	[NSException raise: NSInternalInconsistencyException
-		format: @"non-internet network not supported for %s", host];
-    }
-
-    memset((char*)&sin, '\0', sizeof(sin));
-    sin.sin_family = AF_INET;
-    sin.sin_port = p;
-    memcpy((caddr_t)&sin.sin_addr, hp->h_addr, hp->h_length);
-
-    if (multi) {
-	unsigned short	num;
-	struct in_addr*	b;
-
-	/*
-	 *	A host name of '*' is a special case which should do lookup on
-	 *	all machines on the local network until one is found which has 
-	 *	the specified server on it.
-	 */
-	rval = tryHost(GDO_SERVERS, 0, 0, &sin, &num, (unsigned char**)&b);
-	/*
-	 *	If the connection to the local name server fails,
-	 *	attempt to start it us and retry the lookup.
-	 */
-	if (rval != 0 && host == local_hostname) {
-	    system(make_gdomap_cmd(GNUSTEP_INSTALL_PREFIX));
-	    sleep(5);
-	    rval = tryHost(GDO_SERVERS, 0, 0, &sin, &num, (unsigned char**)&b);
-	}
-	if (rval == 0) {
-	    int	i;
-
-	    for (i = 0; found == 0 && i < num; i++) {
-		memset((char*)&sin, '\0', sizeof(sin));
-		sin.sin_family = AF_INET;
-		sin.sin_port = p;
-		memcpy((caddr_t)&sin.sin_addr, &b[i], sizeof(struct in_addr));
-		if (sin.sin_addr.s_addr == 0) continue;
-
-		if (tryHost(GDO_LOOKUP, len, name, &sin, &port, 0) == 0) {
-		    if (port != 0) {
-			memset((char*)&addr[found], '\0', sizeof(*addr));
-			memcpy((caddr_t)&addr[found].sin_addr, &sin.sin_addr,
-				sizeof(sin.sin_addr));
-			addr[found].sin_family = AF_INET;
-			addr[found].sin_port = htons(port);
-			found++;
-			if (found == max) {
-			    break;
-			}
-		    }
-		}
-	    }
-	    objc_free(b);
-	    return(found);
-	}
-	else {
-	    nameFail(rval);
-	}
-    }
-    else {
-        if (op == GDO_REGISTER) {
-	    port = (unsigned short)pnum;
-	}
-	rval = tryHost(op, len, name, &sin, &port, 0);
-	/*
-	 *	If the connection to the local name server fails,
-	 *	attempt to start it us and retry the lookup.
-	 */
-	if (rval != 0 && host == local_hostname) {
-	    system(make_gdomap_cmd(GNUSTEP_INSTALL_PREFIX));
-	    sleep(5);
-            if (op == GDO_REGISTER) {
-	        port = (unsigned short)pnum;
-	    }
-	    rval = tryHost(op, len, name, &sin, &port, 0);
-	}
-	nameFail(rval);
-    }
-
-    if (op == GDO_REGISTER) {
-	if (port == 0 || (pnum != 0 && port != pnum)) {
-	    [NSException raise: NSInternalInconsistencyException
-		format: @"service already registered"];
-	}
-    }
-    if (port == 0) {
-	return 0;
-    }
-    memset((char*)addr, '\0', sizeof(*addr));
-    memcpy((caddr_t)&addr->sin_addr, &sin.sin_addr, sizeof(sin.sin_addr));
-    addr->sin_family = AF_INET;
-    addr->sin_port = htons(port);
-    return 1;
-}
-
-#else
-/* The old hash code for a name server. */
-
-static unsigned short
-name_2_port_number (const char *name)
-{
-  unsigned int ret = 0;
-  unsigned int ctr = 0;
-        
-  while (*name) 
-    {
-      ret ^= *name++ << ctr;
-      ctr = (ctr + 1) % sizeof (void *);
-    }
-  return (ret % (65535 - IPPORT_USERRESERVED - 1)) + IPPORT_USERRESERVED;
-  /* return strlen (name) + IPPORT_USERRESERVED; */
-}
-#endif	/* GDOMAP */
-
 
 /* Both TcpInPort's and TcpOutPort's are entered in this maptable. */
 
@@ -694,10 +361,17 @@ static NSMapTable* port_number_2_port;
 + (void) initialize
 {
   if (self == [TcpInPort class])
-    port_number_2_port = 
-      NSCreateMapTable (NSIntMapKeyCallBacks,
-			NSNonOwnedPointerMapValueCallBacks, 0);
-  init_port_socket_2_port ();
+    {
+      port_number_2_port = 
+        NSCreateMapTable (NSIntMapKeyCallBacks,
+			  NSNonOwnedPointerMapValueCallBacks, 0);
+      init_port_socket_2_port ();
+      /*
+       *	If SIGPIPE is not ignored, we will abort on any attempt to
+       *	write to a pipe/socket that has been closed by the other end!
+       */
+      signal(SIGPIPE, SIG_IGN);
+    }
 }
 
 /* This is the designated initializer. 
@@ -711,7 +385,7 @@ static NSMapTable* port_number_2_port;
      don't create a new one, just return the old one. */
   if ((p = (id) NSMapGet (port_number_2_port, (void*)((int)n))))
     {
-      assert (p->is_valid);
+      NSAssert(p->is_valid, NSInternalInconsistencyException);
       return [p retain];
     }
 
@@ -732,7 +406,7 @@ static NSMapTable* port_number_2_port;
     }
 
   /* Register the port object according to its socket. */
-  assert (!NSMapGet (socket_2_port, (void*)p->_port_socket));
+  NSAssert(!NSMapGet (socket_2_port, (void*)p->_port_socket), NSInternalInconsistencyException);
   NSMapInsert (socket_2_port, (void*)p->_port_socket, p);
   
   /* Give the socket a name using bind() and INADDR_ANY for the
@@ -809,7 +483,7 @@ static NSMapTable* port_number_2_port;
 	      strerror(errno)];
 
 	  }
-	assert (p->_listening_address.sin_port);
+	NSAssert(p->_listening_address.sin_port, NSInternalInconsistencyException);
 	n = ntohs(p->_listening_address.sin_port);
       }
 
@@ -833,7 +507,7 @@ static NSMapTable* port_number_2_port;
     hp = gethostbyname (hostname);
     if (!hp)
       [self error: "Could not get address of local host \"%s\"", hostname];
-    assert (hp);
+    NSAssert(hp, NSInternalInconsistencyException);
     memcpy (&(p->_listening_address.sin_addr), hp->h_addr, hp->h_length);
   }
 
@@ -863,23 +537,20 @@ static NSMapTable* port_number_2_port;
 
 + newForReceivingFromRegisteredName: (NSString*)name
 {
-#ifdef	GDOMAP
-  TcpInPort*		p = [self newForReceivingFromPortNumber: 0];
+  return [self newForReceivingFromRegisteredName: name fromPort: 0];
+}
+
++ newForReceivingFromRegisteredName: (NSString*)name
+			   fromPort: (int)portn
+{
+  TcpInPort*		p = [self newForReceivingFromPortNumber: portn];
   struct sockaddr_in	sin;
 
   if (p) {
-    int	port = [p portNumber];
-
-    if (nameServer([name cString], 0, GDO_REGISTER, &sin, port, 1) == 0) {
-      [p release];
-      return nil;
-    }
+    [[NSPortNameServer defaultPortNameServer] registerPort: p
+						   forName: name];
   }
   return p;
-#else
-  return [self newForReceivingFromPortNumber: 
-		 name_2_port_number ([name cString])];
-#endif	/* GDOMAP */
 }
 
 + newForReceiving
@@ -910,7 +581,7 @@ static NSMapTable* port_number_2_port;
 
 - (struct sockaddr_in*) _listeningSockaddr
 {
-  assert (is_valid);
+  NSAssert(is_valid, NSInternalInconsistencyException);
   return &_listening_address;
 }
 
@@ -997,10 +668,8 @@ static NSMapTable* port_number_2_port;
       [self _addClientOutPort: op];
       [op release];
       if (debug_tcp_port)
-	fprintf (stderr, 
-		 "%s: Accepted connection from\n %s.\n",
-		 object_get_class_name (self),
-		 [[op description] cString]);
+	NSLog(@"%s: Accepted connection from\n %@.\n",
+		 object_get_class_name (self), [op description]);
       [NotificationDispatcher
 	postNotificationName: InPortAcceptedClientNotification
 	object: self
@@ -1072,8 +741,11 @@ static NSMapTable* port_number_2_port;
 	{
 	  /* No bytes are remaining to be read for this packet; 
 	     the packet is complete; return it. */
-	  assert (packet && [packet class]);
+	  NSAssert(packet && [packet class], NSInternalInconsistencyException);
           NSMapRemove(_client_sock_2_packet, (void*)fd_index);
+	  if (debug_tcp_port > 1)
+	    NSLog(@"%s: Read from socket %d\n",
+		object_get_class_name (self), fd_index);
 	  return packet;
 	}
     }
@@ -1095,7 +767,7 @@ static NSMapTable* port_number_2_port;
   id out_port;
 
   /* Make sure there is enough room in the provided array. */
-  assert (*count > NSCountMapTable (_client_sock_2_out_port));
+  NSAssert(*count > NSCountMapTable (_client_sock_2_out_port), NSInternalInconsistencyException);
 
   /* Put in our listening socket. */
   *count = 0;
@@ -1119,7 +791,7 @@ static NSMapTable* port_number_2_port;
   id arp = [NSAutoreleasePool new];
   id packet;
 
-assert(type == ET_RPORT);
+  NSAssert(type == ET_RPORT, NSInvalidArgumentException);
 
   packet = [self _tryToGetPacketFromReadableFD: (int)extra];
   if (packet) {
@@ -1145,9 +817,9 @@ assert(type == ET_RPORT);
 {
   int s = [p _port_socket];
 
-  assert (is_valid);
+  NSAssert(is_valid, NSInternalInconsistencyException);
   /* Make sure it hasn't already been added. */
-  assert (!NSMapGet (_client_sock_2_out_port, (void*)s));
+  NSAssert(!NSMapGet (_client_sock_2_out_port, (void*)s), NSInternalInconsistencyException);
 
  /* Add it, and put its socket in the set of file descriptors we poll. */
   NSMapInsert (_client_sock_2_out_port, (void*)s, p);
@@ -1159,12 +831,10 @@ assert(type == ET_RPORT);
   id packet;
   int s = [p _port_socket];
 
-  assert (is_valid);
+  NSAssert(is_valid, NSInternalInconsistencyException);
   if (debug_tcp_port)
-    fprintf (stderr, 
-	     "%s: Closed connection from\n %s\n",
-	     object_get_class_name (self),
-	     [[p description] cString]);
+    NSLog(@"%s: Closed connection from\n %@\n",
+	     object_get_class_name (self), [p description]);
 
   packet = NSMapGet (_client_sock_2_packet, (void*)s);
   if (packet)
@@ -1223,7 +893,7 @@ assert(type == ET_RPORT);
 	  /* This will call [self _invalidateConnectedOutPort: for each. */
 	  [out_ports[i] invalidate];
 	}
-      assert (!NSCountMapTable (_client_sock_2_out_port));
+      NSAssert(!NSCountMapTable (_client_sock_2_out_port), NSInternalInconsistencyException);
 
       /* xxx Perhaps should delay this close() to keep another port from
 	 getting it.  This may help Connection invalidation confusion. 
@@ -1246,7 +916,6 @@ assert(type == ET_RPORT);
 - (void) dealloc
 {
   [self invalidate];
-  /* assert that these are empty? */
   NSFreeMapTable (_client_sock_2_out_port);
   NSFreeMapTable (_client_sock_2_packet);
   [super dealloc];
@@ -1296,11 +965,10 @@ assert(type == ET_RPORT);
 
 - (void) encodeWithCoder: aCoder
 {
-  assert (is_valid);
+  NSAssert(is_valid, NSInternalInconsistencyException);
   /* We are actually encoding a "send right" (ala Mach), 
      not a receive right.  
      These values must match those expected by [TcpOutPort +newWithCoder] */
-  [super encodeWithCoder: aCoder];
   /* Encode these at bytes, not as C-variables, because they are
      already in "network byte-order". */
   [aCoder encodeBytes: &_listening_address.sin_port
@@ -1340,6 +1008,11 @@ static NSMapTable *out_port_bag = NULL;
       init_port_socket_2_port ();
       out_port_bag = NSCreateMapTable (NSNonOwnedPointerMapKeyCallBacks,
 				       NSNonOwnedPointerMapValueCallBacks, 0);
+      /*
+       *	If SIGPIPE is not ignored, we will abort on any attempt to
+       *	write to a pipe/socket that has been closed by the other end!
+       */
+      signal(SIGPIPE, SIG_IGN);
     }
 }
 
@@ -1382,7 +1055,7 @@ static NSMapTable *out_port_bag = NULL;
       NSMapEnumerator me = NSEnumerateMapTable (out_port_bag);
       void *k;
 
-      assert (sockaddr);
+      NSAssert(sockaddr, NSInternalInconsistencyException);
       while (NSNextMapEnumeratorPair (&me, &k, (void**)&p))
 	{
 	  /* xxx Do I need to make sure connectedInPort is the same too? */
@@ -1394,7 +1067,7 @@ static NSMapTable *out_port_bag = NULL;
 	    /* Assume that sin_family is equal.  Using memcmp() doesn't
 	       work because sin_zero's may differ. */
 	    {
-	      assert (p->is_valid);
+	      NSAssert(p->is_valid, NSInternalInconsistencyException);
 	      return [p retain];
 	    }
 	}
@@ -1409,11 +1082,11 @@ static NSMapTable *out_port_bag = NULL;
      ivar. */
   if (sock && (p = NSMapGet (socket_2_port, (void*)sock)))
     {
-      assert ([p isKindOfClass: [TcpOutPort class]]);
+      NSAssert([p isKindOfClass: [TcpOutPort class]], NSInternalInconsistencyException);
       if (sockaddr)
 	{
 	  /* Make sure the address we're setting it to is non-zero. */
-	  assert (sockaddr->sin_port);
+	  NSAssert(sockaddr->sin_port, NSInternalInconsistencyException);
 
 	  /* See if the _remote_in_port_address is already set */
 	  if (p->_remote_in_port_address.sin_family)
@@ -1439,8 +1112,10 @@ static NSMapTable *out_port_bag = NULL;
 		          sockaddr,
 		          sizeof (p->_remote_in_port_address));
 	          NSMapInsert (out_port_bag, (void*)p, (void*)p);
+/*
 	          NSLog(@"Out port changed from %@ to %@\n", od,
 			    [p description]);
+*/
 		}
 #endif
 	    }
@@ -1451,13 +1126,13 @@ static NSMapTable *out_port_bag = NULL;
 		      sockaddr,
 		      sizeof (p->_remote_in_port_address));
 	      if (debug_tcp_port)
-		printf ("TcpOutPort setting remote address\n%s\n",
-			[[self description] cString]);
+		NSLog(@"TcpOutPort setting remote address\n%@\n",
+			[self description]);
 	    }
 	}
       if (p)
 	{
-	  assert (p->is_valid);
+	  NSAssert(p->is_valid, NSInternalInconsistencyException);
 	  return [p retain];
 	}
     }
@@ -1488,7 +1163,7 @@ static NSMapTable *out_port_bag = NULL;
   /* Set the port's address. */
   if (sockaddr)
     {
-      assert (sockaddr->sin_port);
+      NSAssert(sockaddr->sin_port, NSInternalInconsistencyException);
       memcpy (&(p->_remote_in_port_address), sockaddr, sizeof(*sockaddr));
     }
   else
@@ -1507,7 +1182,7 @@ static NSMapTable *out_port_bag = NULL;
   if (!sock) {
       int	rval;
 
-      assert (p->_remote_in_port_address.sin_family);
+      NSAssert(p->_remote_in_port_address.sin_family, NSInternalInconsistencyException);
 
       if (connect (p->_port_socket,
 		   (struct sockaddr*)&(p->_remote_in_port_address), 
@@ -1541,7 +1216,7 @@ static NSMapTable *out_port_bag = NULL;
     }
 
   /* Put it in the shared socket->port map table. */
-  assert (!NSMapGet (socket_2_port, (void*)p->_port_socket));
+  NSAssert(!NSMapGet (socket_2_port, (void*)p->_port_socket), NSInternalInconsistencyException);
   NSMapInsert (socket_2_port, (void*)p->_port_socket, p);
 
   /* Put it in TcpOutPort's registry. */
@@ -1594,25 +1269,11 @@ static NSMapTable *out_port_bag = NULL;
 + newForSendingToRegisteredName: (NSString*)name 
 			 onHost: (NSString*)hostname
 {
-#ifdef	GDOMAP
-  struct sockaddr_in	sin[100];
-  int			found;
-  int			i;
-  id			c = nil;
+  id	c;
 
-  found = nameServer([name cString], [hostname cString],
-	GDO_LOOKUP, sin, 0, 100);
-  for (i = 0; c == nil && i < found; i++)
-    {
-      c = [self newForSendingToSockaddr: &sin[i]
-		     withAcceptedSocket: 0
-		     pollingInPort: nil];
-    }
-  return c;
-#else
-  return [self newForSendingToPortNumber: 
-		 name_2_port_number ([name cString]) onHost: hostname];;
-#endif	/* GDOMAP */
+  c = [[NSPortNameServer defaultPortNameServer] portForName: name
+						     onHost: hostname];
+  return [c retain];
 }
 
 + _newWithAcceptedSocket: (int)s 
@@ -1630,9 +1291,9 @@ static NSMapTable *out_port_bag = NULL;
 	  format: @"[TcpInPort newWithAcceptedSocket:] getsockname(): %s",
 	  strerror(errno)];
     }
-  assert (size == sizeof (struct sockaddr_in));
+  NSAssert(size == sizeof (struct sockaddr_in), NSInternalInconsistencyException);
   /* xxx Perhaps I have to get peer name here!! */
-  assert (ntohs (addr.sin_port) != [p portNumber]);
+  NSAssert(ntohs (addr.sin_port) != [p portNumber], NSInternalInconsistencyException);
 #elif 0
   struct sockaddr_in in_port_address;
   c = read (s, &in_port_address, sizeof(struct sockaddr_in));
@@ -1652,7 +1313,7 @@ static NSMapTable *out_port_bag = NULL;
 {
   id reply_port = [packet replyInPort];
 
-  assert (is_valid);
+  NSAssert(is_valid, NSInternalInconsistencyException);
 
   /* If the socket of this TcpOutPort isn't already being polled
      for incoming data by a TcpInPort, and if the packet's REPLY_PORT
@@ -1787,11 +1448,10 @@ static NSMapTable *out_port_bag = NULL;
 
 - (void) encodeWithCoder: aCoder
 {
-  assert (is_valid);
-  [super encodeWithCoder: aCoder];
-  assert (!_polling_in_port 
+  NSAssert(is_valid, NSInternalInconsistencyException);
+  NSAssert(!_polling_in_port 
 	  || (ntohs (_remote_in_port_address.sin_port)
-	      != [_polling_in_port portNumber]));
+      != [_polling_in_port portNumber]), NSInternalInconsistencyException);
   /* Encode these at bytes, not as C-variables, because they are
      already in "network byte-order". */
   [aCoder encodeBytes: &_remote_in_port_address.sin_port
@@ -1801,7 +1461,7 @@ static NSMapTable *out_port_bag = NULL;
 	  count: sizeof (_remote_in_port_address.sin_addr.s_addr)
 	  withName: @"inet address"];
   if (debug_tcp_port)
-    printf ("TcpOutPort encoded port %hd host %s\n",
+    NSLog(@"TcpOutPort encoded port %hd host %s\n",
 	    ntohs (_remote_in_port_address.sin_port),
 	    inet_ntoa (_remote_in_port_address.sin_addr));
 }
@@ -1818,7 +1478,7 @@ static NSMapTable *out_port_bag = NULL;
 	  count: sizeof (addr.sin_addr.s_addr)
 	  withName: NULL];
   if (debug_tcp_port)
-    printf ("TcpOutPort decoded port %hd host %s\n",
+    NSLog(@"TcpOutPort decoded port %hd host %s\n",
 	    ntohs (addr.sin_port),
 	    inet_ntoa (addr.sin_addr));
   return [TcpOutPort newForSendingToSockaddr: &addr
@@ -1849,15 +1509,7 @@ static NSMapTable *out_port_bag = NULL;
   char prefix_buffer[PREFIX_SIZE];
   int c;
   
-#ifdef	GDOMAP
   c = tryRead (s, 3, prefix_buffer, PREFIX_SIZE);
-#else
-#ifdef	__WIN32__
-  c = recv (s, prefix_buffer, PREFIX_SIZE, 0);
-#else
-  c = read (s, prefix_buffer, PREFIX_SIZE);
-#endif	/* __WIN32__ */
-#endif	/* GDOMAP */
   if (c <= 0)
     {
       *packet_size = EOF;  *rp = nil;
@@ -1877,7 +1529,7 @@ static NSMapTable *out_port_bag = NULL;
   /* *size is the number of bytes in the packet, not including 
      the PREFIX_SIZE-byte header. */
   *packet_size = ntohl (*(PREFIX_LENGTH_TYPE*) prefix_buffer);
-  assert (packet_size);
+  NSAssert(packet_size, NSInternalInconsistencyException);
 
   /* If the reply address is non-zero, and the TcpOutPort for this socket
      doesn't already have its _address ivar set, then set it now. */
@@ -1904,16 +1556,7 @@ static NSMapTable *out_port_bag = NULL;
   int remaining;
 
   remaining = [data length] - prefix - eof_position;
-#ifdef	GDOMAP
   c = tryRead(s, 1, [data mutableBytes] + prefix + eof_position, -remaining);
-#else
-  /* xxx We need to make sure this read() is non-blocking. */
-#ifdef	__WIN32__
-  c = recv (s, [data mutableBytes] + prefix + eof_position, remaining, 0);
-#else
-  c = read (s, [data mutableBytes] + prefix + eof_position, remaining);
-#endif	/* __WIN32 */
-#endif	/* GDOMAP */
   if (c <= 0) {
     return EOF;
   }
@@ -1936,8 +1579,11 @@ static NSMapTable *out_port_bag = NULL;
 {
   int c;
 
+  if (debug_tcp_port > 1)
+    NSLog(@"%s: Write to socket %d\n", object_get_class_name (self), s);
+
   /* Put the packet size in the first four bytes of the packet. */
-  assert (prefix == PREFIX_SIZE);
+  NSAssert(prefix == PREFIX_SIZE, NSInternalInconsistencyException);
   *(PREFIX_LENGTH_TYPE*)[data mutableBytes] = htonl (eof_position);
 
   /* Put the sockaddr_in for replies in the next bytes of the prefix
@@ -1950,15 +1596,7 @@ static NSMapTable *out_port_bag = NULL;
     memset ([data mutableBytes]+PREFIX_LENGTH_SIZE, 0, PREFIX_ADDRESS_SIZE);
 
   /* Write the packet on the socket. */
-#ifdef	GDOMAP
-  c = tryWrite (s, (int)timeout, [data bytes], prefix + eof_position);
-#else
-#ifdef	__WIN32__
-  c = send (s, [data bytes], prefix + eof_position, 0);
-#else
-  c = write (s, [data bytes], prefix + eof_position);
-#endif	/* __WIN32__ */
-#endif	/* GDOMAP */
+  c = tryWrite (s, (int)timeout, (unsigned char*)[data bytes], prefix + eof_position);
   if (c == -2) {
     [NSException raise: NSPortTimeoutException
 	format: @"[TcpOutPort -_writeToSocket:] write() timed out"];
