@@ -22,7 +22,7 @@
    */
 
 #include <config.h>
-#include <gnustep/base/preface.h>
+#include <base/preface.h>
 #include <Foundation/NSObject.h>
 #include <Foundation/NSData.h>
 #include <Foundation/NSArray.h>
@@ -38,14 +38,21 @@
 #ifdef WIN32
 #include <Windows32/Sockets.h>
 #else
+#include <time.h>
+#include <sys/time.h>
 #include <sys/param.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
+#include <signal.h>
 #endif /* WIN32 */
 
 #include <sys/file.h>
 #include <sys/stat.h>
 #include <sys/fcntl.h>
+#include <sys/ioctl.h>
+#ifdef	__svr4__
+#include <sys/filio.h>
+#endif
 #include <netdb.h>
 #include <string.h>
 #include <unistd.h>
@@ -83,7 +90,7 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
   struct servent*	sp;
 
   if (pcl)
-    proto = [pcl cStringNoCopy];
+    proto = [pcl cString];
 
   memset(sin, '\0', sizeof(*sin));
   sin->sin_family = AF_INET;
@@ -95,15 +102,15 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
    */
   if (name)
     {
-      NSHost*		host = [NSHost hostWithName:name];
+      NSHost*		host = [NSHost hostWithName: name];
 
       if (host)
 	name = [host address];
 
 #ifndef	HAVE_INET_ATON
-      sin->sin_addr.s_addr = inet_addr([name cStringNoCopy]);
+      sin->sin_addr.s_addr = inet_addr([name cString]);
 #else
-      if (inet_aton([name cStringNoCopy], &sin->sin_addr) == 0)
+      if (inet_aton([name cString], &sin->sin_addr) == 0)
 	return NO;
 #endif
     }
@@ -115,9 +122,9 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
       sin->sin_port = 0;
       return YES;
     }
-  else if ((sp = getservbyname([svc cStringNoCopy], proto)) == 0)
+  else if ((sp = getservbyname([svc cString], proto)) == 0)
     {
-      const char*     ptr = [svc cStringNoCopy];
+      const char*     ptr = [svc cString];
       int             val = atoi(ptr);
 
       while (isdigit(*ptr))
@@ -140,12 +147,24 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
     }
 }
 
-+ allocWithZone:(NSZone*)z
++ (void) initialize
+{
+  if (self == [UnixFileHandle class])
+    {
+      /*
+       *	If SIGPIPE is not ignored, we will abort on any attempt to
+       *	write to a pipe/socket that has been closed by the other end!
+       */
+      signal(SIGPIPE, SIG_IGN);
+    }
+}
+
++ (id) allocWithZone: (NSZone*)z
 {
   return NSAllocateObject ([self class], 0, z);
 }
 
-- (void)dealloc
+- (void) dealloc
 {
   [address release];
   [service release];
@@ -169,7 +188,7 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 	  descriptor = -1;
 	}
       else if (isNonBlocking != wasNonBlocking)
-	[self setNonBlocking:wasNonBlocking];
+	[self setNonBlocking: wasNonBlocking];
     }
   [readInfo release];
   [writeInfo release];
@@ -180,71 +199,118 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 
 - (id)init
 {
-    return [self initWithNullDevice];
+  return [self initWithNullDevice];
 }
 
-- (id)initAsClientAtAddress:a
-		    service:s
-		   protocol:p
-		   forModes:modes
+- (id)initAsClientAtAddress: a
+		    service: s
+		   protocol: p
 {
   int	net;
   struct sockaddr_in	sin;
 
   if (s == nil)
     {
-      [self dealloc];
       NSLog(@"bad argument - service is nil");
+      [self release];
       return nil;
     }
   if (getAddr(a, s, p, &sin) == NO)
     {
-      [self dealloc];
       NSLog(@"bad address-service-protocol combination");
+      [self release];
       return nil;
     }
+  [self setAddr: &sin];
 
   if ((net = socket(AF_INET, SOCK_STREAM, PF_UNSPEC)) < 0)
     {
-      [self dealloc];
       NSLog(@"unable to create socket - %s", strerror(errno));
+      [self release];
       return nil;
     }
 
-  self = [self initWithFileDescriptor:net closeOnDealloc:YES];
+  self = [self initWithFileDescriptor: net closeOnDealloc: YES];
+  if (self)
+    {
+      if (connect(net, (struct sockaddr*)&sin, sizeof(sin)) < 0)
+	{
+	  NSLog(@"unable to make connection to %s:%d - %s",
+		inet_ntoa(sin.sin_addr), ntohs(sin.sin_port), strerror(errno));
+	  [self release];
+	  return nil;
+	}
+	
+      connectOK = NO;
+      readOK = YES;
+      writeOK = YES;
+    }
+  return self;
+}
+
+- (id)initAsClientInBackgroundAtAddress: a
+				service: s
+			       protocol: p
+			       forModes: modes
+{
+  int	net;
+  struct sockaddr_in	sin;
+
+  if (s == nil)
+    {
+      NSLog(@"bad argument - service is nil");
+      [self release];
+      return nil;
+    }
+  if (getAddr(a, s, p, &sin) == NO)
+    {
+      NSLog(@"bad address-service-protocol combination");
+      [self release];
+      return nil;
+    }
+  [self setAddr: &sin];
+
+  if ((net = socket(AF_INET, SOCK_STREAM, PF_UNSPEC)) < 0)
+    {
+      NSLog(@"unable to create socket - %s", strerror(errno));
+      [self release];
+      return nil;
+    }
+
+  self = [self initWithFileDescriptor: net closeOnDealloc: YES];
   if (self)
     {
       NSMutableDictionary*	info;
 
-      [self setNonBlocking:YES];
+      [self setNonBlocking: YES];
       if (connect(net, (struct sockaddr*)&sin, sizeof(sin)) < 0)
 	if (errno != EINPROGRESS)
 	  {
-	    [self dealloc];
-	    NSLog(@"unable to make connection - %s", strerror(errno));
+	    NSLog(@"unable to make connection to %s:%d - %s",
+		inet_ntoa(sin.sin_addr), ntohs(sin.sin_port), strerror(errno));
+	    [self release];
 	    return nil;
 	  }
 	
-      info = [[NSMutableDictionary dictionaryWithCapacity:4] retain];
-      [info setObject:address forKey:NSFileHandleNotificationDataItem];
-      [info setObject:GSFileHandleConnectCompletionNotification
-	        forKey:NotificationKey];
+      info = [[NSMutableDictionary dictionaryWithCapacity: 4] retain];
+      [info setObject: address forKey: NSFileHandleNotificationDataItem];
+      [info setObject: GSFileHandleConnectCompletionNotification
+	       forKey: NotificationKey];
       if (modes)
-        [info setObject:modes forKey:NSFileHandleNotificationMonitorModes];
-      [writeInfo addObject:info];
+        [info setObject: modes forKey: NSFileHandleNotificationMonitorModes];
+      [writeInfo addObject: info];
       [info release];
       [self watchWriteDescriptor];
       connectOK = YES;
       readOK = NO;
       writeOK = NO;
-      [self setAddr: &sin];
     }
   return self;
 }
 
-- (id)initAsServerAtAddress:a
-		    service:s
-		   protocol:p
+- (id)initAsServerAtAddress: a
+		    service: s
+		   protocol: p
 {
   int	status = 1;
   int	net;
@@ -253,15 +319,15 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 
   if (getAddr(a, s, p, &sin) == NO)
     {
-      [self dealloc];
+      [self release];
       NSLog(@"bad address-service-protocol combination");
       return  nil;
     }
 
   if ((net = socket(AF_INET, SOCK_STREAM, PF_UNSPEC)) < 0)
     {
-      [self dealloc];
       NSLog(@"unable to create socket - %s", strerror(errno));
+      [self release];
       return nil;
     }
 
@@ -269,29 +335,30 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 
   if (bind(net, (struct sockaddr *)&sin, sizeof(sin)) < 0)
     {
+      NSLog(@"unable to bind to port %s:%d - %s",
+		inet_ntoa(sin.sin_addr), ntohs(sin.sin_port), strerror(errno));
       (void) close(net);
-      [self dealloc];
-      NSLog(@"unable to bind to port - %s", strerror(errno));
+      [self release];
       return nil;
     }
 
   if (listen(net, 5) < 0)
     {
-      (void) close(net);
-      [self dealloc];
       NSLog(@"unable to listen on port - %s", strerror(errno));
+      (void) close(net);
+      [self release];
       return nil;
     }
 
   if (getsockname(net, (struct sockaddr*)&sin, &size) < 0)
     {
-      (void) close(net);
-      [self dealloc];
       NSLog(@"unable to get socket name - %s", strerror(errno));
+      (void) close(net);
+      [self release];
       return nil;
     }
 
-  self = [self initWithFileDescriptor:net closeOnDealloc:YES];
+  self = [self initWithFileDescriptor: net closeOnDealloc: YES];
   if (self)
     {
       acceptOK = YES;
@@ -302,54 +369,54 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
   return self;
 }
 
-- (id)initForReadingAtPath:(NSString*)path
+- (id)initForReadingAtPath: (NSString*)path
 {
   int	d = open([path fileSystemRepresentation], O_RDONLY);
 
   if (d < 0)
     {
-      [self dealloc];
+      [self release];
       return nil;
     }
   else
     {
-      self = [self initWithFileDescriptor:d closeOnDealloc:YES];
+      self = [self initWithFileDescriptor: d closeOnDealloc: YES];
       if (self)
 	writeOK = NO;
       return self;
     }
 }
 
-- (id)initForWritingAtPath:(NSString*)path
+- (id)initForWritingAtPath: (NSString*)path
 {
   int	d = open([path fileSystemRepresentation], O_WRONLY);
 
   if (d < 0)
     {
-      [self dealloc];
+      [self release];
       return nil;
     }
   else
     {
-      self = [self initWithFileDescriptor:d closeOnDealloc:YES];
+      self = [self initWithFileDescriptor: d closeOnDealloc: YES];
       if (self)
         readOK = NO;
       return self;
     }
 }
 
-- (id)initForUpdatingAtPath:(NSString*)path
+- (id)initForUpdatingAtPath: (NSString*)path
 {
   int	d = open([path fileSystemRepresentation], O_RDWR);
 
   if (d < 0)
     {
-      [self dealloc];
+      [self release];
       return nil;
     }
   else
     {
-      return [self initWithFileDescriptor:d closeOnDealloc:YES];
+      return [self initWithFileDescriptor: d closeOnDealloc: YES];
     }
 }
 
@@ -358,11 +425,11 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
   if (fh_stderr)
     {
       [fh_stderr retain];
-      [self dealloc];
+      [self release];
     }
   else
     {
-      [self initWithFileDescriptor:2 closeOnDealloc:NO];
+      [self initWithFileDescriptor: 2 closeOnDealloc: NO];
       fh_stderr = self;
     }
   self = fh_stderr;
@@ -376,11 +443,11 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
   if (fh_stdin)
     {
       [fh_stdin retain];
-      [self dealloc];
+      [self release];
     }
   else
     {
-      [self initWithFileDescriptor:0 closeOnDealloc:NO];
+      [self initWithFileDescriptor: 0 closeOnDealloc: NO];
       fh_stdin = self;
     }
   self = fh_stdin;
@@ -394,11 +461,11 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
   if (fh_stdout)
     {
       [fh_stdout retain];
-      [self dealloc];
+      [self release];
     }
   else
     {
-      [self initWithFileDescriptor:1 closeOnDealloc:NO];
+      [self initWithFileDescriptor: 1 closeOnDealloc: NO];
       fh_stdout = self;
     }
   self = fh_stdout;
@@ -409,15 +476,15 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 
 - (id)initWithNullDevice
 {
-  self = [self initWithFileDescriptor:open("/dev/null", O_RDWR)
-		       closeOnDealloc:YES];
+  self = [self initWithFileDescriptor: open("/dev/null", O_RDWR)
+		       closeOnDealloc: YES];
   if (self) {
     isNullDevice = YES;
   }
   return self;
 }
 
-- (id)initWithFileDescriptor:(int)desc closeOnDealloc:(BOOL)flag
+- (id)initWithFileDescriptor: (int)desc closeOnDealloc: (BOOL)flag
 {
   self = [super init];
   if (self)
@@ -427,8 +494,8 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 
       if (fstat(desc, &sbuf) < 0)
 	{
-	  [self dealloc];
           NSLog(@"unable to get status of descriptor - %s", strerror(errno));
+	  [self release];
 	  return nil;
 	}
       if (S_ISREG(sbuf.st_mode))
@@ -455,14 +522,14 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
   return self;
 }
 
-- (id)initWithNativeHandle:(void*)hdl
+- (id)initWithNativeHandle: (void*)hdl
 {
-    return [self initWithFileDescriptor:(int)hdl closeOnDealloc:NO];
+    return [self initWithFileDescriptor: (int)hdl closeOnDealloc: NO];
 }
 
-- (id)initWithNativeHandle:(void*)hdl closeOnDealloc:(BOOL)flag
+- (id)initWithNativeHandle: (void*)hdl closeOnDealloc: (BOOL)flag
 {
-  return [self initWithFileDescriptor:(int)hdl closeOnDealloc:flag];
+  return [self initWithFileDescriptor: (int)hdl closeOnDealloc: flag];
 }
 
 - (void)checkAccept
@@ -474,7 +541,7 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
     }
   if (readInfo)
     {
-      id	operation = [readInfo objectForKey:NotificationKey];
+      id	operation = [readInfo objectForKey: NotificationKey];
 
       if (operation == NSFileHandleConnectionAcceptedNotification)
         {
@@ -498,8 +565,8 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
     }
   if ([writeInfo count] > 0)
     {
-      id	info = [writeInfo objectAtIndex:0];
-      id	operation = [info objectForKey:NotificationKey];
+      id	info = [writeInfo objectAtIndex: 0];
+      id	operation = [info objectForKey: NotificationKey];
 
       if (operation == GSFileHandleConnectCompletionNotification)
 	{
@@ -523,7 +590,7 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
     }
   if (readInfo)
     {
-      id	operation = [readInfo objectForKey:NotificationKey];
+      id	operation = [readInfo objectForKey: NotificationKey];
 
       if (operation == NSFileHandleConnectionAcceptedNotification)
         {
@@ -547,8 +614,8 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
     }
   if ([writeInfo count] > 0)
     {
-      id	info = [writeInfo objectAtIndex:0];
-      id	operation = [info objectForKey:NotificationKey];
+      id	info = [writeInfo objectAtIndex: 0];
+      id	operation = [info objectForKey: NotificationKey];
 
       if (operation == GSFileHandleConnectCompletionNotification)
 	{
@@ -581,20 +648,45 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
   int			len;
 
   [self checkRead];
-  d = [NSMutableData dataWithCapacity:0];
+  if (isNonBlocking == YES)
+    [self setNonBlocking: NO];
+  d = [NSMutableData dataWithCapacity: 0];
   if (isStandardFile)
     {
       while ((len = read(descriptor, buf, sizeof(buf))) > 0)
         {
-	  [d appendBytes:buf length:len];
+	  [d appendBytes: buf length: len];
         }
     }
   else
     {
-      if ((len = read(descriptor, buf, sizeof(buf))) > 0)
-        {
-	  [d appendBytes:buf length:len];
-        }
+      int	count;
+
+      /*
+       *	Determine number of bytes readable on descriptor.
+       */
+      if (ioctl(descriptor, FIONREAD, (char*)&count) < 0)
+	{
+	  [NSException raise: NSFileHandleOperationException
+		      format: @"unable to use FIONREAD on descriptor - %s",
+		      strerror(errno)];
+	}
+
+      if (count == 0)
+	{
+	  len = 0;	/* End-of-file	*/
+	}
+      else
+	{
+	  if (count > sizeof(buf))
+	    {
+	      count = sizeof(buf);
+	    }
+	  if ((len = read(descriptor, buf, count)) > 0)
+	    {
+	      [d appendBytes: buf length: len];
+	    }
+	}
     }
   if (len < 0)
     {
@@ -612,10 +704,12 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
   int			len;
 
   [self checkRead];
-  d = [NSMutableData dataWithCapacity:0];
+  if (isNonBlocking == YES)
+    [self setNonBlocking: NO];
+  d = [NSMutableData dataWithCapacity: 0];
   while ((len = read(descriptor, buf, sizeof(buf))) > 0)
     {
-      [d appendBytes:buf length:len];
+      [d appendBytes: buf length: len];
     }
   if (len < 0)
     {
@@ -626,13 +720,15 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
   return d;
 }
 
-- (NSData*)readDataOfLength:(unsigned int)len
+- (NSData*)readDataOfLength: (unsigned int)len
 {
   NSMutableData*	d;
   int			pos;
   char			*buf;
 
   [self checkRead];
+  if (isNonBlocking == YES)
+    [self setNonBlocking: NO];
   buf = objc_malloc(len);
   d = [NSMutableData dataWithBytesNoCopy: buf length: len];
   if ((pos = read(descriptor, [d mutableBytes], len)) < 0)
@@ -641,11 +737,11 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
                   format: @"unable to read from descriptor - %s",
                   strerror(errno)];
     }
-  [d setLength:pos];
+  [d setLength: pos];
   return d;
 }
 
-- (void)writeData:(NSData*)item
+- (void)writeData: (NSData*)item
 {
   int		rval = 0;
   const void*	ptr = [item bytes];
@@ -653,6 +749,8 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
   unsigned int	pos = 0;
 
   [self checkWrite];
+  if (isNonBlocking == YES)
+    [self setNonBlocking: NO];
   while (pos < len)
     {
       int	toWrite = len - pos;
@@ -675,56 +773,74 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 
 // Asynchronous I/O operations
 
-- (void)acceptConnectionInBackgroundAndNotifyForModes:(NSArray*)modes
+- (void)acceptConnectionInBackgroundAndNotifyForModes: (NSArray*)modes
 {
   [self checkAccept];
   readPos = 0;
   [readInfo release];
-  readInfo = [[NSMutableDictionary dictionaryWithCapacity:4] retain];
-  [readInfo setObject:NSFileHandleConnectionAcceptedNotification
-	       forKey:NotificationKey];
-  [self watchReadDescriptorForModes:modes];
+  readInfo = [[NSMutableDictionary dictionaryWithCapacity: 4] retain];
+  [readInfo setObject: NSFileHandleConnectionAcceptedNotification
+	       forKey: NotificationKey];
+  [self watchReadDescriptorForModes: modes];
 }
 
 - (void)acceptConnectionInBackgroundAndNotify
 {
-  [self acceptConnectionInBackgroundAndNotifyForModes:nil];
+  [self acceptConnectionInBackgroundAndNotifyForModes: nil];
 }
 
-- (void)readInBackgroundAndNotifyForModes:(NSArray*)modes
+- (void)readInBackgroundAndNotifyForModes: (NSArray*)modes
 {
   [self checkRead];
   readPos = 0;
   [readInfo release];
-  readInfo = [[NSMutableDictionary dictionaryWithCapacity:4] retain];
-  [readInfo setObject:NSFileHandleReadCompletionNotification
-	       forKey:NotificationKey];
-  [readInfo setObject:[NSMutableData dataWithCapacity:0]
-	       forKey:NSFileHandleNotificationDataItem];
-  [self watchReadDescriptorForModes:modes];
+  readInfo = [[NSMutableDictionary dictionaryWithCapacity: 4] retain];
+  [readInfo setObject: NSFileHandleReadCompletionNotification
+	       forKey: NotificationKey];
+  [readInfo setObject: [NSMutableData dataWithCapacity: 0]
+	       forKey: NSFileHandleNotificationDataItem];
+  [self watchReadDescriptorForModes: modes];
 }
 
 - (void)readInBackgroundAndNotify
 {
-  return [self readInBackgroundAndNotifyForModes:nil];
+  return [self readInBackgroundAndNotifyForModes: nil];
 }
 
-- (void)readToEndOfFileInBackgroundAndNotifyForModes:(NSArray*)modes
+- (void)readToEndOfFileInBackgroundAndNotifyForModes: (NSArray*)modes
 {
   [self checkRead];
   readPos = 0;
   [readInfo release];
-  readInfo = [[NSMutableDictionary dictionaryWithCapacity:4] retain];
-  [readInfo setObject:NSFileHandleReadToEndOfFileCompletionNotification
-	       forKey:NotificationKey];
-  [readInfo setObject:[NSMutableData dataWithCapacity:0]
-	       forKey:NSFileHandleNotificationDataItem];
-  [self watchReadDescriptorForModes:modes];
+  readInfo = [[NSMutableDictionary dictionaryWithCapacity: 4] retain];
+  [readInfo setObject: NSFileHandleReadToEndOfFileCompletionNotification
+	       forKey: NotificationKey];
+  [readInfo setObject: [NSMutableData dataWithCapacity: 0]
+	       forKey: NSFileHandleNotificationDataItem];
+  [self watchReadDescriptorForModes: modes];
 }
 
 - (void)readToEndOfFileInBackgroundAndNotify
 {
-  return [self readToEndOfFileInBackgroundAndNotifyForModes:nil];
+  return [self readToEndOfFileInBackgroundAndNotifyForModes: nil];
+}
+
+- (void)waitForDataInBackgroundAndNotifyForModes: (NSArray*)modes
+{
+  [self checkRead];
+  readPos = 0;
+  [readInfo release];
+  readInfo = [[NSMutableDictionary dictionaryWithCapacity: 4] retain];
+  [readInfo setObject: NSFileHandleDataAvailableNotification
+	       forKey: NotificationKey];
+  [readInfo setObject: [NSMutableData dataWithCapacity: 0]
+	       forKey: NSFileHandleNotificationDataItem];
+  [self watchReadDescriptorForModes: modes];
+}
+
+- (void)waitForDataInBackgroundAndNotify
+{
+  return [self waitForDataInBackgroundAndNotifyForModes: nil];
 }
 
 // Seeking within a file
@@ -759,7 +875,7 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
   return (unsigned long long)result;
 }
 
-- (void)seekToFileOffset:(unsigned long long)pos
+- (void)seekToFileOffset: (unsigned long long)pos
 {
   off_t	result = -1;
 
@@ -803,34 +919,34 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
     (void)sync();
 }
 
-- (void)truncateFileAtOffset:(unsigned long long)pos
+- (void)truncateFileAtOffset: (unsigned long long)pos
 {
   if (isStandardFile && descriptor >= 0)
     (void)ftruncate(descriptor, pos);
-   [self seekToFileOffset:pos];
+   [self seekToFileOffset: pos];
 }
 
-- (void)writeInBackgroundAndNotify:(NSData*)item forModes:(NSArray*)modes
+- (void)writeInBackgroundAndNotify: (NSData*)item forModes: (NSArray*)modes
 {
   NSMutableDictionary*	info;
 
   [self checkWrite];
 
-  info = [[NSMutableDictionary dictionaryWithCapacity:4] retain];
-  [info setObject:item forKey:NSFileHandleNotificationDataItem];
-  [info setObject:GSFileHandleWriteCompletionNotification
-		forKey:NotificationKey];
+  info = [[NSMutableDictionary dictionaryWithCapacity: 4] retain];
+  [info setObject: item forKey: NSFileHandleNotificationDataItem];
+  [info setObject: GSFileHandleWriteCompletionNotification
+		forKey: NotificationKey];
   if (modes)
-    [info setObject:modes forKey:NSFileHandleNotificationMonitorModes];
+    [info setObject: modes forKey: NSFileHandleNotificationMonitorModes];
 
-  [writeInfo addObject:info];
+  [writeInfo addObject: info];
   [info release];
   [self watchWriteDescriptor];
 }
 
-- (void)writeInBackgroundAndNotify:(NSData*)item;
+- (void)writeInBackgroundAndNotify: (NSData*)item;
 {
-  [self writeInBackgroundAndNotify:item forModes:nil];
+  [self writeInBackgroundAndNotify: item forModes: nil];
 }
 
 - (void)postReadNotification
@@ -842,39 +958,39 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 
   [self ignoreReadDescriptor];
   readInfo = nil;
-  modes = (NSArray*)[info objectForKey:NSFileHandleNotificationMonitorModes];
-  name = (NSString*)[info objectForKey:NotificationKey];
+  modes = (NSArray*)[info objectForKey: NSFileHandleNotificationMonitorModes];
+  name = (NSString*)[info objectForKey: NotificationKey];
 
-  n = [NSNotification notificationWithName:name object:self userInfo:info];
+  n = [NSNotification notificationWithName: name object: self userInfo: info];
 
   [info release];	/* Retained by the notification.	*/
 
-  [[NSNotificationQueue defaultQueue] enqueueNotification:n
-		postingStyle:NSPostASAP
-		coalesceMask:NSNotificationNoCoalescing
-		forModes:modes];
+  [[NSNotificationQueue defaultQueue] enqueueNotification: n
+		postingStyle: NSPostASAP
+		coalesceMask: NSNotificationNoCoalescing
+		forModes: modes];
 }
 
 - (void)postWriteNotification
 {
-  NSMutableDictionary*	info = [writeInfo objectAtIndex:0];
+  NSMutableDictionary*	info = [writeInfo objectAtIndex: 0];
   NSNotification*	n;
   NSArray*		modes;
   NSString*		name;
 
   [self ignoreWriteDescriptor];
-  modes = (NSArray*)[info objectForKey:NSFileHandleNotificationMonitorModes];
-  name = (NSString*)[info objectForKey:NotificationKey];
+  modes = (NSArray*)[info objectForKey: NSFileHandleNotificationMonitorModes];
+  name = (NSString*)[info objectForKey: NotificationKey];
 
-  n = [NSNotification notificationWithName:name object:self userInfo:info];
+  n = [NSNotification notificationWithName: name object: self userInfo: info];
 
   writePos = 0;
-  [writeInfo removeObjectAtIndex:0];	/* Retained by notification.	*/
+  [writeInfo removeObjectAtIndex: 0];	/* Retained by notification.	*/
 
-  [[NSNotificationQueue defaultQueue] enqueueNotification:n
-		postingStyle:NSPostASAP
-		coalesceMask:NSNotificationNoCoalescing
-		forModes:modes];
+  [[NSNotificationQueue defaultQueue] enqueueNotification: n
+		postingStyle: NSPostASAP
+		coalesceMask: NSNotificationNoCoalescing
+		forModes: modes];
   [self watchWriteDescriptor];	/* In case of queued writes.	*/
 }
 
@@ -904,7 +1020,7 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
   modes = nil;
 
   if (readInfo)
-    modes = (NSArray*)[readInfo objectForKey:NSFileHandleNotificationMonitorModes];
+    modes = (NSArray*)[readInfo objectForKey: NSFileHandleNotificationMonitorModes];
 
   if (modes && [modes count])
     {
@@ -914,7 +1030,7 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 	{
 	  [l removeEvent: (void*)descriptor
 		    type: ET_RDESC
-		 forMode: [modes objectAtIndex:i]
+		 forMode: [modes objectAtIndex: i]
 		     all: YES];
         }
     }
@@ -938,9 +1054,9 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 
   if ([writeInfo count] > 0)
     {
-      NSMutableDictionary*	info = [writeInfo objectAtIndex:0];
+      NSMutableDictionary*	info = [writeInfo objectAtIndex: 0];
 
-      modes=(NSArray*)[info objectForKey:NSFileHandleNotificationMonitorModes];
+      modes=(NSArray*)[info objectForKey: NSFileHandleNotificationMonitorModes];
     }
 
   if (modes && [modes count])
@@ -951,7 +1067,7 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 	{
 	  [l removeEvent: (void*)descriptor
 		    type: ET_WDESC
-		 forMode: [modes objectAtIndex:i]
+		 forMode: [modes objectAtIndex: i]
 		     all: YES];
         }
     }
@@ -962,7 +1078,7 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 	       all: YES];
 }
 
-- (void)watchReadDescriptorForModes:(NSArray*)modes;
+- (void)watchReadDescriptorForModes: (NSArray*)modes;
 {
   NSRunLoop	*l;
 
@@ -970,7 +1086,7 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
     return;
 
   l = [NSRunLoop currentRunLoop];
-  [self setNonBlocking:YES];
+  [self setNonBlocking: YES];
   if (modes && [modes count])
     {
       int		i;
@@ -980,9 +1096,9 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 	  [l addEvent: (void*)descriptor
 		 type: ET_RDESC
 	      watcher: self
-	      forMode: [modes objectAtIndex:i]];
+	      forMode: [modes objectAtIndex: i]];
         }
-      [readInfo setObject:modes forKey:NSFileHandleNotificationMonitorModes];
+      [readInfo setObject: modes forKey: NSFileHandleNotificationMonitorModes];
     }
   else
     {
@@ -1000,14 +1116,14 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 
   if ([writeInfo count] > 0)
     {
-      NSMutableDictionary*	info = [writeInfo objectAtIndex:0];
+      NSMutableDictionary*	info = [writeInfo objectAtIndex: 0];
       NSRunLoop*		l = [NSRunLoop currentRunLoop];
       NSArray*			modes = nil;
 
 
-      modes = [info objectForKey:NSFileHandleNotificationMonitorModes];
+      modes = [info objectForKey: NSFileHandleNotificationMonitorModes];
 
-      [self setNonBlocking:YES];
+      [self setNonBlocking: YES];
       if (modes && [modes count])
 	{
 	  int		i;
@@ -1017,7 +1133,7 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 	      [l addEvent: (void*)descriptor
 		     type: ET_WDESC
 		  watcher: self
-		  forMode: [modes objectAtIndex:i]];
+		  forMode: [modes objectAtIndex: i]];
 	    }
 	}
       else
@@ -1037,128 +1153,168 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 {
   NSString	*operation;
 
-  if (type == ET_RDESC) {
-    operation = [readInfo objectForKey:NotificationKey];
-    if (operation == NSFileHandleConnectionAcceptedNotification) {
-      struct sockaddr_in	buf;
-      int			desc;
-      int			blen = sizeof(buf);
-      NSFileHandle*		hdl;
+  if (isNonBlocking == NO)
+    [self setNonBlocking: YES];
+  if (type == ET_RDESC)
+    {
+      operation = [readInfo objectForKey: NotificationKey];
+      if (operation == NSFileHandleConnectionAcceptedNotification)
+	{
+	  struct sockaddr_in	buf;
+	  int			desc;
+	  int			blen = sizeof(buf);
+	  NSFileHandle*		hdl;
 
-      desc = accept(descriptor, (struct sockaddr*)&buf, &blen);
-      if (desc < 0) {
-	NSString*	s;
+	  desc = accept(descriptor, (struct sockaddr*)&buf, &blen);
+	  if (desc < 0)
+	    {
+	      NSString	*s;
 
-	s = [NSString stringWithFormat:@"Accept attempt failed - %s",
+	      s = [NSString stringWithFormat: @"Accept attempt failed - %s",
                       strerror(errno)];
-	[readInfo setObject:s forKey:GSFileHandleNotificationError];
-      }
-      else { // Accept attempt completed.
-	UnixFileHandle	*h;
-        struct sockaddr_in	sin;
-        int	size = sizeof(sin);
+	      [readInfo setObject: s forKey: GSFileHandleNotificationError];
+	    }
+	  else
+	    { // Accept attempt completed.
+	      UnixFileHandle		*h;
+	      struct sockaddr_in	sin;
+	      int			size = sizeof(sin);
 
-	h = [[UnixFileHandle alloc] initWithFileDescriptor:desc];
-        getsockname(desc, (struct sockaddr*)&sin, &size);
-        [h setAddr: &sin];
-	[readInfo setObject: h forKey: NSFileHandleNotificationFileHandleItem];
-	[h release];
-      }
-      [self postReadNotification];
-    }
-    else {
-      NSMutableData*	item;
-      int		length;
-      int		received = 0;
-      char		buf[NETBUF_SIZE];
+	      h = [[UnixFileHandle alloc] initWithFileDescriptor: desc];
+	      getpeername(desc, (struct sockaddr*)&sin, &size);
+	      [h setAddr: &sin];
+	      [readInfo setObject: h
+			   forKey: NSFileHandleNotificationFileHandleItem];
+	      [h release];
+	    }
+	  [self postReadNotification];
+	}
+      else if (operation == NSFileHandleDataAvailableNotification)
+	{
+	  [self postReadNotification];
+	}
+      else
+	{
+	  NSMutableData	*item;
+	  int		length;
+	  int		received = 0;
+	  char		buf[NETBUF_SIZE];
 
-      item = [readInfo objectForKey:NSFileHandleNotificationDataItem];
-      length = [item length];
+	  item = [readInfo objectForKey: NSFileHandleNotificationDataItem];
+	  length = [item length];
 
-      received = read(descriptor, buf, sizeof(buf));
-      if (received == 0) { // Read up to end of file.
-	[self postReadNotification];
-      }
-      else if (received < 0) {
-	if (errno != EAGAIN) {
-	  NSString*	s;
+	  received = read(descriptor, buf, sizeof(buf));
+	  if (received == 0)
+	    { // Read up to end of file.
+	      [self postReadNotification];
+	    }
+	  else if (received < 0)
+	    {
+	      if (errno != EAGAIN)
+		{
+		  NSString	*s;
 
-	  s = [NSString stringWithFormat:@"Read attempt failed - %s",
+		  s = [NSString stringWithFormat: @"Read attempt failed - %s",
                           strerror(errno)];
-	  [readInfo setObject:s forKey:GSFileHandleNotificationError];
-	  [self postReadNotification];
+		  [readInfo setObject: s forKey: GSFileHandleNotificationError];
+		  [self postReadNotification];
+		}
+	    }
+	  else
+	    {
+	      [item appendBytes: buf length: received];
+	      if (operation == NSFileHandleReadCompletionNotification)
+		{
+		  // Read a single chunk of data
+		  [self postReadNotification];
+		}
+	    }
 	}
-      }
-      else {
-	[item appendBytes:buf length:received];
-	if (operation == NSFileHandleReadCompletionNotification) {
-	  // Read a single chunk of data
-	  [self postReadNotification];
-	}
-      }
     }
-  }
-  else if (type == ET_WDESC) {
-    NSMutableDictionary*	info;
+  else if (type == ET_WDESC)
+    {
+      NSMutableDictionary	*info;
 
-    info = [writeInfo objectAtIndex:0];
-    operation = [info objectForKey:NotificationKey];
-    if (operation == GSFileHandleWriteCompletionNotification) {
-      NSData*		item;
-      int		length;
-      const void*	ptr;
+      info = [writeInfo objectAtIndex: 0];
+      operation = [info objectForKey: NotificationKey];
+      if (operation == GSFileHandleWriteCompletionNotification)
+	{
+	  NSData	*item;
+	  int		length;
+	  const void	*ptr;
 
-      item = [info objectForKey:NSFileHandleNotificationDataItem];
-      length = [item length];
-      ptr = [item bytes];
-      if (writePos < length) {
-	int	written;
+	  item = [info objectForKey: NSFileHandleNotificationDataItem];
+	  length = [item length];
+	  ptr = [item bytes];
+	  if (writePos < length)
+	    {
+	      int	written;
 
-	written = write(descriptor, (char*)ptr+writePos, length - writePos);
-	if (written <= 0) {
-	  if (errno != EAGAIN) {
-	    NSString*	s;
+	      written = write(descriptor, (char*)ptr+writePos, length-writePos);
+	      if (written <= 0)
+		{
+		  if (errno != EAGAIN)
+		    {
+		      NSString	*s;
 
-	    s = [NSString stringWithFormat:@"Write attempt failed - %s",
-			      strerror(errno)];
-	    [info setObject:s forKey:GSFileHandleNotificationError];
-	    [self postWriteNotification];
-	  }
+		      s = [NSString stringWithFormat:
+				@"Write attempt failed - %s", strerror(errno)];
+		      [info setObject: s forKey: GSFileHandleNotificationError];
+		      [self postWriteNotification];
+		    }
+		}
+	      else
+		{
+		  writePos += written;
+		}
+	    }
+	  if (writePos >= length)
+	    { // Write operation completed.
+	      [self postWriteNotification];
+	    }
 	}
-	else {
-	  writePos += written;
+      else
+	{ // Connection attempt completed.
+	  int	result;
+	  int	len = sizeof(result);
+
+	  if (getsockopt(descriptor, SOL_SOCKET, SO_ERROR,
+		(char*)&result, &len) == 0 && result != 0)
+	    {
+		NSString	*s;
+
+		s = [NSString stringWithFormat: @"Connect attempt failed - %s",
+			      strerror(result)];
+		[info setObject: s forKey: GSFileHandleNotificationError];
+	    }
+	  else
+	    {
+	      readOK = YES;
+	      writeOK = YES;
+	    }
+	  connectOK = NO;
+	  [self postWriteNotification];
 	}
-      }
-      if (writePos >= length) { // Write operation completed.
-	[self postWriteNotification];
-      }
     }
-    else { // Connection attempt completed.
-      connectOK = NO;
-      readOK = YES;
-      writeOK = YES;
-      [self postWriteNotification];
-    }
-  }
 }
 
 - (NSDate*)timedOutEvent: (void*)data
 		    type: (RunLoopEventType)type
 		 forMode: (NSString*)mode
 {
-    return nil;		/* Don't restart timed out events	*/
+  return nil;		/* Don't restart timed out events	*/
 }
 
 - (void) setAddr: (struct sockaddr_in *)sin
 {
-    address = [NSString stringWithCString: inet_ntoa(sin->sin_addr)];
-    [address retain];
-    service = [NSString stringWithFormat: @"%d", (int)ntohs(sin->sin_port)];
-    [service retain];
-    protocol = @"tcp";
+  address = [NSString stringWithCString: (char*)inet_ntoa(sin->sin_addr)];
+  [address retain];
+  service = [NSString stringWithFormat: @"%d", (int)ntohs(sin->sin_port)];
+  [service retain];
+  protocol = @"tcp";
 }
 
-- (void)setNonBlocking:(BOOL)flag
+- (void)setNonBlocking: (BOOL)flag
 {
   int	e;
 
@@ -1189,17 +1345,17 @@ getAddr(NSString* name, NSString* svc, NSString* pcl, struct sockaddr_in *sin)
 
 - (NSString*) socketAddress
 {
-    return address;
+  return address;
 }
 
 - (NSString*) socketProtocol
 {
-    return protocol;
+  return protocol;
 }
 
 - (NSString*) socketService
 {
-    return service;
+  return service;
 }
 
 @end
