@@ -1,5 +1,5 @@
 /* Implementation of functions for dissecting/making method calls 
-   Copyright (C) 1994, 1995, 1996, 1997 Free Software Foundation, Inc.
+   Copyright (C) 1994, 1995, 1996, 1997, 1998 Free Software Foundation, Inc.
    
    Written by:  Andrew Kachites McCallum <mccallum@gnu.ai.mit.edu>
    Created: Oct 1994
@@ -29,24 +29,11 @@
 */
 
 #include <config.h>
-#include <gnustep/base/preface.h>
-#include <gnustep/base/mframe.h>
-#include <gnustep/base/MallocAddress.h>
+#include <base/preface.h>
+#include <mframe.h>
+#include <Foundation/NSData.h>
 #include <Foundation/NSException.h>
 #include <stdlib.h>
-#include <assert.h>
-
-/* Modifications to fix return of floats, doubles, and structures
-   by Richard Frith-Macdonald <richard@brainstorm.co.uk> 1997
-
-   I define STRUCT_ADDR_ON_STACK if functions returns structures, unions
-   and arrays by writing directly to a location passed to them as their
-   first argument.  If this is the case then all the visible arguments
-   are displaced in the arg frame by the size of a pointer.
-
-   I don't know which machines work this way, so you need to alter the
-   define by hand as suits you.  */
-#define	STRUCT_ADDR_ON_STACK	1
 
 /* Deal with strrchr: */
 #if STDC_HEADERS || HAVE_STRING_H
@@ -84,18 +71,406 @@
    value.  The code doesn't currently handle that case.
    */
 
-/* Do we need separate _PASSED_BY_REFERENCE and _RETURNED_BY_REFERENCE? */
 
-#if (sparc) || (hppa) || (AM29K)
-#define MFRAME_STRUCTURES_PASSED_BY_REFERENCE 1
-#else
-#define MFRAME_STRUCTURES_PASSED_BY_REFERENCE 0
-#endif
+char*
+mframe_build_signature(const char *typePtr, int *size, int *narg, char *buf)
+{
+  MFRAME_ARGS	cum;
+  BOOL		doMalloc = NO;
+  const		char	*types;
+  char		*start;
+  char		*dest;
+  int		total = 0;
+  int		count = 0;
+
+  /*
+   *	If we have not been given a buffer - allocate space on the stack for
+   *	the largest concievable type encoding.
+   */
+  if (buf == 0)
+    {
+      doMalloc = YES;
+      buf = alloca((strlen(typePtr)+1)*16);
+    }
+
+  /*
+   *	Copy the return type info (including qualifiers) into the buffer.
+   */
+  types = objc_skip_typespec(typePtr);
+  strncpy(buf, typePtr, types - typePtr);
+  buf[types-typePtr] = '\0'; 
+
+  /*
+   *	Point to the return type, initialise size of stack args, and skip
+   *	to the first argument.
+   */
+  types = objc_skip_type_qualifiers(typePtr);
+  MFRAME_INIT_ARGS(cum, types);
+  types = objc_skip_typespec(types);
+  if (*types == '+')
+    {
+      types++;
+    }
+  while (isdigit(*types))
+    {
+      types++;
+    }
+
+  /*
+   *	Where to start putting encoding information - leave enough room for
+   *	the size of the stack args to be stored after the return type.
+   */
+  start = &buf[strlen(buf)+10];
+  dest = start;
+
+  /*
+   *	Now step through all the arguments - copy any type qualifiers, but
+   *	let the macro write all the other info into the buffer.
+   */
+  while (types && *types)
+    {
+      const char	*qual = types;
+
+      /*
+       *	If there are any type qualifiers - copy the through to the
+       *	destination.
+       */
+      types = objc_skip_type_qualifiers(types);
+      while (qual < types)
+	{
+	  *dest++ = *qual++;
+	}
+      MFRAME_ARG_ENCODING(cum, types, total, dest);
+      count++;
+    }
+  *dest = '\0';
+
+  /*
+   *	Write the total size of the stack arguments after the return type,
+   *	then copy the remaining type information to fill the gap.
+   */
+  sprintf(&buf[strlen(buf)], "%d", total);
+  dest = &buf[strlen(buf)];
+  while (*start)
+    {
+      *dest++ = *start++;
+    }
+  *dest = '\0';
+
+  /*
+   *	If we have written into a local buffer - we need to allocate memory
+   *	in which to return our result.
+   */
+  if (doMalloc)
+    {
+      char	*tmp = objc_malloc(dest - buf + 1);
+
+      strcpy(tmp, buf);
+      buf = tmp;
+    }
+
+  /*
+   *	If the caller wants to know the total size of the stack and/or the
+   *	number of arguments, return them in the appropriate variables.
+   */
+  if (size)
+    {
+      *size = total;
+    }
+  if (narg)
+    {
+      *narg = count;
+    }
+  return buf;
+}
 
 
-#define ROUND(V, A) \
-  ({ typeof(V) __v=(V); typeof(A) __a=(A); \
-     __a*((__v+__a-1)/__a); })
+/*
+ *      Step through method encoding information extracting details.
+ */
+const char *
+mframe_next_arg(const char *typePtr, NSArgumentInfo *info)
+{
+  NSArgumentInfo	local;
+  BOOL			flag;
+
+  if (info == 0)
+    {
+      info = &local;
+    }
+  /*
+   *	Skip past any type qualifiers - if the caller wants them, return them.
+   */
+  flag = YES;
+  info->qual = 0;
+  while (flag)
+    {
+      switch (*typePtr)
+	{
+	  case _C_CONST:  info->qual |= _F_CONST; break;
+	  case _C_IN:     info->qual |= _F_IN; break;
+	  case _C_INOUT:  info->qual |= _F_INOUT; break;
+	  case _C_OUT:    info->qual |= _F_OUT; break;
+	  case _C_BYCOPY: info->qual |= _F_BYCOPY; break;
+#ifdef	_C_BYREF
+	  case _C_BYREF:  info->qual |= _F_BYREF; break;
+#endif
+	  case _C_ONEWAY: info->qual |= _F_ONEWAY; break;
+	  default: flag = NO;
+	}
+      if (flag)
+	{
+	  typePtr++;
+	}
+    }
+
+  info->type = typePtr;
+
+  /*
+   *	Scan for size and alignment information.
+   */
+  switch (*typePtr++)
+    {
+      case _C_ID:
+	info->size = sizeof(id);
+	info->align = __alignof__(id);
+	break;
+
+      case _C_CLASS:
+	info->size = sizeof(Class);
+	info->align = __alignof__(Class);
+	break;
+
+      case _C_SEL:
+	info->size = sizeof(SEL);
+	info->align = __alignof__(SEL);
+	break;
+
+      case _C_CHR:
+	info->size = sizeof(char);
+	info->align = __alignof__(char);
+	break;
+
+      case _C_UCHR:
+	info->size = sizeof(unsigned char);
+	info->align = __alignof__(unsigned char);
+	break;
+
+      case _C_SHT:
+	info->size = sizeof(short);
+	info->align = __alignof__(short);
+	break;
+
+      case _C_USHT:
+	info->size = sizeof(unsigned short);
+	info->align = __alignof__(unsigned short);
+	break;
+
+      case _C_INT:
+	info->size = sizeof(int);
+	info->align = __alignof__(int);
+	break;
+
+      case _C_UINT:
+	info->size = sizeof(unsigned int);
+	info->align = __alignof__(unsigned int);
+	break;
+
+      case _C_LNG:
+	info->size = sizeof(long);
+	info->align = __alignof__(long);
+	break;
+
+      case _C_ULNG:
+	info->size = sizeof(unsigned long);
+	info->align = __alignof__(unsigned long);
+	break;
+
+#ifdef	_C_LNG_LNG
+      case _C_LNG_LNG:
+	info->size = sizeof(long long);
+	info->align = __alignof__(long long);
+	break;
+
+      case _C_ULNG_LNG:
+	info->size = sizeof(unsigned long long);
+	info->align = __alignof__(unsigned long long);
+	break;
+
+#endif
+      case _C_FLT:
+	info->size = sizeof(float);
+	info->align = __alignof__(float);
+	break;
+
+      case _C_DBL:
+	info->size = sizeof(double);
+	info->align = __alignof__(double);
+	break;
+
+      case _C_PTR:
+	info->size = sizeof(char*);
+	info->align = __alignof__(char*);
+	if (*typePtr == '?')
+	  {
+	    typePtr++;
+	  }
+	else
+	  {
+	    typePtr = mframe_next_arg(typePtr, &local);
+	    info->isReg = local.isReg;
+	    info->offset = local.offset;
+	  }
+	break;
+
+      case _C_ATOM:
+      case _C_CHARPTR:
+	info->size = sizeof(char*);
+	info->align = __alignof__(char*);
+	break;
+
+      case _C_ARY_B:
+	{
+	  int	length = atoi(typePtr);
+
+	  while (isdigit(*typePtr))
+	    {
+	      typePtr++;
+	    }
+	  typePtr = mframe_next_arg(typePtr, &local);
+	  info->size = length * ROUND(local.size, local.align);
+	  info->align = local.align;
+	  typePtr++;	/* Skip end-of-array	*/
+	}
+	break; 
+
+      case _C_STRUCT_B:
+	{
+	  struct { int x; double y; } fooalign;
+	  int acc_size = 0;
+	  int acc_align = __alignof__(fooalign);
+
+	  /*
+	   *	Skip "<name>=" stuff.
+	   */
+	  while (*typePtr != _C_STRUCT_E)
+	    {
+	      if (*typePtr++ == '=')
+		{
+		  break;
+		}
+	    }
+	  /*
+	   *	Base structure alignment on first element.
+	   */
+	  if (*typePtr != _C_STRUCT_E)
+	    {
+	      typePtr = mframe_next_arg(typePtr, &local);
+	      if (typePtr == 0)
+		{
+		  return 0;		/* error	*/
+		}
+	      acc_size = ROUND(acc_size, local.align);
+	      acc_size += local.size;
+	      acc_align = MAX(local.align, __alignof__(fooalign));
+	    }
+	  /*
+	   *	Continue accumulating structure size.
+	   */
+	  while (*typePtr != _C_STRUCT_E)
+	    {
+	      typePtr = mframe_next_arg(typePtr, &local);
+	      if (typePtr == 0)
+		{
+		  return 0;		/* error	*/
+		}
+	      acc_size = ROUND(acc_size, local.align);
+	      acc_size += local.size;
+	    }
+	  info->size = acc_size;
+	  info->align = acc_align;
+	  typePtr++;	/* Skip end-of-struct	*/
+	}
+	break;
+
+      case _C_UNION_B:
+	{
+	  int	max_size = 0;
+	  int	max_align = 0;
+
+	  /*
+	   *	Skip "<name>=" stuff.
+	   */
+	  while (*typePtr != _C_UNION_E)
+	    {
+	      if (*typePtr++ == '=')
+		{
+		  break;
+		}
+	    }
+	  while (*typePtr != _C_UNION_E)
+	    {
+	      typePtr = mframe_next_arg(typePtr, &local);
+	      if (typePtr == 0)
+		{
+		  return 0;		/* error	*/
+		}
+	      max_size = MAX(max_size, local.size);
+	      max_align = MAX(max_align, local.align);
+	    }
+	  info->size = max_size;
+	  info->align = max_align;
+	  typePtr++;	/* Skip end-of-union	*/
+	}
+	break;
+
+      case _C_VOID:
+	info->size = 0;
+	info->align = __alignof__(char*);
+	break;
+
+      default:
+	return 0;
+    }
+
+  if (typePtr == 0)
+    {		/* Error condition.	*/
+      return 0;
+    }
+
+  /*
+   *	If we had a pointer argument, we will already have gathered
+   *	(and skipped past) the argframe offset information - so we
+   *	don't need to (and can't) do it here.
+   */
+  if (info->type[0] != _C_PTR || info->type[1] == '?')
+    {
+      /*
+       *	May tell the caller if the item is stored in a register.
+       */
+      if (*typePtr == '+')
+	{
+	  typePtr++;
+	  info->isReg = YES;
+	}
+      else if (info->isReg)
+	{
+	  info->isReg = NO;
+	}
+
+      /*
+       *	May tell the caller what the stack/register offset is for
+       *	this argument.
+       */
+      info->offset = 0;
+      while (isdigit(*typePtr))
+	{
+	  info->offset = info->offset * 10 + (*typePtr++ - '0');
+	}
+    }
+
+  return typePtr;
+}
 
 
 
@@ -106,6 +481,7 @@ int
 method_types_get_number_of_arguments (const char *type)
 {
   int i = 0;
+
   while (*type)
     {
       type = objc_skip_argspec (type);
@@ -130,44 +506,49 @@ int
 method_types_get_size_of_register_arguments(const char *types)
 {
   const char* type = strrchr(types, '+');
+
   if (type)
-    return atoi(++type) + sizeof(void*);
+    {
+      return atoi(++type) + sizeof(void*);
+    }
   else
-    return 0;
+    {
+      return 0;
+    }
 }
 
 
 /* To fix temporary bug in method_get_next_argument() on NeXT boxes */
 /* xxx Perhaps this isn't working with the NeXT runtime? */
-/* Uses 'offset' to adjust for a pointer used to return a structur */
 
 char*
-method_types_get_next_argument (arglist_t argf,
-				const char **type, int offset)
+method_types_get_next_argument (arglist_t argf, const char **type)
 {
   const char *t = objc_skip_argspec (*type);
-   union {
-     char *arg_ptr;
-     char arg_regs[sizeof (char*)];
-   } *argframe;
+  arglist_t	argframe;
 
   argframe = (void*)argf;
 
   if (*t == '\0')
-    return 0;
-
+    {
+      return 0;
+    }
   *type = t;
   t = objc_skip_typespec (t);
 
   if (*t == '+')
-    return argframe->arg_regs + atoi(++t);
+    {
+      return argframe->arg_regs + atoi(++t);
+    }
   else
-    /* xxx What's going on here?  This -8 needed on my 68k NeXT box. */
+    {
+      /* xxx What's going on here?  This -8 needed on my 68k NeXT box. */
 #if NeXT
-    return argframe->arg_ptr + offset + (atoi(t) - 8);
+      return argframe->arg_ptr + (atoi(t) - 8);
 #else
-    return argframe->arg_ptr + offset + atoi(t);
+      return argframe->arg_ptr + atoi(t);
 #endif
+    }
 }
 
 
@@ -186,32 +567,27 @@ method_types_get_next_argument (arglist_t argf,
    e.g. an argument declared (out char*). */
 
 BOOL
-mframe_dissect_call (arglist_t argframe, const char *type,
-		     void (*encoder)(int,void*,const char*,int))
+mframe_dissect_call_opts (arglist_t argframe, const char *type,
+		     void (*encoder)(int,void*,const char*,int),
+			BOOL pass_pointers)
 {
   unsigned flags;
   char *datum;
   int argnum;
   BOOL out_parameters = NO;
-  int off = 0;
 
-#ifdef	STRUCT_ADDR_ON_STACK
-  /* On machines which pass a pointer to a location for returning
-     structures before the first real argument, we need to use an offset
-     into the arg frame. */
   if (*type == _C_STRUCT_B || *type == _C_UNION_B || *type == _C_ARY_B) {
-    off = sizeof(void*);
+    datum = alloca((strlen(type)+1)*10);
+    type = mframe_build_signature(type, 0, 0, datum);
   }
-#endif
-
   /* Enumerate all the arguments in ARGFRAME, and call ENCODER for
      each one.  METHOD_TYPES_GET_NEXT_ARGUEMENT() returns 0 when
      there are no more arguments, otherwise it returns a pointer to the
      argument in the ARGFRAME. */
 
-  for (datum = method_types_get_next_argument(argframe, &type, off), argnum=0;
+  for (datum = method_types_get_next_argument(argframe, &type), argnum=0;
        datum;
-       datum = method_types_get_next_argument(argframe, &type, off), argnum++)
+       datum = method_types_get_next_argument(argframe, &type), argnum++)
     {
       /* Get the type qualifiers, like IN, OUT, INOUT, ONEWAY. */
       flags = objc_get_type_qualifiers(type);
@@ -247,11 +623,6 @@ mframe_dissect_call (arglist_t argframe, const char *type,
 	  break;
 
 	case _C_PTR:
-	  /* Handle an argument that is a pointer to a non-char.  But
-	     (void*) and (anything**) is not allowed. */
-	  /* The argument is a pointer to something; increment TYPE
-	       so we can see what it is a pointer to. */
-	  type++;
 	  /* If the pointer's value is qualified as an OUT parameter,
 	     or if it not explicitly qualified as an IN parameter,
 	     then we will have to get the value pointed to again after
@@ -259,11 +630,22 @@ mframe_dissect_call (arglist_t argframe, const char *type,
 	     it.  Set OUT_PARAMETERS accordingly. */
 	  if ((flags & _F_OUT) || !(flags & _F_IN))
 	    out_parameters = YES;
-	  /* If the pointer's value is qualified as an IN parameter,
-             or not explicity qualified as an OUT parameter, then
-             encode it. */
-	  if ((flags & _F_IN) || !(flags & _F_OUT))
-	    (*encoder) (argnum, *(void**)datum, type, flags);
+	  if (pass_pointers) {
+	    if ((flags & _F_IN) || !(flags & _F_OUT))
+	      (*encoder) (argnum, datum, type, flags);
+	  }
+	  else {
+	    /* Handle an argument that is a pointer to a non-char.  But
+	       (void*) and (anything**) is not allowed. */
+	    /* The argument is a pointer to something; increment TYPE
+		 so we can see what it is a pointer to. */
+	    type++;
+	    /* If the pointer's value is qualified as an IN parameter,
+	       or not explicity qualified as an OUT parameter, then
+	       encode it. */
+	    if ((flags & _F_IN) || !(flags & _F_OUT))
+	      (*encoder) (argnum, *(void**)datum, type, flags);
+	  }
 	  break;
 
 	case _C_STRUCT_B:
@@ -272,9 +654,9 @@ mframe_dissect_call (arglist_t argframe, const char *type,
 	  /* Handle struct and array arguments. */
 	  /* Whether DATUM points to the data, or points to a pointer
 	     that points to the data, depends on the value of
-	     MFRAME_STRUCTURES_PASSED_BY_REFERENCE.  Do the right thing
+	     MFRAME_STRUCT_BYREF.  Do the right thing
 	     so that ENCODER gets a pointer to directly to the data. */
-#if MFRAME_STRUCTURES_PASSED_BY_REFERENCE
+#if MFRAME_STRUCT_BYREF
 	  (*encoder) (argnum, *(void**)datum, type, flags);
 #else
 	  (*encoder) (argnum, datum, type, flags);
@@ -294,6 +676,13 @@ mframe_dissect_call (arglist_t argframe, const char *type,
   return out_parameters;
 }
 
+BOOL
+mframe_dissect_call (arglist_t argframe, const char *type,
+		     void (*encoder)(int,void*,const char*,int))
+{
+  return mframe_dissect_call_opts(argframe, type, encoder, NO);
+}
+
 
 /* mframe_do_call()
 
@@ -340,7 +729,7 @@ mframe_dissect_call (arglist_t argframe, const char *type,
      -decodeValueOfCType:at:withName: to decode a char* string, you
      should remember that -decodeValueOfCType:at:withName: malloc's
      new memory to hold the string, and DECODER should autorelease the
-     malloc'ed pointer, using the MallocAddress class.
+     malloc'ed pointer, using the NSData class.
 
 
    ENCODER should be a pointer to a function that records the method's
@@ -361,12 +750,17 @@ mframe_dissect_call (arglist_t argframe, const char *type,
      pass-by-reference parameters.  The ENCODER function should place
      the value at memory location DATA wherever the user wants to
      record the ARGNUM'th return value.
+
+  PASS_POINTERS is a flag saying whether pointers should be passed
+  as pointers (for local stuff) or should be assumed to point to a
+  single data item (for distributed objects).
 */
 
 void
-mframe_do_call (const char *encoded_types,
+mframe_do_call_opts (const char *encoded_types,
 		void(*decoder)(int,void*,const char*),
-		void(*encoder)(int,void*,const char*,int))
+		void(*encoder)(int,void*,const char*,int),
+		BOOL pass_pointers)
 {
   /* The method type string obtained from the target's OBJC_METHOD 
      structure for the selector we're sending. */
@@ -385,17 +779,8 @@ mframe_do_call (const char *encoded_types,
   int stack_argsize;
   /* The number bytes for holding arguments passed in registers. */
   int reg_argsize;
-  /* Offset for arguments on stack. */
-  int off = 0;
   /* The structure for holding the arguments to the method. */
-#if NeXT_runtime
-  union {
-    char *arg_ptr;
-    char arg_regs[sizeof (char*)];
-  } *argframe;
-#else
   arglist_t argframe;
-#endif
   /* A pointer into the ARGFRAME; points at individual arguments. */
   char *datum;
   /* Type qualifier flags; see <objc/objc-api.h>. */
@@ -472,22 +857,22 @@ mframe_do_call (const char *encoded_types,
   NSCParameterAssert (type);
   NSCParameterAssert (sel_types_match(encoded_types, type));
 
-
-#ifdef	STRUCT_ADDR_ON_STACK
-  /* On machines which pass a pointer to a location for returning
-     structures before the first real argument, we need to use an offset
-     into the arg frame. */
+  /*
+   *	The compiler/runtime doesn't always seem to get the encoding right
+   *	for our purposes - so we generate our own encoding as required by
+   *	__builtin_apply().
+   */
   if (*type == _C_STRUCT_B || *type == _C_UNION_B || *type == _C_ARY_B) {
-    off = sizeof(void*);
+    tmptype = alloca((strlen(type)+1)*10);
+    type = mframe_build_signature(type, 0, 0, (char*)tmptype);
   }
-#endif
 
   /* Allocate an argframe, using memory on the stack */
 
   /* Calculate the amount of memory needed for storing variables that
      are passed in registers, and the amount of memory for storing
      variables that are passed on the stack. */
-  stack_argsize = off + method_types_get_size_of_stack_arguments (type);
+  stack_argsize = method_types_get_size_of_stack_arguments (type);
   reg_argsize = method_types_get_size_of_register_arguments (type);
   /* Allocate the space for variables passed in registers. */
   argframe = (arglist_t) alloca(sizeof(char*) + reg_argsize);
@@ -497,10 +882,15 @@ mframe_do_call (const char *encoded_types,
   else
     argframe->arg_ptr = 0;
 
-  /* If we are passing a pointer to return a structure in, we must allocate
-     the memory for it and put it at the start of the argframe. */
-  if (off)
-    *(void**)argframe->arg_ptr = alloca (objc_sizeof_type(type));
+  if (*type == _C_STRUCT_B || *type == _C_UNION_B || *type == _C_ARY_B)
+    {
+      void	*buf;
+
+    /* If we are passing a pointer to return a structure in, we must allocate
+       the memory for it and put it in the correct place in the argframe. */
+      buf = alloca(objc_sizeof_type(type));
+      MFRAME_SET_STRUCT_ADDR(argframe, type, buf);
+    }
 
   /* Put OBJECT and SELECTOR into the ARGFRAME. */
 
@@ -509,15 +899,15 @@ mframe_do_call (const char *encoded_types,
   etmptype = objc_skip_argspec (encoded_types);
   /* Get a pointer into ARGFRAME, pointing to the location where the
      first argument is to be stored. */
-  datum = method_types_get_next_argument (argframe, &tmptype, off);
+  datum = method_types_get_next_argument (argframe, &tmptype);
   NSCParameterAssert (datum);
   NSCParameterAssert (*tmptype == _C_ID);
   /* Put the target object there. */
   *(id*)datum = object;
-  /* Get a pionter into ARGFRAME, pointing to the location where the
+  /* Get a pointer into ARGFRAME, pointing to the location where the
      second argument is to be stored. */
   etmptype = objc_skip_argspec(etmptype);
-  datum = method_types_get_next_argument(argframe, &tmptype, off);
+  datum = method_types_get_next_argument(argframe, &tmptype);
   NSCParameterAssert (datum);
   NSCParameterAssert (*tmptype == _C_SEL);
   /* Put the selector there. */
@@ -528,10 +918,10 @@ mframe_do_call (const char *encoded_types,
      ARGFRAME.  Step TMPTYPE and ETMPTYPE in lock-step through their
      method type strings. */
 
-  for (datum = method_types_get_next_argument (argframe, &tmptype, off),
+  for (datum = method_types_get_next_argument (argframe, &tmptype),
        etmptype = objc_skip_argspec (etmptype), argnum = 2;
        datum;
-       datum = method_types_get_next_argument (argframe, &tmptype, off),
+       datum = method_types_get_next_argument (argframe, &tmptype),
        etmptype = objc_skip_argspec (etmptype), argnum++)
     {
       /* Get the type qualifiers, like IN, OUT, INOUT, ONEWAY. */
@@ -566,18 +956,13 @@ mframe_do_call (const char *encoded_types,
 	     Note: the decoder allocates memory for holding the
 	     string, and it is also responsible for making sure that
 	     the memory gets freed eventually, (usually through the
-	     autorelease of MallocAddress object). */
+	     autorelease of NSData object). */
 	  if ((flags & _F_IN) || !(flags & _F_OUT))
 	    (*decoder) (argnum, datum, tmptype);
 
 	  break;
 
 	case _C_PTR:
-	  /* Handle an argument that is a pointer to a non-char.  But
-	     (void*) and (anything**) is not allowed. */
-	  /* The argument is a pointer to something; increment TYPE
-	       so we can see what it is a pointer to. */
-	  tmptype++;
 	  /* If the pointer's value is qualified as an OUT parameter,
 	     or if it not explicitly qualified as an IN parameter,
 	     then we will have to get the value pointed to again after
@@ -585,16 +970,29 @@ mframe_do_call (const char *encoded_types,
 	     it.  Set OUT_PARAMETERS accordingly. */
 	  if ((flags & _F_OUT) || !(flags & _F_IN))
 	    out_parameters = YES;
-	  /* Allocate some memory to be pointed to, and to hold the
-	     value.  Note that it is allocated on the stack, and
-	     methods that want to keep the data pointed to, will have
-	     to make their own copies. */
-	  *(void**)datum = alloca (objc_sizeof_type (tmptype));
-	  /* If the pointer's value is qualified as an IN parameter,
-             or not explicity qualified as an OUT parameter, then
-             decode it. */
-	  if ((flags & _F_IN) || !(flags & _F_OUT))
-	    (*decoder) (argnum, *(void**)datum, tmptype);
+	  if (pass_pointers)
+	    {
+	      if ((flags & _F_IN) || !(flags & _F_OUT))
+		(*decoder) (argnum, datum, tmptype);
+	    }
+	  else
+	    {
+	      /* Handle an argument that is a pointer to a non-char.  But
+		 (void*) and (anything**) is not allowed. */
+	      /* The argument is a pointer to something; increment TYPE
+		   so we can see what it is a pointer to. */
+	      tmptype++;
+	      /* Allocate some memory to be pointed to, and to hold the
+		 value.  Note that it is allocated on the stack, and
+		 methods that want to keep the data pointed to, will have
+		 to make their own copies. */
+	      *(void**)datum = alloca (objc_sizeof_type (tmptype));
+	      /* If the pointer's value is qualified as an IN parameter,
+		 or not explicity qualified as an OUT parameter, then
+		 decode it. */
+	      if ((flags & _F_IN) || !(flags & _F_OUT))
+		(*decoder) (argnum, *(void**)datum, tmptype);
+	    }
 	  break;
 
 	case _C_STRUCT_B:
@@ -603,9 +1001,9 @@ mframe_do_call (const char *encoded_types,
 	  /* Handle struct and array arguments. */
 	  /* Whether DATUM points to the data, or points to a pointer
 	     that points to the data, depends on the value of
-	     MFRAME_STRUCTURES_PASSED_BY_REFERENCE.  Do the right thing
+	     MFRAME_STRUCT_BYREF.  Do the right thing
 	     so that ENCODER gets a pointer to directly to the data. */
-#if MFRAME_STRUCTURES_PASSED_BY_REFERENCE
+#if MFRAME_STRUCT_BYREF
 	  /* Allocate some memory to be pointed to, and to hold the
 	     data.  Note that it is allocated on the stack, and
 	     methods that want to keep the data pointed to, will have
@@ -673,11 +1071,18 @@ mframe_do_call (const char *encoded_types,
       break;
 
     case _C_PTR:
-      /* The argument is a pointer to something; increment TYPE
-	 so we can see what it is a pointer to. */
-      tmptype++;
-      /* Encode the value that was pointed to. */
-      (*encoder) (-1, *(void**)retframe, tmptype, flags);
+      if (pass_pointers)
+	{
+	  (*encoder) (-1, retframe, tmptype, flags);
+	}
+      else
+	{
+	  /* The argument is a pointer to something; increment TYPE
+	     so we can see what it is a pointer to. */
+	  tmptype++;
+	  /* Encode the value that was pointed to. */
+	  (*encoder) (-1, *(void**)retframe, tmptype, flags);
+	}
       break;
 
     case _C_STRUCT_B:
@@ -685,16 +1090,7 @@ mframe_do_call (const char *encoded_types,
     case _C_ARY_B:
       /* The argument is a structure or array returned by value.
 	 (In C, are array's allowed to be returned by value?) */
-      /* xxx Does MFRAME_STRUCTURES_PASSED_BY_REFERENCE have
-	 anything to do with how structures are returned?  What about
-	 struct's that are smaller than sizeof(void*)?  Are they also
-	 returned by reference like this? */
-      /* If 'off' is non-zero, the pointer to the stored structure is at
-         the start of the arg frame rather than in retframe. */
-      if (off)
-        (*encoder) (-1, *(void**)argframe->arg_ptr, tmptype, flags);
-      else
-        (*encoder) (-1, *(void**)retframe, tmptype, flags);
+      (*encoder)(-1, MFRAME_GET_STRUCT_ADDR(argframe, tmptype), tmptype, flags);
       break;
 
     case _C_FLT:
@@ -753,11 +1149,11 @@ mframe_do_call (const char *encoded_types,
     {
       /* Step through all the arguments, finding the ones that were
 	 passed by reference. */
-      for (datum = method_types_get_next_argument (argframe, &tmptype, off), 
+      for (datum = method_types_get_next_argument (argframe, &tmptype), 
 	     argnum = 1,
 	     etmptype = objc_skip_argspec (etmptype);
 	   datum;
-	   datum = method_types_get_next_argument (argframe, &tmptype, off), 
+	   datum = method_types_get_next_argument (argframe, &tmptype), 
 	     argnum++,
 	     etmptype = objc_skip_argspec (etmptype))
 	{
@@ -803,6 +1199,14 @@ mframe_do_call (const char *encoded_types,
   return;
 }
 
+void
+mframe_do_call (const char *encoded_types,
+		void(*decoder)(int,void*,const char*),
+		void(*encoder)(int,void*,const char*,int))
+{
+  mframe_do_call_opts(encoded_types, decoder, encoder, NO);
+}
+
 
 /* mframe_build_return()
 
@@ -822,10 +1226,11 @@ mframe_do_call (const char *encoded_types,
 */
 
 retval_t 
-mframe_build_return (arglist_t argframe, 
+mframe_build_return_opts (arglist_t argframe, 
 		     const char *type, 
 		     BOOL out_parameters,
-		     void(*decoder)(int,void*,const char*,int))
+		     void(*decoder)(int,void*,const char*,int),
+		     BOOL pass_pointers)
 {
   /* A pointer to the memory that will hold the return value. */
   retval_t retframe = NULL;
@@ -839,7 +1244,6 @@ mframe_build_return (arglist_t argframe,
   const char *tmptype;
   /* A pointer into the ARGFRAME; points at individual arguments. */
   void *datum;
-  int	off = 0;
   const char *rettype;
   /* For returning strucutres etc */
   typedef struct { id many[8];} __big;
@@ -893,20 +1297,16 @@ mframe_build_return (arglist_t argframe,
       return __builtin_apply((apply_t)return_short, args, sizeof(void*));
     }
 
+  if (*type == _C_STRUCT_B || *type == _C_UNION_B || *type == _C_ARY_B)
+    {
+      tmptype = alloca((strlen(type)+1)*10);
+      type = mframe_build_signature(type, 0, 0, (char*)tmptype);
+    }
   /* Get the return type qualifier flags, and the return type. */
   flags = objc_get_type_qualifiers(type);
   tmptype = objc_skip_type_qualifiers(type);
   rettype = tmptype;
 
-#ifdef	STRUCT_ADDR_ON_STACK
-  /* On machines which pass a pointer to a location for returning
-     structures before the first real argument, we need to use an offset
-     into the arg frame. */
-  if (*rettype==_C_STRUCT_B || *rettype==_C_UNION_B || *rettype==_C_ARY_B) {
-    off = sizeof(void*);
-  }
-#endif
-
   /* Decode the return value and pass-by-reference values, if there
      are any.  OUT_PARAMETERS should be the value returned by
      mframe_dissect_call(). */
@@ -939,17 +1339,28 @@ mframe_build_return (arglist_t argframe,
 	  switch (*tmptype)
 	    {
 	    case _C_PTR:
-	      /* We are returning a pointer to something. */
-	      /* Increment TYPE so we can see what it is a pointer to. */
-	      tmptype++;
-	      /* Allocate some memory to hold the value we're pointing to. */
-	      *(void**)retframe = 
-		objc_malloc (objc_sizeof_type (tmptype));
-	      /* We are responsible for making sure this memory gets free'd
-		 eventually.  Ask MallocAddress class to autorelease it. */
-	      [MallocAddress autoreleaseMallocAddress: *(void**)retframe];
-	      /* Decode the return value into the memory we allocated. */
-	      (*decoder) (-1, *(void**)retframe, tmptype, flags);
+	      if (pass_pointers)
+		{
+		  (*decoder) (-1, retframe, tmptype, flags);
+		}
+	      else
+		{
+		  unsigned retLength;
+
+		  /* We are returning a pointer to something. */
+		  /* Increment TYPE so we can see what it is a pointer to. */
+		  tmptype++;
+		  retLength = objc_sizeof_type(tmptype);
+		  /* Allocate memory to hold the value we're pointing to. */
+		  *(void**)retframe = 
+		    objc_malloc (retLength);
+		  /* We are responsible for making sure this memory gets free'd
+		     eventually.  Ask NSData class to autorelease it. */
+		  [NSData dataWithBytesNoCopy: *(void**)retframe
+				       length: retLength];
+		  /* Decode the return value into the memory we allocated. */
+		  (*decoder) (-1, *(void**)retframe, tmptype, flags);
+		}
 	      break;
 
 	    case _C_STRUCT_B: 
@@ -957,18 +1368,7 @@ mframe_build_return (arglist_t argframe,
 	    case _C_ARY_B:
 	      /* The argument is a structure or array returned by value.
 		 (In C, are array's allowed to be returned by value?) */
-	      /* xxx Does MFRAME_STRUCTURES_PASSED_BY_REFERENCE
-		 have anything to do with how structures are returned?
-		 What about struct's that are smaller than
-		 sizeof(void*)?  Are they also returned by reference
-		 like this? */
-	      if (off)
-		/* If we have been given a pointer to a location to return
-                   the structure in, us it. */
-	        *(void**)retframe = *(void**)argframe->arg_ptr;
-	      else
-	          /* Allocate some memory to hold the struct or array. */
-	          *(void**)retframe = alloca (objc_sizeof_type (tmptype));
+	      *(void**)retframe = MFRAME_GET_STRUCT_ADDR(argframe, tmptype);
 	      /* Decode the return value into the memory we allocated. */
 	      (*decoder) (-1, *(void**)retframe, tmptype, flags);
 	      break;
@@ -1018,10 +1418,10 @@ mframe_build_return (arglist_t argframe,
 	{
 	  /* Step through all the arguments, finding the ones that were
 	     passed by reference. */
-	  for (datum = method_types_get_next_argument(argframe, &tmptype, off), 
+	  for (datum = method_types_get_next_argument(argframe, &tmptype), 
 	       argnum=0;
 	       datum;
-	       (datum = method_types_get_next_argument(argframe, &tmptype, off)), 
+	       (datum = method_types_get_next_argument(argframe, &tmptype)), 
 	       argnum++)
 	    {
 	      /* Get the type qualifiers, like IN, OUT, INOUT, ONEWAY. */
@@ -1090,11 +1490,11 @@ mframe_build_return (arglist_t argframe,
     case _C_SHT:
     case _C_USHT:
 	return apply_short(*(short*)retframe);
-#if 1
+#if 0
     case _C_ARY_B:
     case _C_UNION_B:
     case _C_STRUCT_B:
-	if (off == 0 && objc_sizeof_type(rettype) > 8) {
+	if (objc_sizeof_type(rettype) > 8) {
 	    return apply_block(*(void**)retframe);
 	}
 #endif
@@ -1104,3 +1504,319 @@ mframe_build_return (arglist_t argframe,
   return retframe;
 }
 
+retval_t 
+mframe_build_return (arglist_t argframe, 
+		     const char *type, 
+		     BOOL out_parameters,
+		     void(*decoder)(int,void*,const char*,int))
+{
+  return mframe_build_return_opts(argframe,type,out_parameters,decoder,NO);
+}
+
+
+
+arglist_t
+mframe_create_argframe(const char *types, void** retbuf)
+{
+  arglist_t	argframe = objc_calloc(MFRAME_ARGS_SIZE, 1);
+  const char*	rtype = objc_skip_type_qualifiers(types);
+  int	stack_argsize = atoi(objc_skip_typespec(rtype));
+
+    /*
+     *	Allocate the space for variables passed on the stack.
+     */
+  if (stack_argsize)
+    {
+      argframe->arg_ptr = objc_calloc(stack_argsize, 1);
+    }
+  else
+    {
+      argframe->arg_ptr = 0;
+    }
+  if (*rtype == _C_STRUCT_B || *rtype == _C_UNION_B || *rtype == _C_ARY_B)
+    {
+      /*
+       *	If we haven't been passed a pointer to the location in which
+       *	to store a returned structure - allocate space and return
+       *	the address of the allocated space.
+       */
+      if (*retbuf == 0)
+	{
+	  *retbuf = objc_calloc(objc_sizeof_type(rtype), 1);
+	}
+      MFRAME_SET_STRUCT_ADDR(argframe, rtype, *retbuf);
+    }
+  return argframe;
+}
+
+void
+mframe_destroy_argframe(const char *types, arglist_t argframe)
+{
+  const char*	rtype = objc_skip_type_qualifiers(types);
+  int	stack_argsize = atoi(objc_skip_typespec(rtype));
+
+  if (stack_argsize)
+    {
+      objc_free(argframe->arg_ptr);
+    }
+  objc_free(argframe);
+}
+
+
+
+BOOL
+mframe_decode_return (const char *type, void* buffer, void* retframe)
+{
+  int	size = 0;
+
+  type = objc_skip_type_qualifiers(type);
+  NSGetSizeAndAlignment(type, &size, 0);
+
+  switch (*type)
+    {
+    case _C_ID:
+      {
+	inline id retframe_id(void *rframe)
+	{
+	  __builtin_return (rframe);
+	}
+	*(id*)buffer = retframe_id(retframe);
+	break;
+      }
+
+    case _C_CLASS:
+      {
+	inline Class retframe_Class(void *rframe)
+	{
+	  __builtin_return (rframe);
+	}
+	*(Class*)buffer = retframe_Class(retframe);
+	break;
+      }
+
+    case _C_SEL:
+      {
+	inline SEL retframe_SEL(void *rframe)
+	{
+	  __builtin_return (rframe);
+	}
+	*(SEL*)buffer = retframe_SEL(retframe);
+	break;
+      }
+
+    case _C_CHR:
+    case _C_UCHR:
+      {
+	inline unsigned char retframe_char(void *rframe)
+	{
+	  __builtin_return (rframe);
+	}
+	*(unsigned char*)buffer = retframe_char(retframe);
+	break;
+      }
+
+    case _C_SHT:
+    case _C_USHT:
+      {
+	inline unsigned short retframe_short(void *rframe)
+	{
+	  __builtin_return (rframe);
+	}
+	*(unsigned short*)buffer = retframe_short(retframe);
+	break;
+      }
+
+    case _C_INT:
+    case _C_UINT:
+      {
+	inline unsigned int retframe_int(void *rframe)
+	{
+	  __builtin_return (rframe);
+	}
+	*(unsigned int*)buffer = retframe_int(retframe);
+	break;
+      }
+
+    case _C_LNG:
+    case _C_ULNG:
+      {
+	inline unsigned long retframe_long(void *rframe)
+	{
+	  __builtin_return (rframe);
+	}
+	*(unsigned long*)buffer = retframe_long(retframe);
+	break;
+      }
+
+#ifdef	_C_LNG_LNG
+    case _C_LNG_LNG:
+    case _C_ULNG_LNG:
+      {
+	inline unsigned long long retframe_longlong(void *rframe)
+	{
+	  __builtin_return (rframe);
+	}
+	*(unsigned long long*)buffer = retframe_longlong(retframe);
+	break;
+      }
+#endif
+
+    case _C_FLT:
+      {
+	inline float retframe_float(void *rframe)
+	{
+	  __builtin_return (rframe);
+	}
+	*(float*)buffer = retframe_float(retframe);
+	break;
+      }
+
+    case _C_DBL:
+      {
+	inline double retframe_double(void *rframe)
+	{
+	  __builtin_return (rframe);
+	}
+	*(double*)buffer = retframe_double(retframe);
+	break;
+      }
+
+    case _C_PTR:
+    case _C_ATOM:
+    case _C_CHARPTR:
+      {
+	inline char* retframe_pointer(void *rframe)
+	{
+	  __builtin_return (rframe);
+	}
+	*(char**)buffer = retframe_pointer(retframe);
+	break;
+      }
+
+    case _C_ARY_B:
+    case _C_STRUCT_B:
+    case _C_UNION_B:
+      {
+	typedef struct {
+	  char	val[size];
+	} block;
+	inline block retframe_block(void *rframe)
+	{
+	  __builtin_return (rframe);
+	}
+	*(block*)buffer = retframe_block(retframe);
+	break;
+      }
+
+    case _C_VOID:
+      break;
+
+    default:
+      return NO;		/* Unknown type.	*/
+    }
+  return YES;
+}
+
+
+
+void*
+mframe_handle_return(const char* type, void* retval, arglist_t argframe)
+{
+  retval_t	retframe;
+  typedef struct { id many[8];} __big;
+  __big return_block (void* data)
+  {
+    return *(__big*)data;
+  }
+  /* For returning a char (or unsigned char) */
+  char return_char (char data)
+  {
+    return data;
+  }
+  /* For returning a double */
+  double return_double (double data)
+  {
+    return data;
+  }
+  /* For returning a float */
+  float return_float (float data)
+  {
+    return data;
+  }
+  /* For returning a short (or unsigned short) */
+  short return_short (short data)
+  {
+    return data;
+  }
+  retval_t apply_block(void* data)
+  {
+    void* args = __builtin_apply_args();
+    return __builtin_apply((apply_t)return_block, args, sizeof(void*));
+  }
+  retval_t apply_char(char data)
+  {
+    void* args = __builtin_apply_args();
+    return __builtin_apply((apply_t)return_char, args, sizeof(void*));
+  }
+  retval_t apply_float(float data)
+  {
+    void* args = __builtin_apply_args();
+    return __builtin_apply((apply_t)return_float, args, sizeof(float));
+  }
+  retval_t apply_double(double data)
+  {
+    void* args = __builtin_apply_args();
+    return __builtin_apply((apply_t)return_double, args, sizeof(double));
+  }
+  retval_t apply_short(short data)
+  {
+    void* args = __builtin_apply_args();
+    return __builtin_apply((apply_t)return_short, args, sizeof(void*));
+  }
+
+  retframe = alloca(MFRAME_RESULT_SIZE);
+
+  switch (*type)
+    {
+      case _C_VOID:
+	break;
+      case _C_CHR:
+      case _C_UCHR:
+	return apply_char(*(char*)retval);
+      case _C_DBL:
+	return apply_double(*(double*)retval);
+      case _C_FLT:
+	return apply_float(*(float*)retval);
+      case _C_SHT:
+      case _C_USHT:
+	return apply_short(*(short*)retval);
+      case _C_ARY_B:
+      case _C_UNION_B:
+      case _C_STRUCT_B:
+	{
+	  int    size = objc_sizeof_type(type);
+#if 1
+	  void	*dest;
+
+	  dest = MFRAME_GET_STRUCT_ADDR(argframe, type);
+	  memcpy(dest, retval, size);
+#else
+	  if (size > 8)
+	    {
+	      return apply_block(*(void**)retval);
+	    }
+	  else
+	    {
+	      memcpy(retframe, retval, size);
+	    }
+#endif
+	  break;
+	}
+      default:
+	memcpy(retframe, retval, objc_sizeof_type(type));
+	break;
+    }
+
+  return retframe;
+}
+
