@@ -1,6 +1,6 @@
 #import <Foundation/Foundation.h>
 #import <AppKit/AppKit.h>
-#import "IMLoading.h"
+#import "AppKit/IMLoading.h"
 
 @interface MyView : NSView
 @end
@@ -13,17 +13,18 @@
 }
 @end
 
-
 int main (int argc, char** argv, char** env)
 {
-  id pool;
-  NSArray* arguments;
-  NSProcessInfo* processInfo;
+id pool = [NSAutoreleasePool new];
+NSArray* arguments;
+NSProcessInfo* processInfo;
 
-  pool = [NSAutoreleasePool new];
 #if LIB_FOUNDATION_LIBRARY
   [NSProcessInfo initializeWithArguments:argv count:argc environment:env];
 #endif
+#ifndef NX_CURRENT_COMPILER_RELEASE
+  initialize_gnustep_backend();
+#endif
 
 #if 0
   processInfo = [NSProcessInfo processInfo];
@@ -34,10 +35,6 @@ int main (int argc, char** argv, char** env)
   }
 #endif
 
-#ifndef NX_CURRENT_COMPILER_RELEASE
-  initialize_gnustep_backend();
-#endif
-
 #if 0
   if (![GMModel loadIMFile:[arguments objectAtIndex:1]
 		 owner:[NSApplication sharedApplication]]) {
@@ -46,7 +43,7 @@ int main (int argc, char** argv, char** env)
   }
 #endif
 
-  [[NSDPSContext currentContext] wait];
+  [[GSContext currentContext] wait];
   [[NSApplication sharedApplication] run];
   printf ("exiting...\n");
 
