@@ -30,6 +30,7 @@
 #include <unistd.h>
 #include <sys/file.h>
 #include <fcntl.h>
+#include "config.h"
 
 #include <Foundation/NSDictionary.h>
 #include <Foundation/NSString.h>
@@ -42,20 +43,6 @@
 #include "AFMFileFontInfo.h"
 #include "fonts.h"
 
-#ifndef MAX
-#define MAX(a, b) \
-    ({typedef _ta = (a), _tb = (b); \
-	_ta _a = (a); _tb _b = (b); \
-	_a > _b ? _a : _b; })
-#endif
-
-#ifndef MIN
-#define MIN(a, b) \
-    ({typedef _ta = (a), _tb = (b); \
-	_ta _a = (a); _tb _b = (b); \
-	_a < _b ? _a : _b; })
-#endif
-
 /* This dictionary keeps global font info objects indexed by AFM file names.
    A global font info instance is the representation of AFM file; all the
    dimensions are in the 1x1 space not in the 1000x1000 space of the AFM file.
