@@ -62,7 +62,9 @@ NSString* NSMenuLocationsKey = @"NSMenuLocations";
   NSEvent* lastEvent = nil;
   id aCell;
   BOOL done = NO;
+  BOOL newlySelected = (selectedCell == nil);
   int index;
+  NSWindow* lastWindow;
 
   ASSIGN(lastEvent, theEvent);
 #if !USE_NEEDS_DISPLAY
@@ -74,8 +76,10 @@ NSString* NSMenuLocationsKey = @"NSMenuLocations";
 
     /* Convert the point from the event's window to our window. For the main
        menu window they are the same. */
-    lastLocation = [[lastEvent window] convertBaseToScreen:lastLocation];
-    lastLocation = [window convertScreenToBase:lastLocation];
+    if ((lastWindow = [lastEvent window]) != window) {
+      lastLocation = [[lastEvent window] convertBaseToScreen:lastLocation];
+      lastLocation = [window convertScreenToBase:lastLocation];
+    }
 
     NSDebugLog (@"location = (%f, %f)", lastLocation.x, lastLocation.y);
     lastLocation = [self convertPoint:lastLocation fromView:nil];
@@ -125,6 +129,7 @@ NSString* NSMenuLocationsKey = @"NSMenuLocations";
 	    /* If the current selected cell has a submenu open it */
 	    if ([selectedCell hasSubmenu]) {
 	      [[selectedCell target] display];
+	      newlySelected = YES;
 	    }
 	  }
 	}
@@ -232,23 +237,12 @@ NSString* NSMenuLocationsKey = @"NSMenuLocations";
   else {
     /* Set the state of the selected cell, if 1 this indicates the submenu is
        opened. */
-    if ([selectedCell hasSubmenu]
-	&& [[selectedCell target] hasAnAssociatedTornOffMenu])
+    if ([[selectedCell target] hasAnAssociatedTornOffMenu]||newlySelected==NO) {
       [selectedCell setState:0];
-    else
-      [selectedCell setState:![selectedCell state]];
-
-    if (![selectedCell state]) {
       [selectedCell highlight:NO withFrame:selectedCellRect inView:self];
-
       [[selectedCell target] close];
       ASSIGN(selectedCell, nil);
     }
-    else {
-      [selectedCell highlight:YES withFrame:selectedCellRect inView:self];
-
-      [[selectedCell target] display];
-    }
   }
 #if USE_NEEDS_DISPLAY
   [self setNeedsDisplayInRect:selectedCellRect];
@@ -294,6 +288,12 @@ static titleHeight = 0;
 
 @implementation PXKMenu
 
++ (void)initialize
+{
+	if (self == [PXKMenu class])
+		[self setVersion:1];								// Initial version
+}
+
 - (void)_createWindowWithFrame:(NSRect)winRect
 {
   be_menu_reserved = calloc (sizeof (PXKMenuPrivate), 1);
