@@ -35,6 +35,7 @@
 #include <Foundation/NSArray.h>
 #include <Foundation/NSValue.h>
 #include <Foundation/NSMapTable.h>
+#include <Foundation/NSNotification.h>
 
 #include <AppKit/NSCursor.h>
 
@@ -45,9 +46,11 @@
 #include <gnustep/xdps/PXKColor.h>
 #include <gnustep/xdps/PXKApplication.h>
 #include <gnustep/xdps/PXKMenu.h>
+#include <gnustep/xdps/PXKWindowView.h>
 
 #include "general.h"
 
+
 #ifndef LIB_FOUNDATION_LIBRARY
 
 static void __NSRetainNothing(void *table, const void *anObject)
@@ -90,6 +93,7 @@ static const NSMapTableValueCallBacks NSNonRetainedObjectMapValueCallBacks = {
 
 static NSMapTable* windowsForXWindows = NULL;
 static NSMapTable* windowsForNumbers = NULL;
+static BOOL _needsFlushWindows = YES;
 
 + (void)initialize
 {
@@ -116,6 +120,18 @@ static NSMapTable* windowsForNumbers = NULL;
   return NSMapGet (windowsForNumbers, (void*)windowNumber);
 }
 
++ (void)_setNeedsFlushWindows:(BOOL)flag
+{
+	_needsFlushWindows = flag;
+}
+
++ (BOOL)_needsFlushWindows		{ return _needsFlushWindows; }
+
++ (NSView *)_windowViewWithFrame:(NSRect)frameRect			// create the top
+{															// most view in the
+	return [[PXKWindowView alloc] initWithFrame:frameRect];	// view's heirarchy
+}
+
 - (id)initWithContentRect:(NSRect)contentRect
 		styleMask:(unsigned int)aStyle
 		  backing:(NSBackingStoreType)bufferingType
@@ -364,6 +380,65 @@ static NSMapTable* windowsForNumbers = NULL;
   [self _updateWindowGeometry];
 }
 
+- (void)sendEvent:(NSEvent *)theEvent
+{
+  [super sendEvent:theEvent];
+  [NSWindow _flushWindows];
+}
+
+- (BOOL)_needsFlush				{ return needs_flush; }
+
+- (void)_setNeedsFlush
+{
+//  NSLog (@"NSWindow _setNeedsFlush");
+	needs_flush = YES;
+	_needsFlushWindows = YES;
+}
+
+- (void)_setNeedsDisplay
+{
+//  NSLog (@"NSWindow setNeedsDisplay");
+	needs_display = YES;
+	[self _setNeedsFlush];
+}
+
+- (void)_collectFlushRectangles
+{
+PSMatrix* originMatrix;
+PSMatrix* sizeMatrix;
+
+	if (disable_flush_window || backing_type == NSBackingStoreNonretained)
+		return;
+
+	NSDebugLog (@"_collectFlushRectangles");
+	[_flushRectangles removeAllObjects];
+
+	originMatrix = [PSMatrix new];
+	sizeMatrix = [PSMatrix new];
+
+	[[content_view superview] 
+					_collectInvalidatedRectanglesInArray:_flushRectangles
+					originMatrix:originMatrix
+					sizeMatrix:sizeMatrix];
+
+	[originMatrix release];
+	[sizeMatrix release];
+}
+
+- (void)update
+{
+NSNotificationCenter *nc = [NSNotificationCenter defaultCenter];
+
+	if(is_autodisplay && needs_display)					// if autodisplay is
+		{												// enabled and window
+		[self _collectFlushRectangles];					// is marked as needing
+		[self displayIfNeeded];							// display
+		[self flushWindowIfNeeded];
+    	}
+
+	[nc postNotificationName: NSWindowDidUpdateNotification object: self];
+}
+
 + (BOOL)_flushWindows
 {
   Display* xDisplay;
@@ -396,6 +471,14 @@ static NSMapTable* windowsForNumbers = NULL;
   return YES;
 }
 
+- (void)displayIfNeeded
+{
+  if (needs_display) {
+    [[content_view superview] _displayNeededViews];
+    needs_display = NO;
+  }
+}
+
 - (void)flushWindow
 {
   int x, y, width, height;
@@ -439,7 +522,9 @@ static NSMapTable* windowsForNumbers = NULL;
     XCopyArea (xDisplay, xPixmap, XWINDOW, XGC, x, y, width, height, x, y);
   }
 
-  [super flushWindow];
+  needs_flush = NO;
+  [_flushRectangles removeAllObjects];
+  [[content_view superview] _recursivelyResetNeedsDisplayInAllViews];
 }
 
 - (void)captureMouse:sender
@@ -757,4 +842,21 @@ static NSMapTable* windowsForNumbers = NULL;
 - (void)setXExposed:(BOOL)flag	{ XEXPOSE = flag; }
 - (BOOL)isXExposed		{ return XEXPOSE; }
 
+- (void)_view:(NSView*)view needsFlushInRect:(NSRect)rect
+{
+  [self _setNeedsFlush];
+
+  /* Convert the rectangle to the window coordinates */
+  rect = [view convertRect:rect toView:nil];
+
+  /* If the view is rotated then compute the bounding box of the rectangle in
+     the window's coordinates */
+  if ([view isRotatedFromBase])
+    rect = [view _boundingRectFor:rect];
+
+  /* TODO: Optimize adding rect to the already existing arrays */
+
+  [_flushRectangles addObject:[NSValue valueWithRect:rect]];
+}
+
 @end
